/*============================================================================/
/*= UTILITY ROUTINES (Non-Exportable Functions) =============================*/
/*===========================================================================*/
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "rsafq.h"
#include "rsafq_utility.h"

#define IEEE_ERROR_BITS         0x3C    /* AND mask for ESR checking */
#define BUFFER_SIZE             2048L    /* File I/O buffer size */

/*===========================================================================*/
/* Function: Array Item to Const                                             */
/* Purpose:  Search for item in array and return index. Last element of array*/
/*           must be NULL pointer.                                           */
/*===========================================================================*/
ViStatus rsafq_itemToConst (ViChar **array, ViString buffer, ViPInt32 itemNo)
{
    for ( *itemNo = 0; array[*itemNo]; (*itemNo)++ )
        if (!strcmp (buffer, array[*itemNo]))
            break;

    return (!array[*itemNo]) ? VI_ERROR_INSTR_INTERPRETING_RESPONSE : VI_SUCCESS;
}

/*===========================================================================*/
/* Function: Boolean Value Out Of Range - ViBoolean                          */
/* Purpose:  This function checks a Boolean to see if it is equal to VI_TRUE */
/*           or VI_FALSE. If the value is out of range, the return value is  */
/*           VI_TRUE, otherwise the return value is VI_FALSE.                */
/*===========================================================================*/
ViBoolean rsafq_invalidViBooleanRange (ViBoolean val)
{
    return (ViBoolean)((val != VI_FALSE && val != VI_TRUE) ? VI_TRUE : VI_FALSE);
}

/*===========================================================================*/
/* Function: Long Signed Integer Value Out Of Range - ViInt32                */
/* Purpose:  This function checks a long signed integer value to see if it   */
/*           lies between a minimum and maximum value.  If the value is out  */
/*           of range, the return value is VI_TRUE, otherwise the return     */
/*           value is VI_FALSE.                                              */
/*===========================================================================*/
ViBoolean rsafq_invalidViInt32Range (ViInt32 val, ViInt32 min, ViInt32 max)
{
    return (ViBoolean)((val < min || val > max) ? VI_TRUE : VI_FALSE);
}

/*===========================================================================*/
/* Function: Long Unsigned Integer Value Out Of Range - ViUInt32             */
/* Purpose:  This function checks a long signed integer value to see if it   */
/*           lies between a minimum and maximum value.  If the value is out  */
/*           of range, the return value is VI_TRUE, otherwise the return     */
/*           value is VI_FALSE.                                              */
/*===========================================================================*/
ViBoolean rsafq_invalidViUInt32Range (ViUInt32 val, ViUInt32 min, ViUInt32 max)
{
    return (ViBoolean)((val < min || val > max) ? VI_TRUE : VI_FALSE);
}

/*===========================================================================*/
/* Function: Real (Double) Value Out Of Range - ViReal64                     */
/* Purpose:  This function checks a real (double) value to see if it lies    */
/*           between a minimum and maximum value.  If the value is out of    */
/*           range, the return value is VI_TRUE, otherwise the return value  */
/*           is VI_FALSE.                                                    */
/*===========================================================================*/
ViBoolean rsafq_invalidViReal64Range (ViReal64 val, ViReal64 min, ViReal64 max)
{
    return (ViBoolean)((val < min || val > max) ? VI_TRUE : VI_FALSE);
}

/*===========================================================================*/
/* Function: Check String                                                    */
/* Purpose:  This function checks a string value to see if it is             */
/*           of a defined length and initialized.                            */
/*===========================================================================*/
ViBoolean rsafq_invalidViString (ViString val, ViUInt32 maxLength)
{
    if (val)
        return (ViBoolean)((strlen (val) > maxLength) ? VI_TRUE : VI_FALSE);

    return VI_TRUE;
}

/*====*===========================================================================*/
/* Function: Wait On VISA Event                                                   */
/* Purpose:  This function waits for the occurrence of an enabled VISA Event. The */
/*           event type must be enabled before entering this function. Any        */
/*           programmatic commands, actions, or conditions, necessary to generate */
/*           the specified event type, must be satisified before entering this    */
/*           function. The function will wait for the specified timeout and then  */
/*           return. If the specified event is received then the function will    */
/*           return VI_SUCCESS (0), otherwise the status code for the generated   */
/*           error will be returned. In either case the function will disable the */
/*           event type and deallocate the event handle passed from viWaitOnEvent */
/*           before returning. WARNING: If VI_TMO_INFINITE is passed in as the    */
/*           timeout this function WILL NOT return until the specified event is   */
/*           received, if the event is not received the function will not return  */
/*           and it will be necessary to terminate CVI in order to regain control.*/
/*====*===========================================================================*/
ViStatus rsafq_waitOnVisaEvent (ViSession instrSession,
                                 ViEventType eventTypeIn,
                                 ViUInt32 timeoutIn, ViPUInt16 STB)
{
    ViStatus    rsafq_status = VI_SUCCESS,
                tempStatus = VI_SUCCESS;
    ViEventType eventTypeOut = 0;
    ViEvent     eventHandle = 0;

    /* For debug purposes we want to be able to see the status returned by        */
    /* viDisableEvent and viClose if one of the internal functions fails but do   */
    /* not want to return that value as that is not where the initial error       */
    /* occurs in the function, so we assign it to tempStatus.                     */

    if ((rsafq_status = viWaitOnEvent (instrSession, eventTypeIn, timeoutIn,
                                        &eventTypeOut, &eventHandle)) < 0)
        {
        tempStatus = viDisableEvent (instrSession, eventTypeIn, VI_QUEUE);
        return rsafq_status;
        }

    if (eventTypeIn == VI_EVENT_SERVICE_REQ)
        if ((rsafq_status = viReadSTB (instrSession, STB)) < 0)
            {
            tempStatus = viClose (eventHandle);
            tempStatus = viDisableEvent (instrSession, eventTypeIn, VI_QUEUE);
            return rsafq_status;
            }

    if ((rsafq_status = viClose (eventHandle)) < 0)
        {
        tempStatus = viDisableEvent (instrSession, eventTypeIn, VI_QUEUE);
        return rsafq_status;
        }

    if ((rsafq_status = viDisableEvent (instrSession, eventTypeIn, VI_QUEUE)) < 0)
        return rsafq_status;

    return rsafq_status;
}

/*=========================================================================*/
/* Function: Check Status                                                  */
/* Purpose:  This function reads Event Status Register to check for        */
/*           possible errors.                                              */
/*=========================================================================*/
ViStatus rsafq_checkStatus (ViSession instrSession)
{
    ViStatus            rsafq_status = VI_SUCCESS;
    rsafq_instrRange   instrPtr;

    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));

    /* Error Checking Enabled? */
    if (instrPtr -> errorChecking)
        {
        ViInt32 esrByte = 0;

        /*--------------- Read the Standard Event Register ------------------*/
        CHECKERR (viQueryf (instrSession, "*ESR?\n", "%ld", &esrByte));

        /*---------------- Check if any error or message bit was asserted ---*/
        rsafq_status = (esrByte & IEEE_ERROR_BITS) ? RSAFQ_ERROR_INSTRUMENT_ERROR : VI_SUCCESS;
        }

    return rsafq_status;
}

/*===========================================================================*/
/* Function: Wait for OPC                                                    */
/* Purpose:  This function sends a command to the instrument and waits until */
/*           the instrument response is available (queries MAV in STB).      */
/*===========================================================================*/
ViStatus rsafq_waitOPC (ViSession instrSession, ViPString command,
                ViPString instrResp, ViInt32 timeout)
{
    ViStatus    rsafq_status    = VI_SUCCESS;
    ViUInt16    response        = 0;
    ViInt32     counter         = 0;
    ViInt32     oldTimeout      = 0;
    ViUInt16    intf       = VI_INTF_GPIB;

    CHECKERR (viGetAttribute (instrSession, VI_ATTR_INTF_TYPE, &intf));

    /* Interface specific operations */
    switch (intf)
        {
        case VI_INTF_ASRL: /* Serial intf */
            CHECKERR (viGetAttribute (instrSession, VI_ATTR_TMO_VALUE, &oldTimeout));
            CHECKERR (viSetAttribute (instrSession, VI_ATTR_TMO_VALUE, timeout));
            if ((rsafq_status = viQueryf (instrSession, "%s\n", "%[^\r\n]", command, instrResp)) < 0)
                {
                viSetAttribute (instrSession, VI_ATTR_TMO_VALUE, oldTimeout);
                return rsafq_status;
                }
            CHECKERR (viSetAttribute (instrSession, VI_ATTR_TMO_VALUE, oldTimeout));
        break;
        default: /* Default case */
            CHECKERR (viPrintf (instrSession, "%s\n", command));

            while (!(response & 0x10)) /* (queries MAV in STB) */
                {

                CHECKERR (viReadSTB (instrSession, &response));

                rsafq_delay (instrSession, 0.005);
                counter+= 5;
                if (counter >= timeout)
                    return VI_ERROR_TMO;
                }

            CHECKERR (viScanf (instrSession, "%[^\r\n]", instrResp));
        break;
        }

    return rsafq_status;
}

/*===========================================================================*/
/* Function: Send Command                                                    */
/* Purpose:  This function sends command to the instrument with error check. */
/*===========================================================================*/
ViStatus rsafq_sendCmd (ViSession instrSession, ViString cmd)
{
    ViStatus    rsafq_status    = VI_SUCCESS;
    size_t      length          = 0;
    ViChar      *pbuffer;

    length = strlen (cmd);
    pbuffer = (ViChar*) malloc(length + 1);
    memcpy (pbuffer, cmd, length);
    pbuffer[length] = '\n';

    GOCHECKERR (viWrite (instrSession, (ViPBuf) cmd, (ViUInt32)length + 1, VI_NULL));
    GOCHECKERR (rsafq_checkStatus (instrSession));

ERROR_LABEL:

    free(pbuffer);

    return rsafq_status;
}

/*===========================================================================*/
/* Function: Set ViInt32                                                     */
/* Purpose:  This function sets command with integer parameter.              */
/*===========================================================================*/
ViStatus rsafq_setViInt32 (ViSession instrSession,
                           ViString cmd,
                           ViInt32 value,
                           ViInt32 min,
                           ViInt32 max,
                           ViStatus parameterErrorCode)
{
    ViChar      wrt_buf[BUFFER_SIZE];

    if ((max - min) > 0)
        if (rsafq_invalidViInt32Range (value, min, max))
            return parameterErrorCode;

    sprintf (wrt_buf, "%s %ld", cmd, value);

    return rsafq_sendCmd (instrSession, wrt_buf);
}

/*===========================================================================*/
/* Function: Get ViInt32                                                     */
/* Purpose:  This function gets command with integer parameter.              */
/*===========================================================================*/
ViStatus rsafq_getViInt32 (ViSession instrSession,
                           ViString cmd,
                           ViInt32 *value,
                           ViStatus parameterErrorCode)
{
    ViStatus    rsafq_status = VI_SUCCESS;

    if (value == VI_NULL)
        return parameterErrorCode;

    CHECKERR( viPrintf(instrSession, "%s\n", cmd));
    CHECKERR( viScanf(instrSession, "%ld", value));
    CHECKERR( rsafq_checkStatus (instrSession));

    return rsafq_status;
}

/*===========================================================================*/
/* Function: Set ViInt32 in Hexadecimal                                      */
/* Purpose:  This function sets command with integer parameter.              */
/*===========================================================================*/
ViStatus rsafq_setViInt32Hex (ViSession instrSession,
                              ViString cmd,
                              ViInt32 value,
                              ViInt32 min,
                              ViInt32 max,
                              ViStatus parameterErrorCode)
{
    ViChar      wrt_buf[BUFFER_SIZE];

    if ((max - min) > 0)
        if (rsafq_invalidViInt32Range (value, min, max))
            return parameterErrorCode;

    sprintf (wrt_buf, "%s #H%lX", cmd, value);

    return rsafq_sendCmd (instrSession, wrt_buf);
}

/*===========================================================================*/
/* Function: Get ViInt32 in Hexadecimal                                      */
/* Purpose:  This function gets command with integer parameter.              */
/*===========================================================================*/
ViStatus rsafq_getViInt32Hex (ViSession instrSession,
                           ViString cmd,
                           ViInt32 *value,
                           ViStatus parameterErrorCode)
{
    ViStatus    rsafq_status = VI_SUCCESS;

    if (value == VI_NULL)
        return parameterErrorCode;

    CHECKERR( viPrintf(instrSession, "%s\n", cmd));
    CHECKERR( viScanf(instrSession, "#H%lX", value));
    CHECKERR( rsafq_checkStatus (instrSession));

    return rsafq_status;
}

/*===========================================================================*/
/* Function: Set ViUInt32                                                    */
/* Purpose:  This function sets command with unsigned integer parameter.     */
/*===========================================================================*/
ViStatus rsafq_setViUInt32 (ViSession instrSession,
                           ViString cmd,
                           ViUInt32 value,
                           ViUInt32 min,
                           ViUInt32 max,
                           ViStatus parameterErrorCode)
{
    ViChar      wrt_buf[BUFFER_SIZE];

    if ((max - min) > 0)
        if (rsafq_invalidViUInt32Range (value, min, max))
            return parameterErrorCode;

    sprintf (wrt_buf, "%s %u", cmd, value);

    return rsafq_sendCmd (instrSession, wrt_buf);
}

/*===========================================================================*/
/* Function: Get ViUInt32                                                    */
/* Purpose:  This function gets command with unsigned integer parameter.     */
/*===========================================================================*/
ViStatus rsafq_getViUInt32 (ViSession instrSession,
                           ViString cmd,
                           ViUInt32 *value,
                           ViStatus parameterErrorCode)
{
    ViStatus    rsafq_status = VI_SUCCESS;

    if (value == VI_NULL)
        return parameterErrorCode;

    CHECKERR( viPrintf(instrSession, "%s\n", cmd));
    CHECKERR( viScanf(instrSession, "%u", value));
    CHECKERR( rsafq_checkStatus (instrSession));

    return rsafq_status;
}

/*===========================================================================*/
/* Function: Set ViBoolean                                                   */
/* Purpose:  This function sets command with integer parameter.              */
/*===========================================================================*/
ViStatus rsafq_setViBoolean (ViSession instrSession,
                           ViString cmd,
                           ViBoolean value,
                           ViStatus parameterErrorCode)
{
    if (rsafq_invalidViBooleanRange (value))
        return parameterErrorCode;

    return rsafq_setViString (instrSession,
                              cmd,
                              value ? "ON" : "OFF",
                              0);
}

/*===========================================================================*/
/* Function: Get ViBoolean                                                   */
/* Purpose:  This function gets command with integer parameter.              */
/*===========================================================================*/
ViStatus rsafq_getViBoolean (ViSession instrSession,
                           ViString cmd,
                           ViBoolean *value,
                           ViStatus parameterErrorCode)
{
    ViStatus    rsafq_status;
    ViInt32     bool_val;

    if (value == NULL)
        return parameterErrorCode;

    CHECKERR (rsafq_getViInt32(instrSession,
                               cmd,
                               &bool_val,
                               parameterErrorCode));
    *value = (ViBoolean) bool_val;

    return rsafq_status;
}

/*===========================================================================*/
/* Function: Set ViReal64                                                    */
/* Purpose:  This function sets command with double parameter.               */
/*===========================================================================*/
ViStatus rsafq_setViReal64 (ViSession instrSession,
                            ViString cmd,
                            ViReal64 value,
                            ViReal64 min,
                            ViReal64 max,
                            ViStatus parameterErrorCode)
{
    ViChar      buffer[BUFFER_SIZE];

    /* do range checking only if there's difference between max and min */
    if ((max - min) > 1.0e-14)
        if (rsafq_invalidViReal64Range (value, min, max))
            return parameterErrorCode;

    sprintf(buffer,"%s %f", cmd, value);

    return rsafq_sendCmd(instrSession, buffer);
}

/*===========================================================================*/
/* Function: Get ViReal64                                                    */
/* Purpose:  This function gets command with double parameter.               */
/*===========================================================================*/
ViStatus rsafq_getViReal64 (ViSession instrSession,
                            ViString cmd,
                            ViReal64 *value,
                            ViStatus parameterErrorCode)
{
    ViStatus    rsafq_status = VI_SUCCESS;

    if (value == VI_NULL)
        return parameterErrorCode;

    CHECKERR(viPrintf(instrSession, "%s\n",cmd));
    CHECKERR(viScanf (instrSession, "%le",value));
    CHECKERR(rsafq_checkStatus (instrSession));

    return rsafq_status;
}

/*===========================================================================*/
/* Function: Set ViReal64 with Units                                         */
/* Purpose:  This function sets command with double parameter.               */
/*===========================================================================*/
ViStatus rsafq_setViReal64Units (ViSession instrSession,
                                 ViString cmd,
                                 ViReal64 value,
                                 ViString units,
                                 ViReal64 min,
                                 ViReal64 max,
                                 ViStatus parameterErrorCode)
{
    ViChar      buffer[BUFFER_SIZE];

    /* do range checking only if there's difference between max and min */
    if ((max - min) > 1.0e-14)
        if (rsafq_invalidViReal64Range (value, min, max))
            return parameterErrorCode;

    sprintf(buffer,"%s %.12f %s",cmd, value, units);

    return rsafq_sendCmd(instrSession, buffer);
}

/*===========================================================================*/
/* Function: Set String                                                      */
/* Purpose:  This function sets command with string parameter.               */
/*===========================================================================*/
ViStatus rsafq_setViString (ViSession instrSession,
                            ViString cmd,
                            ViString value,
                            ViStatus parameterErrorValue)
{
    ViStatus    rsafq_status = VI_SUCCESS;
    ViChar*     wrt_buf;
    ViChar*     p2buf;

    if (value == VI_NULL)
        return parameterErrorValue;

    wrt_buf = (ViChar*) malloc (strlen(cmd) + strlen(value) + 4 );
    p2buf = wrt_buf;
    p2buf += sprintf(wrt_buf, "%s %s\n", cmd, value);

    GOCHECKERR(viWrite (instrSession, (ViBuf)wrt_buf, (ViUInt32)(p2buf - wrt_buf), VI_NULL));
    GOCHECKERR(rsafq_checkStatus(instrSession));

ERROR_LABEL:

    free(wrt_buf);

    return rsafq_status;
}

/*===========================================================================*/
/* Function: Get String                                                      */
/* Purpose:  This function gets command with string parameter.               */
/*===========================================================================*/
ViStatus rsafq_getViString (ViSession instrSession,
                            ViString cmd,
                            ViString value,
                            ViInt32 buffer_size,
                            ViStatus parameterErrorCode)
{
    ViStatus    rsafq_status    = VI_SUCCESS;
    ViChar      *pbuffer,
                *p2buffer;
    ViUInt32    count           = 0;
    size_t      length          = 0;

    length = strlen (cmd);
    pbuffer = (ViChar*) malloc (length + 1);
    memcpy (pbuffer, cmd, length);
    pbuffer[length] = '\n';

    if (!value)
        return parameterErrorCode;

    GOCHECKERR (viWrite (instrSession, (ViPBuf)pbuffer, (ViUInt32)length + 1, VI_NULL));

    free (pbuffer);

    GOCHECKERR (rsafq_readDataUnknownLength (instrSession, &pbuffer, &count));
    GOCHECKERR (rsafq_status = rsafq_checkStatus (instrSession));

    if (count > (ViUInt32)buffer_size)
        {
        rsafq_status = RSAFQ_WARNING_SMALL_BUFFER;
        count = buffer_size;
        memcpy (value, pbuffer, count);
        value[count] = '\0';
        }
    else
        if ((p2buffer = strpbrk (pbuffer, "\r\n")) != VI_NULL)
            {
            rsafq_status = VI_SUCCESS;
            memcpy (value, pbuffer, p2buffer - pbuffer);
            value[p2buffer - pbuffer] = '\0';
            }

ERROR_LABEL:

    free (pbuffer);

    return rsafq_status;
}

/*===========================================================================*/
/* Function: Set Quoted String                                               */
/* Purpose:  This function sets command with quoted string parameter.        */
/*===========================================================================*/
ViStatus rsafq_setViStringQuoted (ViSession instrSession,
                            ViString cmd,
                            ViString quotes,
                            ViString value,
                            ViStatus parameterErrorValue)
{
    ViStatus    rsafq_status = VI_SUCCESS;
    ViChar*     wrt_buf;
    ViChar*     p2buf;

    if (value == VI_NULL)
        return parameterErrorValue;

    wrt_buf = (ViChar*)malloc(strlen(cmd) + strlen(value) + 2*strlen(quotes) + 4);
    p2buf = wrt_buf;
    p2buf += sprintf(wrt_buf,"%s %s%s%s\n", cmd, quotes, value, quotes);

    GOCHECKERR (viWrite (instrSession, (ViBuf)wrt_buf, (ViUInt32)(p2buf - wrt_buf), VI_NULL));
    GOCHECKERR (rsafq_checkStatus(instrSession));

ERROR_LABEL:

    free (wrt_buf);

    return rsafq_status;
}

/*===========================================================================*/
/* Function: Get Quoted String                                               */
/* Purpose:  This function gets command with quoted string parameter.        */
/*===========================================================================*/
ViStatus rsafq_getViStringQuoted (ViSession instrSession,
                                  ViString cmd,
                                  ViString quotes,
                                  ViInt32  arraySize,
                                  ViChar   value[],
                                  ViStatus parameterErrorValue)
{
    ViStatus    rsafq_status = VI_SUCCESS;
    ViChar*     rd_buf;
    ViChar      fmt_string[BUFFER_SIZE];

    if (value == VI_NULL)
        return parameterErrorValue;

    rd_buf = (ViChar*)malloc(arraySize + 2*strlen(quotes) + 4);
    sprintf(fmt_string,"%s%s%s%s", cmd, quotes, value, quotes);

    {
        size_t      length      = 0;
        ViChar      *pbuffer;

        length = strlen (cmd);
        pbuffer = (ViChar*) malloc (length + 1);
        memcpy (pbuffer, cmd, length);
        pbuffer[length] = '\n';

        rsafq_status = viWrite (instrSession, (ViPBuf) cmd, (ViUInt32)length + 1, VI_NULL);

        free(pbuffer);
    }

    GOCHECKERR (rsafq_status);
    GOCHECKERR (viScanf (instrSession, fmt_string, value));
    GOCHECKERR (rsafq_checkStatus(instrSession));

ERROR_LABEL:

    free (rd_buf);

    return rsafq_status;
}

/*===========================================================================*/
/* Function: Set Enum                                                        */
/* Purpose:  This function sets command with enumerated parameter.           */
/*===========================================================================*/
ViStatus rsafq_setEnum (ViSession instrSession,
                        ViString cmd,
                        ViInt32 value,
                        ViString list[],
                        ViStatus parameterErrorCode)
{
    ViStatus rsafq_status = VI_SUCCESS;

    CHECKERR (rsafq_checkIndexInList (instrSession, value, list, parameterErrorCode));

    return rsafq_setViString (instrSession, cmd, list[value], parameterErrorCode);
}

/*===========================================================================*/
/* Function: Get Enum                                                        */
/* Purpose:  This function sets command with enumerated parameter.           */
/*===========================================================================*/
ViStatus rsafq_getEnum (ViSession instrSession,
                        ViString cmd,
                        ViInt32 *value,
                        ViString list[],
                        ViStatus parameterErrorCode)
{
    ViStatus    rsafq_status = VI_SUCCESS;
    ViChar      buffer[BUFFER_SIZE];

    CHECKERR( rsafq_getViString(instrSession, cmd, buffer, BUFFER_SIZE, parameterErrorCode));
    CHECKERR( rsafq_getIndexFromString (instrSession, buffer, list, value));

    return rsafq_status;
}

/*===========================================================================*/
/* Function: Check Index List                                                */
/* Purpose:  This function checks if the passed index is valid for given list*/
/*===========================================================================*/
ViStatus rsafq_checkIndexInList (ViSession instrSession,
                                 ViInt32 index,
                                 ViString list[],
                                 ViStatus statusIfError)

{
    ViInt32     i=0;
	instrSession;

    if (index < 0)
        return statusIfError;

    while (list[i++] != VI_NULL);
    if (index >= i)
        return statusIfError;

    return VI_SUCCESS;
}

/*===========================================================================*/
/* Function: Get Index From String                                           */
/* Purpose:  This function returns index for string.                         */
/*===========================================================================*/
ViStatus rsafq_getIndexFromString (ViSession instrSession,
                                   ViString string,
                                   ViString *list,
                                   ViInt32 *index)
{
    ViStatus    rsafq_status = VI_SUCCESS;
    ViUInt32    i;
    size_t      len = 0;
	instrSession;

    for (i = 0; list[i]; i++)
        if ((len = strlen (string)) == strlen (list[i])) {
            if (strncmp(string, list[i], strlen (string)) == 0) break;
        }

    *index = i;

    return (list[i]) ? rsafq_status : VI_ERROR_INSTR_INTERPRETING_RESPONSE;
}

/*===========================================================================*/
/* Function: Get Catalog                                                     */
/*===========================================================================*/
ViStatus rsafq_getCatalog (ViSession instrSession,
                           ViString cmd,
                           ViInt32 arraySize,
                           ViChar* array,
                           ViInt32 parameterErrorCode)
{
    ViStatus    rsafq_status = VI_SUCCESS;
    ViChar      *pbuffer,
                *p2buffer;
    ViUInt32    count           = 0;
    size_t      length          = 0;

    length = strlen (cmd);
    pbuffer = (ViChar*) malloc (length + 1);
    memcpy (pbuffer, cmd, length);
    pbuffer[length] = '\n';

    if (!array)
        return parameterErrorCode;

    GOCHECKERR (viWrite (instrSession, (ViPBuf)pbuffer, (ViUInt32)length + 1, VI_NULL));

    free(pbuffer);

    GOCHECKERR (rsafq_readDataUnknownLength (instrSession, &pbuffer, &count));
    GOCHECKERR (rsafq_checkStatus (instrSession));

    if (count > (ViUInt32)arraySize)
        {
        rsafq_status = RSAFQ_WARNING_SMALL_BUFFER;
        count = arraySize;
        memcpy (array, pbuffer, count);
        array[count-1] = '\0';
        }
    else
        {
        if ((p2buffer = strchr (pbuffer, ';')) != VI_NULL)
            {
            rsafq_status = VI_SUCCESS;
            memcpy (array, pbuffer, p2buffer - pbuffer);
            array[p2buffer - pbuffer] = '\0';
            }
        else
            {
            memcpy (array, pbuffer, count);
            array[count-1] = '\0';
            }
        }
ERROR_LABEL:

    free (pbuffer);

    return rsafq_status;
}

/*===========================================================================*/
/* Function: Get ViReal64 Array                                              */
/*===========================================================================*/
ViStatus rsafq_getViReal64Array (ViSession instrSession,
                            ViString cmd,
                            ViReal64 *value,
                            ViInt32  count,
                            ViStatus parameterErrorCode)
{
    ViStatus    rsafq_status = VI_SUCCESS;
    ViChar*     pbuffer = VI_NULL;
    ViChar*     p2buf;
    ViUInt32    local_count;
    ViInt32     i = 0;

    if (value == VI_NULL)
        return parameterErrorCode;

    GOCHECKERR (viPrintf(instrSession, "%s\n",cmd));
    GOCHECKERR (rsafq_readDataUnknownLength(instrSession, &pbuffer, &local_count));
    p2buf = pbuffer;

    p2buf = strtok(pbuffer, ",");

    while(p2buf && (i < count))
        {
        value[i++] = atof(p2buf);
        p2buf = strtok(NULL, ",");
        };

    GOCHECKERR (rsafq_checkStatus (instrSession));

ERROR_LABEL:

    free(pbuffer);

    return rsafq_status;
}

/*===========================================================================*/
/* Function: Get ViInt32 Array                                               */
/*===========================================================================*/
ViStatus rsafq_getViInt32Array (ViSession instrSession,
                            ViString cmd,
                            ViInt32 *value,
                            ViInt32  count,
                            ViStatus parameterErrorCode)
{
    ViStatus    rsafq_status = VI_SUCCESS;
    ViChar*     pbuffer = VI_NULL;
    ViChar*     p2buf;
    ViUInt32    local_count;
    ViInt32     i = 0;

    if (value == VI_NULL)
        return parameterErrorCode;

    GOCHECKERR (viPrintf(instrSession, "%s\n",cmd));
    GOCHECKERR (rsafq_readDataUnknownLength(instrSession, &pbuffer, &local_count));
    p2buf = pbuffer;

    p2buf = strtok(pbuffer, ",");

    while(p2buf && (i < count))
        {
        value[i++] = atol(p2buf);
        p2buf = strtok(NULL, ",");
        };

    GOCHECKERR (rsafq_checkStatus (instrSession));

ERROR_LABEL:

    free(pbuffer);

    return rsafq_status;
}

/*===========================================================================*/
/* Function: Convert hexa data ("#H...") to binary array ("10010...")        */
/*===========================================================================*/
ViStatus rsafq_readHexArrayToBin (ViSession instrSession,
                            ViString cmd,
                            ViChar   array[],
                            ViInt32  expectedStreamLength,
                            ViInt32  bufferSize)
{
   ViStatus    rsafq_status    = VI_SUCCESS;
    ViInt32     count           = 0;
    ViChar*     p2buf           = VI_NULL;
    ViChar*     psequence       = VI_NULL;
    ViChar*     p2seq           = VI_NULL;
    ViChar*     p2bufpat        = VI_NULL;
    ViChar      character[]     = {'0','\0'};
    ViChar      buffer[BUFFER_SIZE];
    ViChar      bufferPattern[BUFFER_SIZE];
    ViInt32     number          = 0;
    ViUInt32    local_count     = 0;
    ViInt32     idx             = 0;
    size_t     hexLength       = 0;
    ViChar*     pos             = VI_NULL;

    CHECKERR (viQueryf(instrSession, "%s\n", "%[^\r\n]", cmd, buffer));

    if (strncmp (buffer, "#H", 2) != 0)
        {
        return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
        }

    /* Create nibbles */
    p2buf = buffer + 2;

    sscanf(p2buf, "%[^,],%ld",bufferPattern, &count);
    
    p2bufpat = bufferPattern;
  
    hexLength = strlen (p2bufpat) * 4;
    p2seq = psequence = malloc (hexLength + 1);
    memset (psequence, 0, hexLength + 1);

    while(*p2bufpat && ((idx * 4) <= bufferSize))
        {
        memcpy (character, p2bufpat++, 1);
        number = strtol (character, VI_NULL, 16);
        *p2seq++ = (8 & number) ? '1' : '0';
        *p2seq++ = (4 & number) ? '1' : '0';
        *p2seq++ = (2 & number) ? '1' : '0';
        *p2seq++ = (1 & number) ? '1' : '0';
        idx++;
        };
    
    local_count = (count < expectedStreamLength) ? count : expectedStreamLength;
    

    buffer[0] = '\0';
    
    if (local_count == 0)
    {
        /* 
        If the instrument does NOT return the count of the pattern
        then here it's nessessary to get rid of the leading zeros '0'
        from the transformed hexa to bin string
        */
        strncpy(buffer, psequence, hexLength);
        buffer[hexLength] = '\0';
        pos = strchr (buffer, '1');
        strcpy (array, pos); 
    }
    else
    {
        /*
        If the instrument returns the count of the pattern
        then it's just copied into the returned parameter 
        */
        strncpy(array, psequence, local_count);
        array[local_count] = '\0';
    }

    free(psequence);

    if ((rsafq_status = rsafq_checkStatus (instrSession)) < VI_SUCCESS)
        {
        return rsafq_status;
        }
    else
        {
        rsafq_status = VI_SUCCESS;
        }

    return rsafq_status;
}

/*===========================================================================*/
/* Function: Read Data With Unknown length                                   */
/* Purpose:  This function allocates buffer and reads data from the          */
/*           instrument                                                      */
/* WARNING:  You must deallocate buffer, if any pointer has been returned.   */
/*           This function does not work correctly, when you read bin data   */
/*           via serial intf                                                 */
/*===========================================================================*/
ViStatus rsafq_readDataUnknownLength(ViSession instrSession,
                                     ViChar **outBuffer,
                                     ViUInt32 *count)
{
    ViStatus rsafq_status = VI_SUCCESS;
    ViUInt32 actSize, retCnt = 0;
    ViChar   *buffer = VI_NULL, *p2buffer;

    *outBuffer = VI_NULL;
    if (count) *count = 0;

    /* Allocate Buffer */
    actSize = BUFFER_SIZE;
    buffer = (ViChar *)calloc(actSize, sizeof(ViChar));
    if (buffer == VI_NULL) return VI_ERROR_ALLOC;

    p2buffer = buffer;
    do
        {
        if ((rsafq_status = viRead (instrSession, (ViPBuf)p2buffer, BUFFER_SIZE, &retCnt)) < 0)
            {
            free(buffer);
            *outBuffer = VI_NULL;
            return rsafq_status;
            }
        if (retCnt == BUFFER_SIZE) /* Buffer was small, reallocate it */
        {
            actSize += BUFFER_SIZE;
            buffer = (ViChar *)realloc(buffer, actSize);
            if (buffer == VI_NULL) return VI_ERROR_ALLOC;
            p2buffer = buffer + (actSize - BUFFER_SIZE); /* Set pointer to end of data in rallocated buffer */
        }
        else
        {
            *(p2buffer+retCnt) = '\0'; /* For sure set end of string at the end of data */
        }
       } while (rsafq_status == VI_SUCCESS_MAX_CNT); /* if buffer was small, continue with next read */

    *outBuffer = buffer;
    if (count) *count = (actSize - BUFFER_SIZE) + retCnt;

    return rsafq_status;
}

/*===========================================================================*/
/* Function: Check Instrument                                                */
/* Purpose:  This function checks presence of selected instrument. According */
/*           to generateErrorIfPresent value it either generates error or    */
/*           returns success if instrumet is present.                        */
/*===========================================================================*/
ViStatus rsafq_checkInstrument (ViSession instrSession, ViInt32 instrument, ViBoolean generateErrorIfPresent)
{
    ViStatus            rsafq_status = VI_SUCCESS;
    rsafq_instrRange    instrPtr;
    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));

    if ((instrPtr->errorChecking != 0) && (instrPtr->optionChecking != 0))
    {
        if (generateErrorIfPresent != VI_FALSE)
        {
            if (instrPtr->instrVer == instrument)
                rsafq_status = RSAFQ_ERROR_INSTRUMENT_NSUP_MODEL;
        }
        else
        {
            if (instrPtr->instrVer != instrument)
                rsafq_status = RSAFQ_ERROR_INSTRUMENT_NSUP_MODEL;
        }
    }

    return rsafq_status;
}
