/* R&S AFQ I/Q Modulation Generator =========================================*/
/* LabWindows/CVI 8.0 VXIpnp Instrument Driver
/* Original Release: May 2006
/* By: Jiri Kominek
/*
/* Should you have any technical questions please contact the hotline of
/* Rohde & Schwarz Vertriebs-GmbH Rohde & Schwarz Support Center
/*
/* e-mail: CustomerSupport@rsd.rohde-schwarz.com
/*
/* Modification History:
/*
/*   see file rsafq_vxi.chm
/*===========================================================================*/
#ifdef _WIN32
    #ifdef _CVI_
        #include <userint.h>
    #else
        #include <windows.h>
    #endif
#endif
#include <math.h>
#include <visa.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include "rsafq.h"
#include "rsafq_utility.h"
#include <limits.h>
#include <locale.h>  

/*= Macros ==================================================================*/
#define CHECKERR(fCal) if (rsafq_status = (fCal), rsafq_status < VI_SUCCESS) \
    return rsafq_status; else

#define CHECKERR_CU(fCal) if (rsafq_status = (fCal), rsafq_status < VI_SUCCESS) \
    return rsafq_initCleanUp (rmSession, instrSession, rsafq_status); else

/*= DEFINES =================================================================*/
#define RSAFQ_REVISION          "Rev 2.09.02, 04/2015" /* Instrument driver revision */
#define BUFFER_SIZE             2048L    /* File I/O buffer size */

#define RSAFQ_DEFAULT_TIMEOUT   10000   /* Default Timeout value in ms (10 sec) */

/*****************************************************************************/
/*= INSTRUMENT-DEPENDENT COMMAND ARRAYS =====================================*/
/*****************************************************************************/
static ViString triggerStatusArr[] = {"STOP","RUN", VI_NULL};
static ViString triggerModeArr[] = {"CONT","SING", VI_NULL};
static ViString triggerSourceArr[] = {"BUS","EXT","MAN","AUTO", VI_NULL};
static ViString triggerInputImpedanceArr[] =  {"G50","G1K", VI_NULL};
static ViString inputSlopeArr[] = {"NEG","POS", VI_NULL};
static ViString markerTypeArr[] = {"UNCH","REST","PULS","PATT","RAT","LIST", VI_NULL};
static ViString multiSegmentTriggerModeArr[] = {"CONT","SEAM","SING","AUTO", VI_NULL};
static ViString segmentTriggerSourceArr[] = {"BUS","MAN","EXT","AUTO", VI_NULL};
static ViString sourceArr[] = {"INT","EXT", VI_NULL};
static ViString roscSynchronizationBandwidthArr[] = {"WIDE","NARR", VI_NULL};
static ViString frequencyEntryModeArr[] = {"ABS","REL", VI_NULL};
static ViString activeOutputArr[] = {"ANAL","DIG","ANALOG1","ANALOG2","DIGITAL1","DIGITAL2","BBO", VI_NULL};
static ViString typeArr[] = {"BAL","UNB", VI_NULL};
static ViString amplitudeModeArr[] = {"ELEC","ATT", VI_NULL};
static ViString outputImpedanceArr[] = {"R50","HIGH", VI_NULL};
static ViString resolutionArr[] = {"BIT6","BIT7","BIT8","BIT9","BIT10","BIT11","BIT12","BIT13","BIT14","BIT15","BIT16", VI_NULL};
static ViString portArr[] = {"PORT1","PORT2", VI_NULL};
static ViString iMapstoArr[] = {"SIGN","MARK", VI_NULL};
static ViString bertModeArr[] = {"AUTO","SING", VI_NULL};
static ViString prbsArr[] = {"PRBS9","PRBS11","PRBS15","PRBS16","PRBS20","PRBS21","PRBS23", VI_NULL};
static ViString bertClockSlopeArr[] = {"RIS","FALL", VI_NULL};
static ViString dataPolarityArr[] = {"NORM","INV", VI_NULL};
static ViString dataEnableArr[] = {"OFF","HIGH","LOW", VI_NULL};
static ViString patternIgnoreArr[] = {"ONE","OFF","ZERO", VI_NULL};
static ViString impedanceArr[] = {"G50","G1K", VI_NULL};
static ViString displayUnitArr[] = {"ENG","SCI","PCT","PPM", VI_NULL};
static ViString fanSpeedArr[] = {"AUTO","FSP", VI_NULL};
static ViString regGroupArr[] = {"OPER","QUES", VI_NULL};
static ViString regTypeArr[] = {"COND","EVEN","ENAB","NTR","PTR", VI_NULL};
static ViString instrVersionList[] = {"AFQ100A","AFQ100B","UNKNOWN", VI_NULL};
static ViString nextWaveformSegmentModeArr[] = {"AINC","FPAR","FSER", VI_NULL};
static ViString ROSCSynchronizationModeArr[] = {"NORM","MAST","SLAV","LAST", VI_NULL};

/*****************************************************************************/
/*= UTILITY ROUTINE DECLARATIONS (Non-Exportable Functions) =================*/
/*****************************************************************************/
ViStatus rsafq_initCleanUp (ViSession openRMSession,
                             ViPSession openInstrSession,
                             ViStatus currentStatus);
ViStatus rsafq_defaultInstrSetup (ViSession openInstrSession);

/*****************************************************************************
 * Function: Initialize
 * Purpose:  This function performs the following initialization actions:
 *           
 *           - Opens a session to the Default Resource Manager resource and a 
 *           session to the specified device using the intf and address 
 *           specified in the Resource_Name control.
 *           
 *           - Performs an identification query on the Instrument.
 *           
 *           - Resets the instrument to a known state.
 *           
 *           - Sends initialization commands to the instrument that set any 
 *           necessary programmatic variables such as Headers Off, Short 
 *           Command form, and Data Transfer Binary to the state necessary for 
 *           the operation of the instrument driver.
 *           
 *           - Returns an Instrument Handle which is used to differentiate 
 *           between different sessions of this instrument driver.
 *           
 *           - Each time this function is invoked a Unique Session is opened.  
 *           It is possible to have more than one session open for the same 
 *           resource.
 *           
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_init(
    ViRsrc  resourceName,
    ViBoolean   idQuery,
    ViBoolean   resetDevice,
    ViSession*  instrSession
)
{
{
    ViStatus            rsafq_status = VI_SUCCESS;
    ViSession           rmSession = 0;
    ViChar              buffer[BUFFER_SIZE];
    rsafq_instrRange    instrPtr = NULL;

    /* --- Set standard "C" locale --- */
    {
       char *locale;

       locale = setlocale(LC_ALL, "C");
    }
    
    /*- Check input parameter ranges ----------------------------------------*/
    CHECKPARAM (rsafq_invalidViBooleanRange (idQuery), 2);
    CHECKPARAM (rsafq_invalidViBooleanRange (resetDevice), 3)

    /*- Open instrument session ---------------------------------------------*/
    CHECKERR( viOpenDefaultRM (&rmSession));

    if ((rsafq_status = viOpen (rmSession, resourceName, VI_LOAD_CONFIG,
                                 VI_NULL, instrSession)) < 0)
    {
        viClose (rmSession);
        return rsafq_status;
    }

    /* Determine if the structure has been initialized for the current VISA  */
    /* Session and malloc if it has not.                                     */

    if (instrPtr == NULL)
        instrPtr = calloc (1, sizeof (struct rsafq_statusDataRanges));
        

    CHECKERR( viSetAttribute (*instrSession, VI_ATTR_USER_DATA, (ViAttrState)instrPtr));

    //With remote control via Ethernet, remote control is not automatically
    //set by means of a command. The instrument must be explicitly set to
    //the REMOTE state, e.g. by sending the intf command &GTR
    CHECKERR (viPrintf (*instrSession, "&GTR\n"));

    /*- Configure VISA Formatted I/O ----------------------------------------*/
    CHECKERR( viSetAttribute (*instrSession, VI_ATTR_TMO_VALUE,
                                 RSAFQ_DEFAULT_TIMEOUT));
    CHECKERR( viSetBuf (*instrSession, VI_READ_BUF|VI_WRITE_BUF, 4000));
    CHECKERR( viSetAttribute (*instrSession, VI_ATTR_WR_BUF_OPER_MODE,
                                 VI_FLUSH_ON_ACCESS));
    CHECKERR( viSetAttribute (*instrSession, VI_ATTR_RD_BUF_OPER_MODE,
                                 VI_FLUSH_ON_ACCESS));

    /* -------------------------------------------------------------------- */
    /* Drawback of Agilent VISA is patched by adding extra termination      */
    /* character "\n" to each I/O operation                                 */
    /*                                                                      */
    /* VI_ATTR_SEND_END_EN will stay with this attribute set to VI_TRUE     */
    /* -------------------------------------------------------------------- */

    CHECKERR( viSetAttribute (*instrSession, VI_ATTR_SEND_END_EN, VI_TRUE));

    // read/reset ESR which may still indicate device dependent error from 
    // previous handling!
    (void) rsafq_checkStatus(*instrSession);
    
    /*- Identification Query ------------------------------------------------*/
    if (idQuery)
    {
        CHECKERR( viQueryf (*instrSession, "*IDN?\n", "%s",
                               buffer));

        /* Check if instrument is supported or not */
        if (strncmp (buffer, "Rohde&Schwarz,AFQ", 17))
            return rsafq_initCleanUp (rmSession, instrSession, VI_ERROR_FAIL_ID_QUERY);
    }

    /*- Reset instrument ----------------------------------------------------*/
    if (resetDevice)
        CHECKERR( rsafq_reset (*instrSession));
    else  /*- Send Default Instrument Setup ---------------------------------*/
        CHECKERR( rsafq_defaultInstrSetup (*instrSession));

    return rsafq_status;
}
}

/*****************************************************************************
 * Function: Set ARB State
 * Purpose:  This function switches ARB modulation on or off. 
 *           
 *           The output is based on the waveform file that is loaded. The name 
 *           of the waveform file is indicated in the header section of the 
 *           display.
 *           
 *           If no waveform file is loaded, ARB modulation cannot be powered 
 *           up. None will be displayed next to Load Waveform.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetARBState(
    ViSession   instrSession,
    ViBoolean   state
)
{
    return rsafq_setViBoolean (instrSession,
                               "SOUR:STAT",
                               state,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get ARB State
 * Purpose:  This function returns the ARB modulation state
 *           
 *           The output is based on the waveform file that is loaded. The name 
 *           of the waveform file is indicated in the header section of the 
 *           display.
 *           
 *           If no waveform file is loaded, ARB modulation cannot be powered 
 *           up. None will be displayed next to Load Waveform.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetARBState(
    ViSession   instrSession,
    ViBoolean*  state
)
{
    return rsafq_getViBoolean (instrSession,
                               "SOUR:STAT?",
                               state,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: ARB Set To Default
 * Purpose:  This function sets all ARB generator parameters to their default 
 *           values. The default settings are listed in section "Default 
 *           instrument settings" of chapter 4.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ARBSetToDefault(
    ViSession   instrSession
)
{
    return rsafq_sendCmd (instrSession, ":PRES");
}

/*****************************************************************************
 * Function: Get Crest Factor
 * Purpose:  This function queries the crest factor of the generated signal. 
 *           The crest factor of a waveform is equal to the peak amplitude 
 *           divided by the root mean square value. The program calculates the 
 *           power ratio, not the voltage ratio. Hence, the crest factor of a 
 *           sine function amounts to 2 instead of sqrt(2), equal to 3.01 dB.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetCrestFactor(
    ViSession   instrSession,
    ViReal64*   crestFactor
)
{
    return rsafq_getViReal64 (instrSession,
                              ":CRES?",
                              crestFactor,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: ARB Waveform Catalog
 * Purpose:  This function reads out the files in the default directory. The 
 *           default directory is set using command MMEM:CDIRectory. When the 
 *           names of the waveform files are returned they are separated by 
 *           commas. Only files with the file extension *.wv will be listed.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ARBWaveformCatalog(
    ViSession   instrSession,
    ViString    directoryPath,
    ViInt32 arraySize,
    ViChar  waveformCatalog[]
)
{
    ViChar  buffer [BUFFER_SIZE];

    if (!directoryPath)
        sprintf (buffer, ":SOUR:WAV:CAT?;*OPC?");
    else
        sprintf (buffer, ":SOUR:WAV:CAT? '%s';*OPC?", directoryPath);

    return rsafq_getCatalog (instrSession,
                             buffer,
                             arraySize,
                             waveformCatalog,
                             VI_ERROR_PARAMETER4);
}

/*****************************************************************************
 * Function: ARB Waveform Catalog Length
 * Purpose:  This function reads out the number of files in the default 
 *           directory for waveform files. Only files with the file extension 
 *           *.wv will be counted.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ARBWaveformCatalogLength(
    ViSession   instrSession,
    ViString    filePath,
    ViInt32*    waveformCatalogLength
)
{
    ViChar  buffer [BUFFER_SIZE];

    if (!filePath)
        sprintf (buffer, ":SOUR:WAV:CAT:LENG?");
    else
        sprintf (buffer, ":SOUR:WAV:CAT:LENG? '%s'", filePath);

    return rsafq_getViInt32 (instrSession,
                             buffer,
                             waveformCatalogLength,
                             VI_ERROR_PARAMETER3);

}

/*****************************************************************************
 * Function: ARB Waveform Data Write
 * Purpose:  This function writes the block data <binary block data> to the 
 *           file identified by <file_name>.
 *           
 *           Using this command, waveforms can be sent directly to the 
 *           instrument.
 *           
 *           Only the file name has to be entered, the file extension may be 
 *           omitted. Waveform data is stored only in files with the specific 
 *           file extensions *.wv.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ARBWaveformDataWrite(
    ViSession   instrSession,
    ViString    fileName,
    ViInt32 arraySize,
    ViString    waveformData
)
{
    ViUInt16    intfType     = VI_INTF_GPIB;
    ViChar*     pbuffer;
    ViChar*     p2buf;

    INIT;
    
    CHECKNULL (fileName, 3);
    CHECKNULL (waveformData, 5);

    CHECKERR (viGetAttribute (instrSession, VI_ATTR_INTF_TYPE, &intfType));

    if (intfType == VI_INTF_GPIB)
        CHECKERR (viWrite (instrSession, (ViBuf)":SYST:COMM:GPIB:LTER EOI\n", 25, VI_NULL));

    p2buf = pbuffer = (ViChar*) malloc (strlen (fileName) + arraySize + 64);

    p2buf += sprintf (pbuffer, ":FORM PACK;:SOUR:WAV:DATA '%s',#%ld%ld",
                               fileName,
                               (ViInt32)(floor(log10(arraySize)) + 1),
                               arraySize);

    memcpy (p2buf, waveformData, arraySize);
    p2buf[arraySize] = '\n';

    rsafq_status = viWrite (instrSession, (ViPBuf)pbuffer, (ViUInt32)((p2buf - pbuffer) + arraySize +1), NULL);

    if (pbuffer)
        free (pbuffer);

    if (intfType == VI_INTF_GPIB)
        viWrite (instrSession, (ViBuf)":SYST:COMM:GPIB:LTER STAN\n", 26, VI_NULL);

    return rsafq_status;
}

/*****************************************************************************
 * Function: ARB Waveform Delete
 * Purpose:  This function deletes the specified waveform file. If the file is 
 *           not on the default path, the path must be specified at the same 
 *           time.  The file extension may be omitted. Only files with the file 
 *           extension *.wv will be deleted.  
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ARBWaveformDelete(
    ViSession   instrSession,
    ViString    waveformName
)
{
    ViChar  *pbuffer;
    INIT;

    pbuffer = (ViChar*) malloc (strlen (waveformName) + 3);
    sprintf (pbuffer, "'%s'", waveformName);

    rsafq_status = rsafq_setViString (instrSession,
                                "SOUR:WAV:DEL",
                                pbuffer,
                                VI_ERROR_PARAMETER2);

    free (pbuffer);
    RETURN;
}

/*****************************************************************************
 * Function: ARB Waveform Points
 * Purpose:  This function queries the number of samples in the waveform file 
 *           selected using command :WAV:SEL. Only the file name has to be 
 *           entered. Only files with the file extension *.wv will be read out.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ARBWaveformPoints(
    ViSession   instrSession,
    ViInt32*    waveformPoints
)
{
    return rsafq_getViInt32 (instrSession,
                             "WAV:POIN?",
                             waveformPoints,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: ARB Waveform Select
 * Purpose:  This function selects the waveform file. If the file is not on 
 *           the default path, the path must be specified at the same time. If 
 *           no file of the specified name exists, it is created. The file 
 *           extension may be omitted. Only files with the file extension *.wv 
 *           will be created or loaded.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ARBWaveformSelect(
    ViSession   instrSession,
    ViString    waveformSelect
)
{
    ViChar* pbuffer;
    INIT;

    pbuffer = (ViChar*) malloc (strlen (waveformSelect) + 3);
    sprintf (pbuffer, "'%s'", waveformSelect);

    rsafq_status = rsafq_setViString (instrSession,
                                      "SOUR:WAV:SEL",
                                      pbuffer,
                                      VI_ERROR_PARAMETER2);
    free (pbuffer);
    RETURN;
}

/*****************************************************************************
 * Function: ARB Waveform Tag
 * Purpose:  This function queries the content of the specified tag of the 
 *           selected Waveform file (see following section Baseband Block - ARB 
 *           Waveform and List Format).
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ARBWaveformTag(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViString    waveformTag,
    ViChar  waveformTagData[]
)
{
    ViChar  buffer [BUFFER_SIZE];

    CHECKPARAM ( arraySize <= 0, 3);
    CHECKNULL ( waveformTag, 4);

    sprintf (buffer, ":FORM ASC;:SOUR:WAV:TAG? '%s'", waveformTag);

    return rsafq_getViString (instrSession,
                              buffer,
                              waveformTagData,
                              arraySize,
                              VI_ERROR_PARAMETER5);

}

/*****************************************************************************
 * Function: ARB Waveform Resample
 * Purpose:  This function selects the waveform file to be resampled and 
 *           starts resampling. The software resampler interpolates the 
 *           waveform to a clock rate of 600 MHz. Then the new calculated 
 *           waveform data is saved in a new file. The file is stored in the 
 *           same directory as the source file. For identification the 
 *           extension @600M is applied to the file name. After completing 
 *           recalculation the new waveform file is loaded and ready for 
 *           operation.
 *           
 *           Note(s):
 *           (1) This function is available on AFQ100B only.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ARBWaveformResample(
    ViSession   instrSession,
    ViString    waveformSelect
)
{
    ViChar* pbuffer;
	INIT;
	CHECKVER(AFQ100B);
	
    pbuffer = (ViChar*) malloc (strlen (waveformSelect) + 3);
    sprintf (pbuffer, "'%s'", waveformSelect);

    rsafq_status = rsafq_setViString (instrSession,
                                      "SOUR:WAV:RES",
                                      pbuffer,
                                      VI_ERROR_PARAMETER2);
    free (pbuffer);
    RETURN;
}
	   
/*****************************************************************************
 * Function: Get ARB Waveform Resample Frequency 
 * Purpose:  This function queries the resampling frequency set for the new 
 *           waveform.
 *           
 *           Note(s):
 *           (1) This function is available on AFQ100B only.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetARBWaveformResampleFrequency(
    ViSession   instrSession,
    ViReal64*   resampleFrequency
)
{
    INIT;
	CHECKVER(AFQ100B);
    return rsafq_getViReal64 (instrSession,
                          "SOUR:WAV:RES:FREQ?",
                          resampleFrequency,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set TSIG Sample Clock
 * Purpose:  This function enters the sample clock frequency for generating 
 *           either a sine, rectangular or a table sine test signal. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetTSIGSampleClock(
    ViSession   instrSession,
    ViReal64    sampleClock
)
{
    return rsafq_setViReal64 (instrSession,
                              "SOUR:TSIG:CLOC",
                              sampleClock,
                              1.0e3,600.0e6,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get TSIG Sample Clock
 * Purpose:  This function queries the sample clock frequency for generating 
 *           either a sine, rectangular or a table sine test signal. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetTSIGSampleClock(
    ViSession   instrSession,
    ViReal64*   sampleClock
)
{
    return rsafq_getViReal64 (instrSession,
                          "SOUR:TSIG:CLOC?",
                          sampleClock,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Configure TSIG Sine
 * Purpose:  This function configures sine test signal parameters. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ConfigureTSIGSine(
    ViSession   instrSession,
    ViReal64    sampleClock,
    ViInt32 samplesperPeriod,
    ViReal64    phaseOffsetQ
)								
{
	ViStatus rsafq_status = VI_SUCCESS;
    CHECKERR( rsafq_SetTSIGSampleClock(instrSession,sampleClock));

	rsafq_status = rsafq_SetTSIGSineSamplesPerPeriod(instrSession,
					samplesperPeriod);
	REMAP(2,3);
	
	rsafq_status = rsafq_SetTSIGSinePhaseOffsetQ(instrSession,phaseOffsetQ);
	REMAP(2,4);
    
    return rsafq_status;
}

/*****************************************************************************
 * Function: Get TSIG Sine Resulting Frequency
 * Purpose:  This function queries the resulting frequency of the test signal. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetTSIGSineResultingFrequency(
    ViSession   instrSession,
    ViReal64*   resultingFrequency
)
{
    return rsafq_getViReal64 (instrSession,
                          "SOUR:TSIG:SINE:FREQ?",
                          resultingFrequency,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Generate TSIG Sine
 * Purpose:  This function generates a signal and eventually stores the file 
 *           on the local hard disk (HD). 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GenerateTSIGSine(
    ViSession   instrSession,
    ViBoolean   storeToHDD,
    ViString    waveformFilename
)
{
	CHECKPARAM (rsafq_invalidViBooleanRange (storeToHDD), 2);

	if(storeToHDD==VI_TRUE){
		return rsafq_setViStringQuoted (instrSession,
									"SOUR:TSIG:SINE:CRE:NAM",
									"'",
                                    waveformFilename,VI_ERROR_PARAMETER3);
	}else{
		return rsafq_sendCmd (instrSession, "SOUR:TSIG:SINE:CRE");	
	}
}

/*****************************************************************************
 * Function: Set TSIG Sine Samples Per Period
 * Purpose:  This function sets the number of sample values required from the 
 *           sine wave per period. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetTSIGSineSamplesPerPeriod(
    ViSession   instrSession,
    ViInt32 samplesperPeriod
)
{
	return rsafq_setViInt32(instrSession,"SOUR:TSIG:SINE:SAMP",samplesperPeriod
							,4,1000,VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get TSIG Sine Samples Per Period
 * Purpose:  This function queries the number of sample values required from 
 *           the sine wave per period. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetTSIGSineSamplesPerPeriod(
    ViSession   instrSession,
    ViInt32*    samplesperPeriod
)
{
	return rsafq_getViInt32(instrSession,"SOUR:TSIG:SINE:SAMP?",samplesperPeriod
							,VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set TSIG Sine Phase Offset Q
 * Purpose:  This function sets the phase offset of the sinewave signal on the 
 *           Q channel relative to the sinewave signal on the I channel. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetTSIGSinePhaseOffsetQ(
    ViSession   instrSession,
    ViReal64    phaseOffsetQ
)
{
	return rsafq_setViReal64(instrSession,"SOUR:TSIG:SINE:PHAS",phaseOffsetQ
							,-180,180,VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get TSIG Sine Phase Offset Q
 * Purpose:  This function queries the phase offset of the sinewave signal on 
 *           the Q channel relative to the sinewave signal on the I channel. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetTSIGSinePhaseOffsetQ(
    ViSession   instrSession,
    ViReal64*   phaseOffsetQ
)
{
	return rsafq_getViReal64(instrSession,"SOUR:TSIG:SINE:PHAS?",phaseOffsetQ
							,VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Configure TSIG Rectangular
 * Purpose:  This function configures rectangular test signal parameters. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ConfigureTSIGRectangular(
    ViSession   instrSession,
    ViReal64    sampleClock,
    ViInt32 	samplesperPeriod,
    ViReal64    amplitude,
    ViReal64    offset
)
{
	ViStatus rsafq_status = VI_SUCCESS;
    CHECKERR( rsafq_SetTSIGSampleClock(instrSession,sampleClock));

	rsafq_status = rsafq_SetTSIGRectangularSamplesPerPeriod(instrSession,
					samplesperPeriod);
	REMAP(2,3);
	rsafq_status = 	rsafq_SetTSIGRectangularAmplitude(instrSession,
					amplitude);
	REMAP(2,4);
	rsafq_status = rsafq_SetTSIGRectangularOffset(instrSession,
					offset);
	REMAP(2,5);
    return rsafq_status;
}

/*****************************************************************************
 * Function: Generate TSIG Rectangular
 * Purpose:  This function generates a signal and eventually stores the file 
 *           on the local hard disk (HD). 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GenerateTSIGRectangular(
    ViSession   instrSession,
    ViBoolean   storeToHDD,
    ViString    waveformFilename
)
{
	CHECKPARAM (rsafq_invalidViBooleanRange (storeToHDD), 2);

	if(storeToHDD==VI_TRUE){
		return rsafq_setViStringQuoted (instrSession,
									"SOUR:TSIG:RECT:CRE:NAM",
									"'",
                                    waveformFilename,VI_ERROR_PARAMETER3);
	}else{
		return rsafq_sendCmd (instrSession, "SOUR:TSIG:RECT:CRE");	
	}
}

/*****************************************************************************
 * Function: Set TSIG Rectangular Samples Per Period
 * Purpose:  This function sets the number of sample values required from the 
 *           rectangular wave per period. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetTSIGRectangularSamplesPerPeriod(
    ViSession   instrSession,
    ViInt32 samplesperPeriod
)
{
	return rsafq_setViInt32(instrSession,"SOUR:TSIG:RECT:SAMP",samplesperPeriod
							,4,1000,VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get TSIG Rectangular Samples Per Period
 * Purpose:  This function queries the number of sample values required from 
 *           the rectangular wave per period. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetTSIGRectangularSamplesPerPeriod(
    ViSession   instrSession,
    ViInt32*    samplesperPeriod
)
{
	return rsafq_getViInt32(instrSession,"SOUR:TSIG:RECT:SAMP?",samplesperPeriod
							,VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set TSIG Rectangular Amplitude
 * Purpose:  This function sets the digital amplitude of the rectangular wave. 
 *           The abbreviation FS means full scale. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetTSIGRectangularAmplitude(
    ViSession   instrSession,
    ViReal64    amplitude
)
{
	return rsafq_setViReal64(instrSession,"SOUR:TSIG:RECT:AMPL",amplitude
							,0.0,1.0,VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get TSIG Rectangular Amplitude
 * Purpose:  This function queries the digital amplitude of the rectangular 
 *           wave. The abbreviation FS means full scale. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetTSIGRectangularAmplitude(
    ViSession   instrSession,
    ViReal64*   amplitude
)
{
	return rsafq_getViReal64(instrSession,"SOUR:TSIG:RECT:AMPL?",amplitude
							,VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set TSIG Rectangular Offset
 * Purpose:  This function sets a DC component to the rectangular test signal. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetTSIGRectangularOffset(
    ViSession   instrSession,
    ViReal64    offset
)
{
	return rsafq_setViReal64(instrSession,"SOUR:TSIG:RECT:OFFS",offset
							,-1.0,1.0,VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get TSIG Rectangular Offset
 * Purpose:  This function queries a DC component to the rectangular test 
 *           signal. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetTSIGRectangularOffset(
    ViSession   instrSession,
    ViReal64*   offset
)
{
	return rsafq_getViReal64(instrSession,"SOUR:TSIG:RECT:OFFS?",offset
							,VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Configure TSIG Table Sine Common
 * Purpose:  This function configures table sine test signal parameters. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ConfigureTSIGTableSineCommon(
    ViSession   instrSession,
    ViReal64    sampleClock,
    ViInt32 	tableLength,
    ViReal64    offset
)
{
	ViStatus rsafq_status = VI_SUCCESS;
    CHECKERR( rsafq_SetTSIGSampleClock(instrSession,sampleClock));

	rsafq_status = rsafq_SetTSIGTableSineLength(instrSession,
					tableLength);
	REMAP(2,3);
	rsafq_status = 	rsafq_SetTSIGTableSineOffset(instrSession,
					offset);
	REMAP(2,4);
    return rsafq_status;
}

/*****************************************************************************
 * Function: Configure TSIG Table Sine
 * Purpose:  This function configures selected sine wave.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ConfigureTSIGTableSine(
    ViSession   instrSession,
    ViInt32 sine,
    ViInt32 periods,
    ViReal64    amplitude,
    ViReal64    phase
)
{
	ViStatus rsafq_status = VI_SUCCESS;
    
	CHECKERR( rsafq_SetTSIGTableSinePeriods(instrSession,sine,periods));

	rsafq_status = rsafq_SetTSIGTableSineAmplitude(instrSession,sine,
					amplitude);
	REMAP(2,3);
	rsafq_status = rsafq_SetTSIGTableSinePhase(instrSession,sine,
					phase);
	REMAP(2,4);
    return rsafq_status;
}

/*****************************************************************************
 * Function: Get TSIG Table Sine Resulting Frequency
 * Purpose:  This function queries the resulting frequency of the selected 
 *           sine wave.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetTSIGTableSineResultingFrequency(
    ViSession   instrSession,
    ViInt32 sine,
    ViReal64*   resultingFrequency
)
{
	ViChar      buffer [BUFFER_SIZE];

	CHECKPARAM (rsafq_invalidViInt32Range (sine, 1, 2), 2);

    sprintf (buffer, "SOUR:TSIG:TSIN%d:FREQ?", sine);

    return rsafq_getViReal64 (instrSession,buffer,resultingFrequency,
                                      VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Generate TSIG Table Sine
 * Purpose:  This function generates a signal and eventually stores the file 
 *           on the local hard disk (HD). 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GenerateTSIGTableSine(
    ViSession   instrSession,
    ViBoolean   storeToHDD,
    ViString    waveformFilename
)
{
	CHECKPARAM (rsafq_invalidViBooleanRange (storeToHDD), 2);

	if(storeToHDD==VI_TRUE){
		return rsafq_setViStringQuoted (instrSession,
									"SOUR:TSIG:TSIN:CRE:NAM",
									"'",
                                    waveformFilename,VI_ERROR_PARAMETER3);
	}else{
		return rsafq_sendCmd (instrSession, "SOUR:TSIG:TSIN:CRE");	
	}
}

/*****************************************************************************
 * Function: Set TSIG Table Sine Length
 * Purpose:  This function sets the number of sample values of a waveform. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetTSIGTableSineLength(
    ViSession   instrSession,
    ViInt32 tableLength
)
{
	return rsafq_setViInt32(instrSession,"SOUR:TSIG:TSIN:TLENgth",tableLength
							,20,256000,VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get TSIG Table Sine Length
 * Purpose:  This function queries the number of sample values of a waveform. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetTSIGTableSineLength(
    ViSession   instrSession,
    ViInt32*    tableLength
)
{
	return rsafq_getViInt32(instrSession,"SOUR:TSIG:TSIN:TLENgth?",tableLength
							,VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set TSIG Table Sine Offset
 * Purpose:  This function sets a DC component to the test signal. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetTSIGTableSineOffset(
    ViSession   instrSession,
    ViReal64    offset
)
{
	return rsafq_setViReal64(instrSession,"SOUR:TSIG:TSIN:OFFS",offset
							,-1.0,1.0,VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get TSIG Table Sine Offset
 * Purpose:  This function queries a DC component of the test signal. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetTSIGTableSineOffset(
    ViSession   instrSession,
    ViReal64*   offset
)
{
	return rsafq_getViReal64(instrSession,"SOUR:TSIG:TSIN:OFFS?",offset
							,VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set TSIG Table Sine Periods
 * Purpose:  This function sets the number of periods for the sine waves.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetTSIGTableSinePeriods(
    ViSession   instrSession,
    ViInt32 sine,
    ViInt32 periods
)
{
	ViChar      buffer [BUFFER_SIZE];

	CHECKPARAM (rsafq_invalidViInt32Range (sine, 1, 2), 2);

    sprintf (buffer, "SOUR:TSIG:TSIN%d:PER", sine);

    return rsafq_setViInt32 (instrSession,buffer,periods,1,50,
                                      VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get TSIG Table Sine Periods
 * Purpose:  This function queries the number of periods for the sine waves.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetTSIGTableSinePeriods(
    ViSession   instrSession,
    ViInt32 sine,
    ViInt32*    periods
)
{
	ViChar      buffer [BUFFER_SIZE];

	CHECKPARAM (rsafq_invalidViInt32Range (sine, 1, 2), 2);

    sprintf (buffer, "SOUR:TSIG:TSIN%d:PER?", sine);

    return rsafq_getViInt32 (instrSession,buffer,periods,VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set TSIG Table Sine Amplitude 
 * Purpose:  This function sets the amplitude values of the sine waves.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetTSIGTableSineAmplitude(
    ViSession   instrSession,
    ViInt32 	sine,
    ViReal64    amplitude
)
{
	ViChar      buffer [BUFFER_SIZE];

	CHECKPARAM (rsafq_invalidViInt32Range (sine, 1, 2), 2);

    sprintf (buffer, "SOUR:TSIG:TSIN%d:AMPL", sine);

    return rsafq_setViReal64 (instrSession,buffer,amplitude,0.0,0.5,
                                      VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get TSIG Table Sine Amplitude
 * Purpose:  This function queries the amplitude values of the sine waves.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetTSIGTableSineAmplitude(
    ViSession   instrSession,
    ViInt32 sine,
    ViReal64*   amplitude
)
{
	ViChar      buffer [BUFFER_SIZE];

	CHECKPARAM (rsafq_invalidViInt32Range (sine, 1, 2), 2);

    sprintf (buffer, "SOUR:TSIG:TSIN%d:AMPL?", sine);

    return rsafq_getViReal64 (instrSession,buffer,amplitude,
									VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set TSIG Table Sine Phase
 * Purpose:  This function sets the phase shifts of the sine waves.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetTSIGTableSinePhase(
    ViSession   instrSession,
    ViInt32 	sine,
    ViReal64    phase
)
{
	ViChar      buffer [BUFFER_SIZE];

	CHECKPARAM (rsafq_invalidViInt32Range (sine, 1, 2), 2);

    sprintf (buffer, "SOUR:TSIG:TSIN%d:PHAS", sine);

    return rsafq_setViReal64 (instrSession,buffer,phase,-360,360,
                                      VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get TSIG Table Sine Phase
 * Purpose:  This function queries the phase shifts of the sine waves.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetTSIGTableSinePhase(
    ViSession   instrSession,
    ViInt32 	sine,
    ViReal64*   phase
)
{
	ViChar      buffer [BUFFER_SIZE];

	CHECKPARAM (rsafq_invalidViInt32Range (sine, 1, 2), 2);

    sprintf (buffer, "SOUR:TSIG:TSIN%d:PHAS?", sine);

    return rsafq_getViReal64 (instrSession,buffer,phase,VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Configure TSIG Pattern
 * Purpose:  This function configures the pattern test signal.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ConfigureTSIGPattern(
    ViSession   instrSession,
    ViInt32 i,
    ViInt32 q
)
{
	ViStatus rsafq_status = VI_SUCCESS;
    
	CHECKERR( rsafq_SetTSIGPatternI(instrSession,i));

	rsafq_status = rsafq_SetTSIGPatternQ(instrSession,q);
	REMAP(2,3);
    return rsafq_status;
}

/*****************************************************************************
 * Function: Generate TSIG Pattern
 * Purpose:  This function generates a signal and eventually stores the file 
 *           on the local hard disk (HD). 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GenerateTSIGPattern(
    ViSession   instrSession,
    ViBoolean   storeToHDD,
    ViString    waveformFilename
)
{
	CHECKPARAM (rsafq_invalidViBooleanRange (storeToHDD), 2);

	if(storeToHDD==VI_TRUE){
		return rsafq_setViStringQuoted (instrSession,
									"SOUR:TSIG:PATT:CRE:NAM",
									"'",
                                    waveformFilename,VI_ERROR_PARAMETER3);
	}else{
		return rsafq_sendCmd (instrSession, "SOUR:TSIG:PATT:CRE");	
	}
}

/*****************************************************************************
 * Function: Set TSIG Pattern I
 * Purpose:  This function sets the pattern number of the I component. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetTSIGPatternI(
    ViSession   instrSession,
    ViInt32 i
)
{
	return rsafq_setViInt32(instrSession,"SOUR:TSIG:PATT:I",i
							,0,65535,VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get TSIG Pattern I
 * Purpose:  This function queries the pattern number of the I component. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetTSIGPatternI(
    ViSession   instrSession,
    ViInt32*    i
)
{
	return rsafq_getViInt32(instrSession,"SOUR:TSIG:PATT:I?",i
							,VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set TSIG Pattern Q
 * Purpose:  This function sets the pattern number of the Q component. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetTSIGPatternQ(
    ViSession   instrSession,
    ViInt32 q
)
{
	return rsafq_setViInt32(instrSession,"SOUR:TSIG:PATT:Q",q
							,0,65535,VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get TSIG Pattern Q
 * Purpose:  This function queries the pattern number of the Q component. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetTSIGPatternQ(
    ViSession   instrSession,
    ViInt32*    q
)
{
	return rsafq_getViInt32(instrSession,"SOUR:TSIG:PATT:Q?",q
							,VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Trigger Status
 * Purpose:  This function queries the status of waveform output for all 
 *           trigger modes.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetTriggerStatus(
    ViSession   instrSession,
    ViInt32*    triggerStatus
)
{
    return rsafq_getEnum (instrSession, 
                             "TRIG:RMOD?",
                             triggerStatus,
                             triggerStatusArr,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Configure Data Input
 * Purpose:  This function configures the data inputs. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ConfigureDataInput(
    ViSession   instrSession,
    ViInt32 impedanceData,
    ViReal64    thresholdData
)
{
    ViStatus    rsafq_status;

    CHECKERR (rsafq_SetImpedanceDataInput (instrSession, impedanceData));

    rsafq_status = rsafq_SetThresholdDataInput (instrSession, thresholdData);
    REMAP (2, 3);
    
    RETURN;
}


/*****************************************************************************
 * Function: Configure Trigger
 * Purpose:  This function sets all parameters of trigger.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ConfigureTrigger(
    ViSession   instrSession,
    ViInt32 triggerMode,
    ViBoolean   retrigger,
    ViInt32 triggerSource,
    ViReal64    externalDelay,
    ViReal64    externalInhibit,
    ViInt32 inputTriggerSlope
)
{
    ViStatus    rsafq_status;

    CHECKERR (rsafq_SetTriggerMode (instrSession, triggerMode));

    rsafq_status = rsafq_SetRetriggerEnable (instrSession, retrigger);
    REMAP (2, 3);
    
    rsafq_status = rsafq_SetTriggerSource (instrSession, triggerSource);
    REMAP (2, 4);
    
    rsafq_status = rsafq_SetExternalDelay (instrSession, externalDelay);
    REMAP (2, 5);
    
    rsafq_status = rsafq_SetExternalInhibit (instrSession, externalInhibit);
    REMAP (2, 6);
    
    rsafq_status = rsafq_SetInstrumentTriggerInputSlope (instrSession, inputTriggerSlope);
    REMAP (2, 7);
    
    RETURN;
}

/*****************************************************************************
 * Function: Configure Segment Trigger
 * Purpose:  This function sets all parameters of segment trigger.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ConfigureSegmentTrigger(
    ViSession   instrSession,
    ViInt32 multiSegmentTriggerMode,
    ViInt32 segmentTriggerSource,
    ViReal64    segmentExternalDelay,
    ViReal64    segmentExternalInhibit,
    ViInt32 inputTriggerSlope
)
{
    ViStatus    rsafq_status;

    rsafq_status = rsafq_SetSegmentTriggerSource (instrSession, segmentTriggerSource);
    REMAP (2, 3);
    
    CHECKERR (rsafq_SetMultiSegmentTriggerMode (instrSession, multiSegmentTriggerMode));

    rsafq_status = rsafq_SetSegmentExternalDelay (instrSession, segmentExternalDelay);
    REMAP (2, 4);
    
    rsafq_status = rsafq_SetSegmentExternalInhibit (instrSession, segmentExternalInhibit);
    REMAP (2, 5);
    
    rsafq_status = rsafq_SetSegmentTriggerInputSlope (instrSession, inputTriggerSlope);
    REMAP (2, 6);
    
    RETURN;
}

/*****************************************************************************
 * Function: Trigger Execute
 * Purpose:  This function executes a trigger event. However, this is possible 
 *           only if the Bus trigger source is selected.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_TriggerExecute(
    ViSession   instrSession
)
{
    return rsafq_sendCmd (instrSession, "TRIG:EXEC");
}

/*****************************************************************************
 * Function: Set Trigger Mode
 * Purpose:  This function selects selects the trigger mode. i.e. continuous 
 *           or single. After a trigger event, the waveform will be output 
 *           continuously or once.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetTriggerMode(
    ViSession   instrSession,
    ViInt32 triggerMode
)
{
    return rsafq_setEnum (instrSession,
                          "TRIG:MODE",
                          triggerMode,
                          triggerModeArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Trigger Mode
 * Purpose:  This function returns the trigger mode.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetTriggerMode(
    ViSession   instrSession,
    ViInt32*    triggerMode
)
{
    return rsafq_getEnum (instrSession,
                          "TRIG:MODE?",
                          triggerMode,
                          triggerModeArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Trigger Source
 * Purpose:  This function selects the trigger source. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetTriggerSource(
    ViSession   instrSession,
    ViInt32 triggerSource
)
{
    return rsafq_setEnum (instrSession,
                          ":TRIG:SOUR",
                          triggerSource,
                          triggerSourceArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Trigger Source
 * Purpose:  This function returns the trigger source. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetTriggerSource(
    ViSession   instrSession,
    ViInt32*    triggerSource
)
{
    return rsafq_getEnum (instrSession,
                          ":TRIG:SOUR?",
                          triggerSource,
                          triggerSourceArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Trigger Count
 * Purpose:  This function sets a repeat rate for restarting the waveform 
 *           after a trigger event occurs.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetTriggerCount(
    ViSession   instrSession,
    ViInt32 triggerCount
)
{
	return rsafq_setViInt32(instrSession,"SOUR:TRIG:RCO",triggerCount
							,1,100,VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Trigger Count
 * Purpose:  This function queries a repeat rate for restarting the waveform 
 *           after a trigger event occurs.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetTriggerCount(
    ViSession   instrSession,
    ViInt32*    triggerCount
)
{
	return rsafq_getViInt32(instrSession,"SOUR:TRIG:RCO?",triggerCount
							,VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set External Delay
 * Purpose:  This function specifies the trigger delay for external 
 *           triggering. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetExternalDelay(
    ViSession   instrSession,
    ViReal64    externalDelay
)
{
    return rsafq_setViReal64 (instrSession,
                              ":TRIG:DEL",
                              externalDelay,
                              0.0,10.0,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get External Delay
 * Purpose:  This function returns the trigger delay for external triggering. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetExternalDelay(
    ViSession   instrSession,
    ViReal64*   externalDelay
)
{
    return rsafq_getViReal64 (instrSession,
                              ":TRIG:DEL?",
                              externalDelay,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set External Inhibit
 * Purpose:  This function specifies the time interval by which a restart is 
 *           to be inhibited following a trigger event. This function applies 
 *           only in the case of external triggering.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetExternalInhibit(
    ViSession   instrSession,
    ViReal64    externalInhibit
)
{
    return rsafq_setViReal64 (instrSession,
                              "TRIG:INH",
                              externalInhibit,
                              0.0,10.0,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get External Inhibit
 * Purpose:  This function returns the time interval by which a restart is to 
 *           be inhibited following a trigger event. This function applies only 
 *           in the case of external triggering.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetExternalInhibit(
    ViSession   instrSession,
    ViReal64*   externalInhibit
)
{
    return rsafq_getViReal64 (instrSession,
                              ":TRIG:INH?",
                              externalInhibit,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Impedance Trigger Data Input
 * Purpose:  This function sets the impedance for the three inputs CLOCK IN, 
 *           TRIG, and NEXT (segment trigger). The lower value is recommended 
 *           for high clock frequencies. The meanings of the abbreviations G50 
 *           and G1k are 0.05 and 1.00 k, respectively.
 *****************************************************************************/
 ViStatus _VI_FUNC rsafq_SetImpedanceDataInput(
    ViSession   instrSession,
    ViInt32 impedanceData
)
{
    return rsafq_setEnum (instrSession,
                          ":INP:CONT:IMP",
                          impedanceData,
                          impedanceArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Impedance Trigger Data Input
 * Purpose:  This function returns the impedance for the three inputs CLOCK IN, 
 *           TRIG, and NEXT (segment trigger). The lower value is recommended 
 *           for high clock frequencies. The meanings of the abbreviations G50 
 *           and G1k are 0.05 and 1.00 k, respectively.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetImpedanceDataInput (ViSession instrSession,
                                               ViInt32 *dataInputImpedance)
{
    return rsafq_getEnum (instrSession,
                          ":INP:CONT:IMP?",
                          dataInputImpedance,
                          impedanceArr,
                          VI_ERROR_PARAMETER2);    
}

/*****************************************************************************
 * Function: Set Threshold Trigger Data Input
 * Purpose:  This function sets the high/low threshold for the three inputs
 *           CLOCK IN, TRIG, and NEXT (segment trigger).           
 *****************************************************************************/
 ViStatus _VI_FUNC rsafq_SetThresholdDataInput(
    ViSession   instrSession,
    ViReal64 thresholdData
)
{
    return rsafq_setViReal64 (instrSession,
                              ":INP:CONT:THR",
                              thresholdData,
                              0.0,2.0,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Threshold Trigger Data Input
 * Purpose:  This function returns the high/low threshold for the three inputs
 *           CLOCK IN, TRIG, and NEXT (segment trigger).           
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetThresholdDataInput (ViSession instrSession,
                                               ViReal64 *dataInputThreshold)
{
    return rsafq_getViReal64 (instrSession,
                              ":INP:CONT:THR?",
                              dataInputThreshold,
                              VI_ERROR_PARAMETER2);    
}

/*****************************************************************************
 * Function: Set Instrument Trigger Input Slope
 * Purpose:  This function selects the slope of the active edge of the 
 *           external trigger signal. The trigger signal is fed in via the TRIG 
 *           connector on the rear panel.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetInstrumentTriggerInputSlope(
    ViSession   instrSession,
    ViInt32 inputTriggerSlope
)
{
    return rsafq_setEnum (instrSession,
                          "INP:TRIG:SLOP",
                          inputTriggerSlope,
                          inputSlopeArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Instrument Trigger Input Slope
 * Purpose:  This function returns the slope of the active edge of the 
 *           external trigger signal. The trigger signal is fed in via the TRIG 
 *           connector on the rear panel.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetInstrumentTriggerInputSlope(
    ViSession   instrSession,
    ViInt32*    inputTriggerSlope
)
{
    return rsafq_getEnum (instrSession,
                          "INP:TRIG:SLOP?",
                          inputTriggerSlope,
                          inputSlopeArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Retrigger Enable
 * Purpose:  This function determines whether a following trigger event shall 
 *           restart the current signal output. This setting is always ON for 
 *           multi segment waveforms.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetRetriggerEnable(
    ViSession   instrSession,
    ViBoolean   retrigger
)
{
    return rsafq_setViBoolean (instrSession,
                               "TRIG:REN",
                               retrigger,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Retrigger Enable
 * Purpose:  This function returns the retrigger enable state.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetRetriggerEnable(
    ViSession   instrSession,
    ViBoolean*  rettriger
)
{
    return rsafq_getViBoolean (instrSession,
                               "TRIG:REN?",
                               rettriger,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Segment Trigger Execute
 * Purpose:  This function executes a manual segment trigger. A manual segment 
 *           trigger can be executed only when a manual trigger source has been 
 *           selected. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SegmentTriggerExecute(
    ViSession   instrSession
)
{
    return rsafq_sendCmd (instrSession, "TRIG:SEGM:EXEC");
}

/*****************************************************************************
 * Function: Set Multi Segment Trigger Mode
 * Purpose:  This function selects the extended trigger mode for multi-segment 
 *           waveforms.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetMultiSegmentTriggerMode(
    ViSession   instrSession,
    ViInt32 multiSegmentTriggerMode
)
{
    if (multiSegmentTriggerMode == RSAFQ_MSEGMENT_TRIGGER_MODE_AUTO)
        return RSAFQ_ERROR_INSTRUMENT_NSUP_FUNC;
    
    return rsafq_setEnum (instrSession,
                          "TRIG:SEGM:MODE",
                          multiSegmentTriggerMode,
                          multiSegmentTriggerModeArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Multi Segment Trigger Mode
 * Purpose:  This function returns the extended trigger mode for multi-segment 
 *           waveforms. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetMultiSegmentTriggerMode(
    ViSession   instrSession,
    ViInt32*    multiSegmentTriggerMode
)
{
    return rsafq_getEnum (instrSession,
                          "TRIG:SEGM:MODE?",
                          multiSegmentTriggerMode,
                          multiSegmentTriggerModeArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Segment Trigger Source
 * Purpose:  This function selects the segment trigger source.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetSegmentTriggerSource(
    ViSession   instrSession,
    ViInt32 segmentTriggerSource
)
{
    return rsafq_setEnum (instrSession,
                          "TRIG:SEGM:SOUR",
                          segmentTriggerSource,
                          segmentTriggerSourceArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Segment Trigger Source
 * Purpose:  This function returns the segment trigger source. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetSegmentTriggerSource(
    ViSession   instrSession,
    ViInt32*    segmentTriggerSource
)
{
    return rsafq_getEnum (instrSession,
                          "TRIG:SEGM:SOUR?",
                          segmentTriggerSource,
                          segmentTriggerSourceArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Segment Trigger Count 
 * Purpose:  This function sets the repeat rate for each particular segment of 
 *           a multi segment waveform. The repeat count is entered in a list of 
 *           numbers. Each segment can be repeated up to 100 times.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetSegmentTriggerCount(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViInt32 triggerCount[]
)
{
    ViChar* pbuffer;
    ViChar* p2buf;
    ViInt32 i;

    CHECKNULL (triggerCount, 3);
    
    p2buf = pbuffer = malloc(arraySize*20 + 30);
    p2buf += sprintf(pbuffer, ":SOUR:TRIG:SEGM:RCO ");

    for (i = 0; i < arraySize; i++)
        p2buf += sprintf (p2buf, "%ld,", triggerCount[i]);

    *--p2buf = '\0';

    return rsafq_sendCmd (instrSession, pbuffer);
}

/*****************************************************************************
 * Function: Get Segment Trigger Count 
 * Purpose:  This function queries the repeat rate for each particular segment 
 *           of a multi segment waveform.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetSegmentTriggerCount(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViInt32 triggerCount[]
)
{
    return rsafq_getViInt32Array (instrSession,
                                   ":SOUR:TRIG:SEGM:RCO?",
                                   triggerCount,
                                   arraySize,
                                   VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Segment External Delay
 * Purpose:  This function specifies a segment trigger delay for a multi 
 *           segment waveform.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetSegmentExternalDelay(
    ViSession   instrSession,
    ViReal64    segmentExternalDelay
)
{
    return rsafq_setViReal64 (instrSession,
                              "TRIG:SEGM:DEL",
                              segmentExternalDelay,
                              0,10,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Segment External Delay
 * Purpose:  This function returns a segment trigger delay for a multi segment 
 *           waveform.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetSegmentExternalDelay(
    ViSession   instrSession,
    ViReal64*   segmentExternalDelay
)
{
    return rsafq_getViReal64 (instrSession,
                              "TRIG:SEGM:DEL?",
                              segmentExternalDelay,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Segment External Inhibit
 * Purpose:  This function sets the duration for inhibiting a new trigger 
 *           event subsequent to triggering.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetSegmentExternalInhibit(
    ViSession   instrSession,
    ViReal64    segmentExternalInhibit
)
{
    return rsafq_setViReal64 (instrSession,
                              "TRIG:SEGM:INH",
                              segmentExternalInhibit,
                              0,10,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Segment External Inhibit
 * Purpose:  This function returns the duration for inhibiting a new trigger 
 *           event subsequent to triggering. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetSegmentExternalInhibit(
    ViSession   instrSession,
    ViReal64*   segmentExternalInhibit
)
{
    return rsafq_getViReal64 (instrSession,
                              "TRIG:SEGM:INH?",
                              segmentExternalInhibit,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Segment Trigger Input Slope
 * Purpose:  This function selects slope of the active edge of the external 
 *           trigger signal. The external segment trigger signal is fed in via 
 *           the NEXT connector on the rear panel.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetSegmentTriggerInputSlope(
    ViSession   instrSession,
    ViInt32 inputTriggerSlope
)
{
    return rsafq_setEnum (instrSession,
                          "INP:TRIG:SEGM:SLOP",
                          inputTriggerSlope,
                          inputSlopeArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Segment Trigger Input Slope
 * Purpose:  This function returns the segment trigger slope.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetSegmentTriggerInputSlope(
    ViSession   instrSession,
    ViInt32*    inputTriggerSlope
)
{
    return rsafq_getEnum (instrSession,
                          "INP:TRIG:SEGM:SLOP?",
                          inputTriggerSlope,
                          inputSlopeArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Waveform Segments
 * Purpose:  This function indicates which segment the instrument is 
 *           generating.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetWaveformSegments(
    ViSession   instrSession,
    ViInt32*    segment
)
{
    return rsafq_getViInt32 (instrSession,
                             "WSEG?",
                             segment,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Next Waveform Segment
 * Purpose:  This function determines which segment the instrument shall 
 *           generate after a following trigger event.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetNextWaveformSegment(
    ViSession   instrSession,
    ViInt32 nextSegment
)
{
    return rsafq_setViInt32 (instrSession,
                             "WSEG:NEXT",
                             nextSegment,
                             0,0,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Next Waveform Segment
 * Purpose:  This function returns a next waveform segment which the 
 *           instrument shall generate after a following trigger event.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetNextWaveformSegment(
    ViSession   instrSession,
    ViInt32*    nextSegment
)
{
    return rsafq_getViInt32 (instrSession,
                             "WSEG:NEXT?",
                             nextSegment,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Next Waveform Segment Mode
 * Purpose:  This function sets the segment sequence mode of a multisegment 
 *           waveform.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetNextWaveformSegmentMode (ViSession instrSession,
                                                    ViInt32 nextWaveformSegmentMode)
{
    return rsafq_setEnum (instrSession,
                          "SOUR:WSEG:NEXT:MODE",
                          nextWaveformSegmentMode,
                          nextWaveformSegmentModeArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Next Waveform Segment Mode
 * Purpose:  This function returns the segment sequence mode of a multisegment 
 *           waveform.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetNextWaveformSegmentMode (ViSession instrSession,
                                                    ViInt32 *nextWaveformSegmentMode)
{
    return rsafq_getEnum (instrSession,
                          "SOUR:WSEG:NEXT:MODE?",
                          nextWaveformSegmentMode,
                          nextWaveformSegmentModeArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Configure Marker Mode
 * Purpose:  This function sets all parameters for selected marker output.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ConfigureMarkerMode(
    ViSession   instrSession,
    ViInt32 markerOutput,
    ViInt32 markerType,
    ViInt32 markerPulseDivider,
    ViInt32 patternCount,
    ViBoolean    markerPattern[],
    ViInt32 markerONTime,
    ViInt32 markerOFFTime
)
{
    ViStatus rsafq_status;
    
    
    CHECKERR( rsafq_SetMarkerMode (instrSession, markerOutput, markerType));
    
    switch (markerType)
    {
        case RSAFQ_MARKER_TYPE_UNCHANGED:
        case RSAFQ_MARKER_TYPE_RESTART:
        break;
        
        case RSAFQ_MARKER_TYPE_PULSE:
            rsafq_status = rsafq_SetMarkerPulseDivider (instrSession, markerOutput, markerPulseDivider);
            REMAP(3,4);
        break;
        
        case RSAFQ_MARKER_TYPE_PATTERN:
            rsafq_status = rsafq_SetMarkerPattern (instrSession, markerOutput, patternCount, markerPattern);
            REMAP(3,5);
        break;
        
        case RSAFQ_MARKER_TYPE_RATIO:
            rsafq_status = rsafq_SetMarkerONTime (instrSession, markerOutput, markerONTime);
            REMAP(3,6);

            rsafq_status = rsafq_SetMarkerOFFTime (instrSession, markerOutput, markerOFFTime);
            REMAP(3,7);
        break;
		case RSAFQ_MARKER_TYPE_LIST:
			// The pattern is entered with the aid of the function SetMarkerList.
		break;
        default: 
            return VI_ERROR_PARAMETER3;
    }
    
    return rsafq_status;    

}

/*****************************************************************************
 * Function: Set Marker Mode
 * Purpose:  This function defines the signal for the selected marker output. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetMarkerMode(
    ViSession   instrSession,
    ViInt32 markerOutput,
    ViInt32 markerType
)
{
    ViChar  buffer[BUFFER_SIZE];
    
    CHECKPARAM (rsafq_invalidViInt32Range (markerOutput,1,4), 2);
    
    sprintf (buffer,":SOUR:TRIG:OUTP%ld:MODE", markerOutput);
    
    return rsafq_setEnum (instrSession,
                          buffer,
                          markerType,
                          markerTypeArr,
                          VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Marker Mode
 * Purpose:  This function returns the signal for the selected marker output.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetMarkerMode(
    ViSession   instrSession,
    ViInt32 markerOutput,
    ViInt32*    markerType
)
{
    ViChar  command[256];
    CHECKPARAM (rsafq_invalidViInt32Range (markerOutput,1,4), 2);
    
    sprintf (command,"SOUR:TRIG:OUTP%ld:MODE?", markerOutput);
    
    return rsafq_getEnum (instrSession,
                          command,
                          markerType,
                          markerTypeArr,
                          VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Marker Pattern
 * Purpose:  This function defines the bit pattern used to generate the marker 
 *           signal in the setting SOURce:TRIGger:OUTPut:MODE PATTern 0 is 
 *           marker off, 1 is marker on.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetMarkerPattern (ViSession instrSession,
                                          ViInt32 markerOutput,
                                          ViInt32 patternCount,
                                          ViBoolean markerPattern[])
{
    ViChar  command[256];
    ViChar  *p2buf;
    ViInt32 i;
    
    CHECKPARAM (rsafq_invalidViInt32Range (markerOutput, 1, 4), 2);
    CHECKPARAM (rsafq_invalidViInt32Range (patternCount, 1, 32), 3);

    p2buf = command;
    p2buf += sprintf(command, ":SOUR:TRIG:OUTP%ld:PATT ", markerOutput);
    
    for (i = 0; i < patternCount; i++)
        p2buf += sprintf (p2buf, "%d,", markerPattern[i]);

    *--p2buf = '\0';

    return rsafq_sendCmd (instrSession, command);
}

/*****************************************************************************
 * Function: Get Marker Pattern
 * Purpose:  This function returns the bit pattern used to generate the marker 
 *           signal in the setting  SOURce:TRIGger:OUTPut:MODE PATTern 0 is 
 *           marker off, 1 is marker on.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetMarkerPattern(
    ViSession   instrSession,
    ViInt32 markerOutput,
    ViInt32 patternCount,
    ViBoolean  markerPattern[]
)
{
    ViStatus    rsafq_status = VI_SUCCESS;
    ViChar*     pbuffer = VI_NULL;
    ViChar*     p2buf;
    ViUInt32    local_count;
    ViInt32     i = 0;

    if (markerPattern == VI_NULL)
        return VI_ERROR_PARAMETER4;

    CHECKERR (viPrintf(instrSession, ":SOUR:TRIG:OUTP%ld:PATT?", markerOutput));
    CHECKERR (rsafq_readDataUnknownLength(instrSession, &pbuffer, &local_count));
    p2buf = pbuffer;

    p2buf = strtok(pbuffer, ",");

    while(p2buf && (i < patternCount))
        {
        markerPattern[i++] = (ViBoolean)atoi(p2buf);
        p2buf = strtok(NULL, ",");
        };

    free(pbuffer);
    
    return rsafq_status;
}

/*****************************************************************************
 * Function: Set Marker OFF Time
 * Purpose:  This function sets the number of symbols in a period (ON time + 
 *           OFF time) during which the marker signal in setting 
 *           TRIGger:OUTPut:MODE RATio on the marker outputs is OFF.  
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetMarkerOFFTime(
    ViSession   instrSession,
    ViInt32 markerOutput,
    ViInt32 markerOFFTime
)
{
    ViChar  buffer[BUFFER_SIZE];
    
    CHECKPARAM (rsafq_invalidViInt32Range (markerOutput,1,4), 2);

    sprintf (buffer,":SOUR:TRIG:OUTP%ld:OFFT", markerOutput);
    
    return rsafq_setViInt32 (instrSession,
                             buffer,
                             markerOFFTime,
                             1,16777215,             // 2^24-1
                             VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Marker OFF Time
 * Purpose:  This function returnsthe number of symbols in a period (ON time + 
 *           OFF time) during which the marker signal in setting 
 *           TRIGger:OUTPut:MODE RATio  on the marker outputs is OFF.  
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetMarkerOFFTime(
    ViSession   instrSession,
    ViInt32 markerOutput,
    ViInt32*    markerOFFTime
)
{
    ViChar  command[256];
    
    CHECKPARAM (rsafq_invalidViInt32Range (markerOutput, 1, 4), 2);
    
    sprintf (command,":SOUR:TRIG:OUTP%ld:OFFT?", markerOutput);
    
    return rsafq_getViInt32 (instrSession,
                             command,
                             markerOFFTime,
                             VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Marker ON Time
 * Purpose:  This function sets the number of symbols in a period (ON time + 
 *           OFF time) during which the marker signal in setting 
 *           TRIGger:OUTPut:MODE RATio on the marker outputs is ON.  
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetMarkerONTime(
    ViSession   instrSession,
    ViInt32 markerOutput,
    ViInt32 markerONTime
)
{
    ViChar  buffer[BUFFER_SIZE];
    
    CHECKPARAM (rsafq_invalidViInt32Range (markerOutput,1,4), 2);
    
    sprintf (buffer,":SOUR:TRIG:OUTP%ld:ONT", markerOutput);
    
    return rsafq_setViInt32 (instrSession,
                             buffer,
                             markerONTime,
                             1,16777215,             // 2^24-1
                             VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Marker ON Time
 * Purpose:  This function returns the number of symbols in a period (ON time 
 *           + OFF time) during which the marker signal in setting 
 *           TRIGger:OUTPut:MODE RATio on the marker outputs is ON.  
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetMarkerONTime(
    ViSession   instrSession,
    ViInt32 markerOutput,
    ViInt32*    markerONTime
)
{
    ViChar  command[256];
    
    CHECKPARAM (rsafq_invalidViInt32Range (markerOutput,1,4), 2);
    
    sprintf (command,":SOUR:TRIG:OUTP%ld:ONT?", markerOutput);
    
    return rsafq_getViInt32 (instrSession,
                             command,
                             markerONTime,
                             VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Marker Pulse Divider
 * Purpose:  This function sets the divider for the pulsed marker signal in 
 *           the setting SOURce:TRIGger:OUTPut:MODE PULSe. The pulse frequency 
 *           is derived by  dividing the symbol rate by the divider. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetMarkerPulseDivider(
    ViSession   instrSession,
    ViInt32 markerOutput,
    ViInt32 markerPulseDivider
)
{
    ViChar  buffer[BUFFER_SIZE];
    
    CHECKPARAM (rsafq_invalidViInt32Range (markerOutput,1,4), 2);

    sprintf (buffer,":SOUR:TRIG:OUTP%ld:PULS:DIV", markerOutput);
    
    return rsafq_setViInt32 (instrSession,
                             buffer,
                             markerPulseDivider,
                             2, 1024,           /* to 2^10 */
                             VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Marker Pulse Divider
 * Purpose:  This function returns the divider for the pulsed marker signal in 
 *           the setting SOURce:TRIGger:OUTPut:MODE PULSe. The pulse frequency 
 *           is derived by  dividing the symbol rate by the divider. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetMarkerPulseDivider(
    ViSession   instrSession,
    ViInt32 markerOutput,
    ViInt32*    markerPulseDivider
)
{
    ViChar    command[256];
    
    CHECKPARAM (rsafq_invalidViInt32Range (markerOutput, 1, 4), 2);
    
    sprintf(command, "TRIG:OUTP%ld:PULS:DIV? ", markerOutput);

    return rsafq_getViInt32 (instrSession,
                             command,
                             markerPulseDivider,
                             VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Marker Pulse Frequency
 * Purpose:  This function queries the pulse frequency of the pulsed marker 
 *           signal in the setting SOURce:TRIGger:OUTPut:MODE PULSe. The pulse 
 *           frequency  is derived by dividing the symbol rate by the divider. 
 *           The divider is defined with command :TRIG:OUTP:PULS:DIV. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetMarkerPulseFrequency(
    ViSession   instrSession,
    ViInt32 markerOutput,
    ViReal64*   markerPulseFrequency
)
{
    ViChar  command[256];
    
    CHECKPARAM (rsafq_invalidViInt32Range (markerOutput,1,4), 2);
    
    sprintf (command,":SOUR:TRIG:OUTP%ld:PULS:FREQ?", markerOutput);
    
    return rsafq_getViReal64 (instrSession,
                              command,
                              markerPulseFrequency,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Marker List
 * Purpose:  This function defines the pattern of samples used to generate the 
 *           marker signal in the setting TRIGger:OUTPut:MODE LIST for the 
 *           selected marker output.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetMarkerList(
    ViSession   instrSession,
    ViInt32 	markerOutput,
    ViString    markerList
)
{
	ViChar      buffer [BUFFER_SIZE];

	CHECKPARAM (rsafq_invalidViInt32Range (markerOutput, 1, 4), 2);
	CHECKPARAM (rsafq_invalidViString(markerList,BUFFER_SIZE),3);

    sprintf (buffer, "SOUR:TRIG:OUTP%d:LIST '%s'", markerOutput, markerList);

    return rsafq_sendCmd(instrSession,buffer);
}

/*****************************************************************************
 * Function: Get Marker List 
 * Purpose:  This function queries the pattern of samples used to generate the 
 *           marker signal in the setting TRIGger:OUTPut:MODE LIST for the 
 *           selected marker output.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetMarkerList (ViSession instrSession,
                                       ViInt32 markerOutput, ViChar markerList[],
                                       ViInt32 arraySize)
{
    ViChar  buffer [BUFFER_SIZE];

	CHECKPARAM (rsafq_invalidViInt32Range (markerOutput, 1, 4), 2);
	CHECKNULL ( markerList, 3);

    sprintf (buffer, "SOUR:TRIG:OUTP%d:LIST?", markerOutput);

    return rsafq_getViString (instrSession,
                              buffer,
                              markerList,
                              arraySize,
                              VI_ERROR_PARAMETER3);
    
}

/*****************************************************************************
 * Function: Set Marker Delay
 * Purpose:  This function defines the delay between the signal on the marker 
 *           outputs and the start of the signals, expressed in terms of 
 *           samples. Command :TRIGger:OUTPut:DELay:FIXed ON can be used to 
 *           restrict the range of values to the dynamic range, i.e. the range 
 *           within which a delay of the marker signals can be set without 
 *           restarting the marker and signal.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetMarkerDelay(
    ViSession   instrSession,
    ViInt32 	markerOutput,
    ViInt32 	markerDelay
)
{
    ViChar  buffer[BUFFER_SIZE];
    
    CHECKPARAM (rsafq_invalidViInt32Range (markerOutput,1,4), 2);

    sprintf (buffer,":SOUR:TRIG:OUTP%ld:DEL", markerOutput);
    
    return rsafq_setViInt32 (instrSession,
                             buffer,
                             markerDelay,
                             1,1048575,
                             VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Marker Delay
 * Purpose:  This function returns the delay between the signal on the marker 
 *           outputs and the start of the signals, expressed in terms of 
 *           samples. Command :TRIGger:OUTPut:DELay:FIXed ON can be used to 
 *           restrict the range of values to the dynamic range, i.e. the range 
 *           within which a delay of the marker signals can be set without 
 *           restarting the marker and signal.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetMarkerDelay(
    ViSession   instrSession,
    ViInt32 	markerOutput,
    ViInt32*    markerDelay
)
{
    ViChar  command[256];
    
    CHECKPARAM (rsafq_invalidViInt32Range (markerOutput,1,4), 2);
    
    sprintf (command,":SOUR:TRIG:OUTP%ld:DEL?", markerOutput);
    
    return rsafq_getViInt32 (instrSession,
                             command,
                             markerDelay,
                             VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Fix Marker Delay To Current Range
 * Purpose:  This function restricts the marker delay setting range to the 
 *           dynamic range. In this range the delay can be set without 
 *           restarting the marker and signal. If a delay is entered in setting 
 *           ON but is outside this range, the maximum possible delay is set 
 *           and an error message is output.
 *           
 *           The numeric suffix in OUTPut has no significance for this 
 *           command, since the setting always affects every marker.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetFixMarkerDelayToCurrentRange(
    ViSession   instrSession,
    ViBoolean   fixedMarkerDelay
)
{
    return rsafq_setViBoolean (instrSession,
                               ":TRIG:OUTP:DEL:FIX",
                               fixedMarkerDelay,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Fix Marker Delay To Current Range
 * Purpose:  This function returns the marker delay setting range to the 
 *           dynamic range. In this range the delay can be set without 
 *           restarting the marker and signal. If a delay is entered in setting 
 *           ON but is outside this range, the maximum possible delay is set 
 *           and an error message is output.
 *           
 *           The numeric suffix in OUTPut has no significance for this 
 *           command, since the setting always affects every marker.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetFixMarkerDelayToCurrentRange(
    ViSession   instrSession,
    ViBoolean*  fixedMarkerDelay
)
{
    return rsafq_getViBoolean (instrSession,
                               ":TRIG:OUTP:DEL:FIX?",
                               fixedMarkerDelay,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Fixed Marker Delay Maximum
 * Purpose:  This function queries the maximum marker delay for setting 
 *           :TRIGger:OUTPut:DELay:FIXed ON.
*****************************************************************************/
ViStatus _VI_FUNC rsafq_GetFixedMarkerDelayMaximum(
    ViSession   instrSession,
    ViInt32 	markerOutput,
    ViReal64*   fixedMarkerDelayMaximum
)
{
    ViChar  command[256];
    
    CHECKPARAM (rsafq_invalidViInt32Range (markerOutput,1,4), 2);
    
    sprintf (command,"SOUR:TRIG:OUTP%ld:DEL:MAX?", markerOutput);
    
    return rsafq_getViReal64 (instrSession,
                              command,
                              fixedMarkerDelayMaximum,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Fixed Marker Delay Minimum
 * Purpose:  This function queries the minimum marker delay for setting 
 *           :TRIGger:OUTPut:DELay:FIXed ON.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetFixedMarkerDelayMinimum(
    ViSession   instrSession,
    ViInt32 markerOutput,
    ViReal64*   fixedMarkerDelayMinimum
)
{
    ViChar  command[256];
    
    CHECKPARAM (rsafq_invalidViInt32Range (markerOutput,1,4), 2);
    
    sprintf (command,"SOUR:TRIG:OUTP%ld:DEL:MIN?", markerOutput);
    
    return rsafq_getViReal64 (instrSession,
                              command,
                              fixedMarkerDelayMinimum,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Configure Clock
 * Purpose:  This function  is used to select and configure the clock source.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ConfigureClock(
    ViSession   instrSession,
    ViInt32 clockSource,
    ViReal64    clockFrequency,
    ViInt32 inputSlope,
    ViReal64    thresholdDataInput,
    ViInt32 impedanceDataInput,
    ViInt32 roscSource,
    ViInt32 roscSynchronizationBandwidth
)
{
    ViStatus      rsafq_status;

    
    CHECKERR (rsafq_SetClockSource (instrSession, clockSource));
    
    rsafq_status = rsafq_SetClockFrequency (instrSession, clockFrequency);
    REMAP(2,3);
    
    rsafq_status = rsafq_SetClockInputSlope (instrSession, inputSlope);
    REMAP(2,4);
    
    rsafq_status = rsafq_SetThresholdDataInput (instrSession, thresholdDataInput);
    REMAP(2,5);
    
    rsafq_status = rsafq_SetImpedanceDataInput (instrSession, impedanceDataInput);
    REMAP(2,6);
    
    rsafq_status = rsafq_SetROSCSource (instrSession, roscSource);
    REMAP(2,7);
    
    rsafq_status = rsafq_SetROSCSynchronizationBandwidth (instrSession, roscSynchronizationBandwidth);
    REMAP(2,8);
    
    return rsafq_status;
    
}

/*****************************************************************************
 * Function: Set Clock Source
 * Purpose:  This function selects the memory clock source (internal or 
 *           external).
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetClockSource(
    ViSession   instrSession,
    ViInt32 clockSource
)
{
    return rsafq_setEnum (instrSession,
                          ":CLOC:SOUR",
                          clockSource,
                          sourceArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Clock Source
 * Purpose:  This function returns the memory clock source (internal or 
 *           external).
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetClockSource(
    ViSession   instrSession,
    ViInt32*    clockSource
)
{
    return rsafq_getEnum (instrSession,
                          ":CLOC:SOUR?",
                          clockSource,
                          sourceArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Clock Frequency
 * Purpose:  This function sets the output clock rate of the ARB. By default, 
 *           the clock rate defined by clock tag of waveform.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetClockFrequency(
    ViSession   instrSession,
    ViReal64    clockFrequency
)
{
    return rsafq_setViReal64 (instrSession,
                              "CLOC",
                              clockFrequency,
                              1.0e3,300.0e6,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Clock Frequency
 * Purpose:  This function returns the output clock rate.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetClockFrequency(
    ViSession   instrSession,
    ViReal64*   clockFrequency
)
{
    return rsafq_getViReal64 (instrSession,
                              "CLOC?",
                              clockFrequency,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Clock Input Slope
 * Purpose:  This function sets the active slope of an externally applied 
 *           clock signal at the CLOCK connector. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetClockInputSlope(
    ViSession   instrSession,
    ViInt32 inputSlope
)
{
    return rsafq_setEnum (instrSession,
                          "CLOC:INP:SLOP",
                          inputSlope,
                          inputSlopeArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Clock Input Slope
 * Purpose:  This function returns the active slope of an externally applied 
 *           clock signal at the CLOCK connector. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetClockInputSlope(
    ViSession   instrSession,
    ViInt32*    inputSlope
)
{
    return rsafq_getEnum (instrSession,
                          "CLOC:INP:SLOP?",
                          inputSlope,
                          inputSlopeArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Clock Wideband Mode 
 * Purpose:  This function queries the the current operating mode extended 
 *           (wideband) / normal of the instrument. The operating mode is 
 *           defined by the samlple rate of the loaded waveform.
 *           
 *           Note(s):
 *           (1) This function is available on AFQ100B only.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetClockWidebandMode(
    ViSession   instrSession,
    ViBoolean*  widebandMode
)
{
	INIT;
	CHECKVER(AFQ100B);
    return rsafq_getViBoolean (instrSession,
                                      "SOUR:CLOC:WBAN:STAT?",
                                      widebandMode,
                                      VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Clock Input Frequency
 * Purpose:  This function returns the measured frequency of the external 
 *           clock signal. An external clock reference must be supplied at the 
 *           CLOCK input.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ClockInputFrequency(
    ViSession   instrSession,
    ViReal64*   clockInputFrequency
)
{
    return rsafq_getViReal64 (instrSession,
                              "CLOC:INP:FREQ?",
                              clockInputFrequency,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set ROSC Source
 * Purpose:  This function sets selects an internal or external reference 
 *           oscillator source.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetROSCSource(
    ViSession   instrSession,
    ViInt32 roscSource
)
{
    return rsafq_setEnum (instrSession,
                          "ROSC:SOUR",
                          roscSource,
                          sourceArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get ROSC Source
 * Purpose:  This function returns the reference clock type.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetROSCSource(
    ViSession   instrSession,
    ViInt32*    roscSource
)
{
    return rsafq_getEnum (instrSession,
                          "ROSC:SOUR?",
                          roscSource,
                          sourceArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set ROSC Synchronization Bandwidth
 * Purpose:  This function sets the synchronization bandwidth, i.e. the target 
 *           range of the synchronization circuit.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetROSCSynchronizationBandwidth(
    ViSession   instrSession,
    ViInt32 roscSynchronizationBandwidth
)
{
    return rsafq_setEnum (instrSession,
                          "ROSC:EXT:SBAN",
                          roscSynchronizationBandwidth,
                          roscSynchronizationBandwidthArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get ROSC Synchronization Bandwidth
 * Purpose:  This function returns synchronization bandwidth.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetROSCSynchronizationBandwidth(
    ViSession   instrSession,
    ViInt32*    roscSynchronizationBandwidth
)
{
    return rsafq_getEnum (instrSession,
                          "ROSC:EXT:SBAN?",
                          roscSynchronizationBandwidth,
                          roscSynchronizationBandwidthArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set ROSC Synchronization Mode
 * Purpose:  This function selects the mode for synchronization with other 
 *           R&S AFQ waveform generators.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetROSCSynchronizationMode (ViSession instrSession,
                                                    ViInt32 ROSCSynchronizationMode)
{
    return rsafq_setEnum (instrSession,
                          "SOUR:ROSC:SYNC:MODE",
                          ROSCSynchronizationMode,
                          ROSCSynchronizationModeArr,
                          VI_ERROR_PARAMETER2);    
}

/*****************************************************************************
 * Function: Get ROSC Synchronization Mode
 * Purpose:  This function returns the mode for synchronization with other 
 *           R&S AFQ waveform generators.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetROSCSynchronizationMode (ViSession instrSession,
                                                    ViInt32 *ROSCSynchronizationMode)
{
    return rsafq_getEnum (instrSession,
                          "SOUR:ROSC:SYNC:MODE?",
                          ROSCSynchronizationMode,
                          ROSCSynchronizationModeArr,
                          VI_ERROR_PARAMETER2);    
}

/*****************************************************************************
 * Function: Set ROSC Delay
 * Purpose:  This function sets a time shift (delay) to the reference clock 
 *           signal for adjusting the clocks in the slaves.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetROSCDelay(
    ViSession   instrSession,
    ViReal64    delay
)
{
    return rsafq_setViReal64 (instrSession,"SOUR:ROSC:DEL",delay,0.0,5.0,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get ROSC Delay
 * Purpose:  This function queries a time shift (delay) to the reference clock 
 *           signal for adjusting the clocks in the slaves.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetROSCDelay(
    ViSession   instrSession,
    ViReal64*   delay
)
{
    return rsafq_getViReal64 (instrSession,"SOUR:ROSC:DEL?",delay,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Configure Time Shifts
 * Purpose:  This function set time and frequency shifts.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ConfigureTimeShifts(
    ViSession   instrSession,
    ViReal64    iPathDelay,
    ViReal64    qPathDelay,
    ViReal64    frequencyOffset
)
{
    ViStatus    rsafq_status;

    CHECKERR (rsafq_SetIPathDelay (instrSession, iPathDelay));

    rsafq_status = rsafq_SetQPathDelay (instrSession, qPathDelay);
    REMAP (2, 3);
    
    rsafq_status = rsafq_SetFrequencyOffset (instrSession, frequencyOffset);
    REMAP (2, 4);
    
    RETURN;
}

/*****************************************************************************
 * Function: Set I Path Delay
 * Purpose:  This function defines the delay in the I path.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetIPathDelay(
    ViSession   instrSession,
    ViReal64    iPathDelay
)
{
    return rsafq_setViReal64 (instrSession,
                              "IMP:DEL:I",
                              iPathDelay,
                              -2,2,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get I Path Delay
 * Purpose:  This function returns the delay in the I path
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetIPathDelay(
    ViSession   instrSession,
    ViReal64*   iPathDelay
)
{
    return rsafq_getViReal64 (instrSession,
                              "IMP:DEL:I?",
                              iPathDelay,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Q Path Delay
 * Purpose:  This function defines the delay in the Q path.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetQPathDelay(
    ViSession   instrSession,
    ViReal64    qPathDelay
)
{
    return rsafq_setViReal64 (instrSession,
                              "IMP:DEL:Q",
                              qPathDelay,
                              -2,2,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Q Path Delay
 * Purpose:  This function returns the delay in the Q path
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetQPathDelay(
    ViSession   instrSession,
    ViReal64*   qPathDelay
)
{
    return rsafq_getViReal64 (instrSession,
                              "IMP:DEL:Q?",
                              qPathDelay,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Frequency Offset
 * Purpose:  This function sets an intermediate frequency.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetFrequencyOffset(
    ViSession   instrSession,
    ViReal64    frequencyOffset
)
{
    return rsafq_setViReal64 (instrSession,
                              "FOFF",
                              frequencyOffset,
                              0.0,100e6,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Frequency Offset
 * Purpose:  This function returns an intermediate frequency.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetFrequencyOffset(
    ViSession   instrSession,
    ViReal64*   frequencyOffset
)
{
    return rsafq_getViReal64 (instrSession,
                              "FOFF?",
                              frequencyOffset,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Configure IQ Impairment
 * Purpose:  This function configures and activates the three impairment or 
 *           correction values for IQ modulation.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ConfigureIQImpairment(
    ViSession   instrSession,
    ViReal64    iGain,
    ViReal64    qGain,
    ViReal64    iOffset,
    ViReal64    qOffset,
    ViReal64    phaseMismatch
)
{
    ViStatus    rsafq_status;

    CHECKERR (rsafq_SetIQModulationGainIPath (instrSession, iGain));

    rsafq_status = rsafq_SetIQModulationGainQPath (instrSession, qGain);
    REMAP (2, 3);
    
    rsafq_status = rsafq_SetIQModulationIOffset (instrSession, iOffset);
    REMAP (2, 4);
    
    rsafq_status = rsafq_SetIQModulationQOffset (instrSession, qOffset);
    REMAP (2, 5);
    
    rsafq_status = rsafq_SetIQModulationPhase (instrSession, phaseMismatch);
    REMAP (2, 6);
    
    RETURN;
}

/*****************************************************************************
 * Function: Set IQ Modulation Impairments State
 * Purpose:  This function activates (ON) and deactivates (OFF) impairments.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetIQModulationImpairmentsState(
    ViSession   instrSession,
    ViBoolean   impairmentsState
)
{
    return rsafq_setViBoolean (instrSession,
                               ":IMP:STAT",
                               impairmentsState,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get IQ Modulation Impairments State
 * Purpose:  This function returns impairments state.
*****************************************************************************/
ViStatus _VI_FUNC rsafq_GetIQModulationImpairmentsState(
    ViSession   instrSession,
    ViBoolean*  impairmentsState
)
{
    return rsafq_getViBoolean (instrSession,
                               ":IMP:STAT?",
                               impairmentsState,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set IQ Modulation I Offset
 * Purpose:  This function sets the sets a DC offset in the I path.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetIQModulationIOffset(
    ViSession   instrSession,
    ViReal64    iOffset
)
{
    return rsafq_setViReal64 (instrSession,
                              ":IMP:OFFS:I",
                              iOffset,
                              -0.1,0.1,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get IQ Modulation I Offset
 * Purpose:  This function returns the DC offset in the I path.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetIQModulationIOffset(
    ViSession   instrSession,
    ViReal64*   iOffset
)
{
    return rsafq_getViReal64 (instrSession,
                              ":IMP:OFFS:I?",
                              iOffset,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set IQ Modulation Q Offset
 * Purpose:  This function sets a DC offset in the Q path.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetIQModulationQOffset(
    ViSession   instrSession,
    ViReal64    qOffset
)
{
    return rsafq_setViReal64 (instrSession,
                              ":IMP:OFFS:Q",
                              qOffset,
                              -0.1,0.1,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get IQ Modulation Q Offset
 * Purpose:  This function returns a DC offset in the Q path.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetIQModulationQOffset(
    ViSession   instrSession,
    ViReal64*   qOffset
)
{
    return rsafq_getViReal64 (instrSession,
                              ":IMP:OFFS:Q?",
                              qOffset,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set IQ Modulation Gain I Path
 * Purpose:  This function sets a gain in the I path.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetIQModulationGainIPath(
    ViSession   instrSession,
    ViReal64    iGain
)
{
    return rsafq_setViReal64 (instrSession,
                              ":IMP:GAIN:I",
                              iGain,
                              -2,0,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get IQ Modulation Gain I Path
 * Purpose:  This function returns a gain in the I path.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetIQModulationGainIPath(
    ViSession   instrSession,
    ViReal64*   iGain
)
{
    return rsafq_getViReal64 (instrSession,
                              ":IMP:GAIN:I?",
                              iGain,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set IQ Modulation Gain Q Path
 * Purpose:  This function sets a gain in the Q path.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetIQModulationGainQPath(
    ViSession   instrSession,
    ViReal64    qGain
)
{
    return rsafq_setViReal64 (instrSession,
                              ":IMP:GAIN:Q",
                              qGain,
                              -2,0,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get IQ Modulation Gain Q Path
 * Purpose:  This function returns a gain in the Q path.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetIQModulationGainQPath(
    ViSession   instrSession,
    ViReal64*   qGain
)
{
    return rsafq_getViReal64 (instrSession,
                              ":IMP:GAIN:Q?",
                              qGain,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set IQ Modulation Phase
 * Purpose:  This function sets an I/Q mismatch phase.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetIQModulationPhase(
    ViSession   instrSession,
    ViReal64    phaseMismatch
)
{
    return rsafq_setViReal64 (instrSession,
                              ":IMP:PHAS",
                              phaseMismatch,
                              -5,5,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get IQ Modulation Phase
 * Purpose:  This function returns an I/Q mismatch phase.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetIQModulationPhase(
    ViSession   instrSession,
    ViReal64*   phaseMismatch
)
{
    return rsafq_getViReal64 (instrSession,
                              ":IMP:PHAS?",
                              phaseMismatch,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Equalizer State
 * Purpose:  This function switches on or off both the I/Q and modulator 
 *           corrections.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetEqualizerState(
    ViSession   instrSession,
    ViBoolean   equalizerState
)
{
    return rsafq_setViBoolean (instrSession,
                               "EQU:STAT",
                               equalizerState,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Equalizer State
 * Purpose:  This function returns the equalizer state setting.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetEqualizerState(
    ViSession   instrSession,
    ViBoolean*  equalizerState
)
{
    return rsafq_getViBoolean (instrSession,
                               "EQU:STAT?",
                               equalizerState,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Equalizer IQ State
 * Purpose:  This function switches on or off I/Q corrections.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetEqualizerIQState(
    ViSession   instrSession,
    ViBoolean   iqState
)
{
    return rsafq_setViBoolean (instrSession,
                               "EQU:IQ:STAT",
                               iqState,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Equalizer IQ State
 * Purpose:  This function returns the IQ state setting.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetEqualizerIQState(
    ViSession   instrSession,
    ViBoolean*  iqState
)
{
    return rsafq_getViBoolean (instrSession,
                               "EQU:IQ:STAT?",
                               iqState,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Equalizer IQ Catalog
 * Purpose:  This function lists all files containing I/Q equalizer data 
 *           tables saved on local hard disk.
 *           
 *           This function reads out the subdirectories and files in the 
 *           specified directory.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_EqualizerIQCatalog(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViChar  catalog[]
)
{
    return rsafq_getViString (instrSession,
                              "EQU:IQ:CAT?",
                              catalog,
                              arraySize,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Equalizer IQ Catalog Length
 * Purpose:  This function queries the number of files containing I/Q 
 *           equalizer data tables saved on local hard disk.
 *           
 *           This command reads out the number of files in the specified 
 *           directory.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetEqualizerIQCatalogLength(
    ViSession   instrSession,
    ViString    filePath,
    ViInt32*    catalogLength
)
{
    ViChar  buffer [BUFFER_SIZE];

    if (!filePath)
        sprintf (buffer, "EQU:IQ:CAT:LENG?");
    else
        sprintf (buffer, "EQU:IQ:CAT:LENG? '%s'", filePath);

    return rsafq_getViInt32 (instrSession,
                             buffer,
                             catalogLength,
                             VI_ERROR_PARAMETER3);

}

/*****************************************************************************
 * Function: Equalizer IQ I Select
 * Purpose:  This function loads the specified file containing a data table 
 *           for the I path. All new data points will be saved in the selected 
 *           file, hence there is no 'save' or 'save as...' command. If the 
 *           specified file does not exist, the instrument will create a new 
 *           one, assigning the entered file name.
 *           
 *           If no directory is specified, the default directory selected with 
 *           the MMEM:CDIR command is read out on the default drive selected 
 *           with the MMEM:MSIS command.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_EqualizerIQISelect(
    ViSession   instrSession,
    ViString    fileName
)
{
    return rsafq_setViStringQuoted (instrSession,
                                    "EQU:IQ:I:SEL",
                                    "'",
                                    fileName,
                                    VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Equalizer IQ Q Select
 * Purpose:  This function loads the specified file containing a data table 
 *           for the Q path. All new data points will be saved in the selected 
 *           file, hence there is no 'save' or 'save as...' command. If the 
 *           specified file does not exist, the instrument will create a new 
 *           one, assigning the entered file name.
 *           
 *           If no directory is specified, the default directory selected with 
 *           the MMEM:CDIR command is read out on the default drive selected 
 *           with the MMEM:MSIS command.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_EqualizerIQQSelect(
    ViSession   instrSession,
    ViString    fileName
)
{
    return rsafq_setViStringQuoted (instrSession,
                              "EQU:IQ:Q:SEL",
                              "'",
                              fileName,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Delete IQ Equalizer Data
 * Purpose:  This function deletes the selected I/Q equalizer data table from 
 *           local hard disk.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_DeleteIQEqualizerData(
    ViSession   instrSession,
    ViString    fileName
)
{
    return rsafq_setViStringQuoted (instrSession,
                              "EQU:IQ:DEL",
                              "'",
                              fileName,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Configure Equalizer IQ I Table
 * Purpose:  This function configures equalizer table I data.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ConfigureEqualizerIQITable(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViReal64    iFrequency[],
    ViReal64    iGain[],
    ViReal64    iPhase[]
)
{
    ViStatus    rsafq_status;

    CHECKERR (rsafq_SetEqualizerIQIFrequency (instrSession, arraySize, iFrequency));

    rsafq_status = rsafq_SetEqualizerIQIGain (instrSession, arraySize, iGain);
    REMAP (2, 3);
    
    rsafq_status = rsafq_SetEqualizerIQIPhase (instrSession, arraySize, iPhase);
    REMAP (2, 4);
    
    RETURN;
}

/*****************************************************************************
 * Function: Configure Equalizer IQ Q Table
 * Purpose:  This function configures equalizer table Q data.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ConfigureEqualizerIQQTable(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViReal64    qFrequency[],
    ViReal64    qGain[],
    ViReal64    qPhase[]
)
{
    ViStatus    rsafq_status;

    CHECKERR (rsafq_SetEqualizerIQQFrequency (instrSession, arraySize, qFrequency));

    rsafq_status = rsafq_SetEqualizerIQQGain (instrSession, arraySize, qGain);
    REMAP (2, 3);
    
    rsafq_status = rsafq_SetEqualizerIQQPhase (instrSession, arraySize, qPhase);
    REMAP (2, 4);
    
    RETURN;
}

/*****************************************************************************
 * Function: Set Equalizer IQ I Frequency
 * Purpose:  This function enters frequency values into the current data table 
 *           of the I path selected using the EQU:IQ:I:SEL command.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetEqualizerIQIFrequency(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViReal64    _VI_FAR iFrequency[]
)
{
    ViChar* pbuffer;
    ViChar* p2buf;
    ViInt32 i;

    CHECKNULL (iFrequency, 3);
    
    p2buf = pbuffer = malloc(arraySize*20 + 30);
    p2buf += sprintf(pbuffer, ":FORM ASC;EQU:IQ:I:FREQ ");

    for (i = 0; i < arraySize; i++)
        p2buf += sprintf (p2buf, "%lf,", iFrequency[i]);

    *--p2buf = '\0';

    return rsafq_sendCmd (instrSession, pbuffer);
}

/*****************************************************************************
 * Function: Get Equalizer IQ I Frequency
 * Purpose:  This function returns I frequency.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetEqualizerIQIFrequency(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViReal64    _VI_FAR iFrequency[]
)
{
    return rsafq_getViReal64Array (instrSession,
                                   ":FORM ASC;EQU:IQ:I:FREQ?",
                                   iFrequency,
                                   arraySize,
                                   VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Equalizer IQ Q Frequency
 * Purpose:  This function enters frequency values into the current data table 
 *           of the Q path.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetEqualizerIQQFrequency(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViReal64    _VI_FAR qFrequency[]
)
{
    ViChar* pbuffer;
    ViChar* p2buf;
    ViInt32 i;

    CHECKNULL (qFrequency, 3);
    
    p2buf = pbuffer = malloc(arraySize*20 + 30);
    p2buf += sprintf(pbuffer, ":FORM ASC;EQU:IQ:Q:FREQ ");

    for (i = 0; i < arraySize; i++)
        p2buf += sprintf (p2buf, "%lf,", qFrequency[i]);

    *--p2buf = '\0';

    return rsafq_sendCmd (instrSession, pbuffer);
}

/*****************************************************************************
 * Function: Get Equalizer IQ Q Frequency
 * Purpose:  This function returns Q frequency.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetEqualizerIQQFrequency(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViReal64    _VI_FAR qFrequency[]
)
{
    return rsafq_getViReal64Array (instrSession,
                                   ":FORM ASC;EQU:IQ:Q:FREQ?",
                                   qFrequency,
                                   arraySize,
                                   VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Equalizer IQ I Gain
 * Purpose:  This function enters gain values into the current data table of 
 *           the I path.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetEqualizerIQIGain(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViReal64    _VI_FAR iGain[]
)
{
    ViChar* pbuffer;
    ViChar* p2buf;
    ViInt32 i;

    CHECKNULL (iGain, 3);
    
    p2buf = pbuffer = malloc(arraySize*20 + 30);
    p2buf += sprintf(pbuffer, ":FORM ASC;EQU:IQ:I:GAIN ");

    for (i = 0; i < arraySize; i++)
        p2buf += sprintf (p2buf, "%lf,", iGain[i]);

    *--p2buf = '\0';

    return rsafq_sendCmd (instrSession, pbuffer);
}

/*****************************************************************************
 * Function: Get Equalizer IQ I Gain
 * Purpose:  This function returns I gain.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetEqualizerIQIGain(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViReal64    _VI_FAR iGain[]
)
{
    return rsafq_getViReal64Array (instrSession,
                                   ":FORM ASC;EQU:IQ:I:GAIN?",
                                   iGain,
                                   arraySize,
                                   VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Equalizer IQ Q Gain
 * Purpose:  This function enters gain values into the current data table of 
 *           the Q path.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetEqualizerIQQGain(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViReal64    _VI_FAR qGain[]
)
{
    ViChar* pbuffer;
    ViChar* p2buf;
    ViInt32 i;

    CHECKNULL (qGain, 3);
    
    p2buf = pbuffer = malloc(arraySize*20 + 30);
    p2buf += sprintf(pbuffer, ":FORM ASC;EQU:IQ:Q:GAIN ");

    for (i = 0; i < arraySize; i++)
        p2buf += sprintf (p2buf, "%lf,", qGain[i]);

    *--p2buf = '\0';

    return rsafq_sendCmd (instrSession, pbuffer);
}

/*****************************************************************************
 * Function: Get Equalizer IQ Q Gain
 * Purpose:  This function returns Q gain.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetEqualizerIQQGain(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViReal64    _VI_FAR qGain[]
)
{
    return rsafq_getViReal64Array (instrSession,
                                   ":FORM ASC;EQU:IQ:Q:GAIN?",
                                   qGain,
                                   arraySize,
                                   VI_ERROR_PARAMETER3);
                                   
}                                   

/*****************************************************************************
 * Function: Set Equalizer IQ I Phase
 * Purpose:  This function enters phase values into the current data table of 
 *           the I path.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetEqualizerIQIPhase(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViReal64    _VI_FAR iPhase[]
)
{
    ViChar* pbuffer;
    ViChar* p2buf;
    ViInt32 i;

    CHECKNULL (iPhase, 3);
    
    p2buf = pbuffer = malloc(arraySize*20 + 30);
    p2buf += sprintf(pbuffer, ":FORM ASC;EQU:IQ:I:PHAS ");

    for (i = 0; i < arraySize; i++)
        p2buf += sprintf (p2buf, "%lf,", iPhase[i]);

    *--p2buf = '\0';

    return rsafq_sendCmd (instrSession, pbuffer);
}

/*****************************************************************************
 * Function: Get Equalizer IQ I Phase
 * Purpose:  This function returns I phase.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetEqualizerIQIPhase(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViReal64    _VI_FAR iPhase[]
)
{
    return rsafq_getViReal64Array (instrSession,
                               ":FORM ASC;EQU:IQ:I:PHAS?",
                               iPhase,
                               arraySize,
                               VI_ERROR_PARAMETER3);
                                   
}

/*****************************************************************************
 * Function: Set Equalizer IQ Q Phase
 * Purpose:  This function enters phase values into the current data table of 
 *           the Q path.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetEqualizerIQQPhase(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViReal64    _VI_FAR qPhase[]
)
{
    ViChar* pbuffer;
    ViChar* p2buf;
    ViInt32 i;

    CHECKNULL (qPhase, 3);
    
    p2buf = pbuffer = malloc(arraySize*20 + 30);
    p2buf += sprintf(pbuffer, ":FORM ASC;EQU:IQ:Q:PHAS ");

    for (i = 0; i < arraySize; i++)
        p2buf += sprintf (p2buf, "%lf,", qPhase[i]);

    *--p2buf = '\0';

    return rsafq_sendCmd (instrSession, pbuffer);
}

/*****************************************************************************
 * Function: Get Equalizer IQ Q Phase
 * Purpose:  This function returns q phase.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetEqualizerIQQPhase(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViReal64    _VI_FAR qPhase[]
)
{
    return rsafq_getViReal64Array (instrSession,
                               ":FORM ASC;EQU:IQ:Q:PHAS?",
                               qPhase,
                               arraySize,
                               VI_ERROR_PARAMETER3);
                                   
}

/*****************************************************************************
 * Function: Get Equalizer IQ I Points
 * Purpose:  This function queries the number of data points of the current 
 *           data table of the I path, i.e. the number of filled rows.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetEqualizerIQIPoints(
    ViSession   instrSession,
    ViInt32*    numberOfDataIPoints
)
{
    return rsafq_getViInt32 (instrSession,
                             "EQU:IQ:I:POIN?",
                             numberOfDataIPoints,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Equalizer IQ Q Points
 * Purpose:  This function queries the number of data points of the current 
 *           data table of the Q path.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetEqualizerIQQPoints(
    ViSession   instrSession,
    ViInt32*    numberOfDataQPoints
)
{
    return rsafq_getViInt32 (instrSession,
                             "EQU:IQ:Q:POIN?",
                             numberOfDataQPoints,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Equalizer Modulator State
 * Purpose:  This function switches on or off the corrections for the 
 *           modulator.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetEqualizerModulatorState(
    ViSession   instrSession,
    ViBoolean   modulatorState
)
{
    return rsafq_setViBoolean (instrSession,
                               "EQU:MOD:STAT",
                               modulatorState,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Equalizer Modulator State
 * Purpose:  This function returns the modulator state setting.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetEqualizerModulatorState(
    ViSession   instrSession,
    ViBoolean*  modulatorState
)
{
    return rsafq_getViBoolean (instrSession,
                               "EQU:MOD:STAT?",
                               modulatorState,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Frequency Entry Mode
 * Purpose:  This function Selects the frequency entry mode.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetFrequencyEntryMode(
    ViSession   instrSession,
    ViInt32 frequencyEntryMode
)
{
    return rsafq_setEnum (instrSession,
                          "EQU:MOD:FEM",
                          frequencyEntryMode,
                          frequencyEntryModeArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Frequency Entry Mode
 * Purpose:  This function returns the frequency entry mode.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetFrequencyEntryMode(
    ViSession   instrSession,
    ViInt32*    frequencyEntryMode
)
{
    return rsafq_getEnum (instrSession,
                          "EQU:MOD:FEM?",
                          frequencyEntryMode,
                          frequencyEntryModeArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Equalizer Modulator Catalog
 * Purpose:  This function lists all files containing equalizer data tables 
 *           for the modulator saved on local hard disk.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_EqualizerModulatorCatalog(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViChar  catalog[]
)
{
    return rsafq_getViString (instrSession,
                              "EQU:MOD:CAT?",
                              catalog,
                              arraySize,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Equalizer Modulator Catalog Length
 * Purpose:  This function queries the number of files containing equalizer 
 *           data tables for the modulator saved on local hard disk.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetEqualizerModulatorCatalogLength(
    ViSession   instrSession,
    ViString    filePath,
    ViInt32*    catalogLength
)
{
    ViChar  buffer [BUFFER_SIZE];

    if (!filePath)
        sprintf (buffer, "EQU:MOD:CAT:LENG?");
    else
        sprintf (buffer, "EQU:MOD:CAT:LENG? '%s'", filePath);

    return rsafq_getViInt32 (instrSession,
                             buffer,
                             catalogLength,
                             VI_ERROR_PARAMETER3);

}

/*****************************************************************************
 * Function: Equalizer Modulator Select
 * Purpose:  This function selects a file containing the equalizer data table 
 *           for the modulator previously entered by the user. All new data 
 *           points will be saved in the selected one, hence there is no 'save' 
 *           or 'save as...' command.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_EqualizerModulatorSelect(
    ViSession   instrSession,
    ViString    fileName
)
{
    return rsafq_setViStringQuoted (instrSession,
                                    "EQU:MOD:SEL",
                                    "'",
                                    fileName,
                                    VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Delete Equalizer Modulator Data
 * Purpose:  This function deletes an equalizer data table for the modulator 
 *           from local hard disk.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_DeleteEqualizerModulatorData(
    ViSession   instrSession,
    ViString    fileName
)
{
    return rsafq_setViStringQuoted (instrSession,
                                    "EQU:MOD:DEL",
                                    "'",
                                    fileName,
                                    VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Configure Equalizer Modulator Table
 * Purpose:  This function configures equalizer table data.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ConfigureEqualizerModulatorTable(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViReal64    frequency[],
    ViReal64    gain[],
    ViReal64    phase[]
)
{
    ViStatus    rsafq_status;

    CHECKERR (rsafq_SetEqualizerModulatorFrequency (instrSession, arraySize, frequency));

    rsafq_status = rsafq_SetEqualizerModulatorGain (instrSession, arraySize, gain);
    REMAP (3, 4);
    
    rsafq_status = rsafq_SetEqualizerModulatorPhase (instrSession, arraySize, phase);
    REMAP (3, 5);
    
    RETURN;
}

/*****************************************************************************
 * Function: Set Equalizer Modulator Frequency
 * Purpose:  This function enters frequency values into the current equalizer 
 *           data table of the modulator.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetEqualizerModulatorFrequency(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViReal64    _VI_FAR frequency[]
)
{
    ViChar* pbuffer;
    ViChar* p2buf;
    ViInt32 i;

    CHECKNULL (frequency, 3);
    
    p2buf = pbuffer = malloc(arraySize*20 + 30);
    p2buf += sprintf(pbuffer, ":FORM ASC;EQU:MOD:FREQ ");

    for (i = 0; i < arraySize; i++)
        p2buf += sprintf (p2buf, "%lf,", frequency[i]);

    *--p2buf = '\0';

    return rsafq_sendCmd (instrSession, pbuffer);
}

/*****************************************************************************
 * Function: Get Equalizer Modulator Frequency
 * Purpose:  This function returns modulator frequency.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetEqualizerModulatorFrequency(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViReal64    _VI_FAR frequency[]
)
{
    return rsafq_getViReal64Array (instrSession,
                                   ":FORM ASC;EQU:MOD:FREQ?",
                                   frequency,
                                   arraySize,
                                   VI_ERROR_PARAMETER3);
  
}

/*****************************************************************************
 * Function: Set Equalizer Modulator Gain
 * Purpose:  This function enters gain values into the current equalizer data 
 *           table of the modulator.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetEqualizerModulatorGain(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViReal64    gain[]
)
{
    ViChar* pbuffer;
    ViChar* p2buf;
    ViInt32 i;

    CHECKNULL (gain, 3);
    
    p2buf = pbuffer = malloc(arraySize*20 + 30);
    p2buf += sprintf(pbuffer, ":FORM ASC;EQU:MOD:GAIN ");

    for (i = 0; i < arraySize; i++)
        p2buf += sprintf (p2buf, "%lf,", gain[i]);

    *--p2buf = '\0';

    return rsafq_sendCmd (instrSession, pbuffer);
}

/*****************************************************************************
 * Function: Get Equalizer Modulator Gain
 * Purpose:  This function returns gain value.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetEqualizerModulatorGain(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViReal64    gain[]
)
{
    return rsafq_getViReal64Array (instrSession,
                                   ":FORM ASC;EQU:MOD:GAIN?",
                                   gain,
                                   arraySize,
                                   VI_ERROR_PARAMETER3);
  
}

/*****************************************************************************
 * Function: Set Equalizer Modulator Phase
 * Purpose:  This function enters phase values into the current equalizer data 
 *           table of the modulator.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetEqualizerModulatorPhase(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViReal64    phase[]
)
{
    ViChar* pbuffer;
    ViChar* p2buf;
    ViInt32 i;

    CHECKNULL (phase, 3);
    
    p2buf = pbuffer = malloc(arraySize*20 + 30);
    p2buf += sprintf(pbuffer, ":FORM ASC;EQU:MOD:PHAS ");

    for (i = 0; i < arraySize; i++)
        p2buf += sprintf (p2buf, "%lf,", phase[i]);

    *--p2buf = '\0';

    return rsafq_sendCmd (instrSession, pbuffer);
}

/*****************************************************************************
 * Function: Get Equalizer Modulator Phase
 * Purpose:  This function returns phase value.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetEqualizerModulatorPhase(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViReal64    phase[]
)
{
    return rsafq_getViReal64Array (instrSession,
                                   ":FORM ASC;EQU:MOD:PHAS?",
                                   phase,
                                   arraySize,
                                   VI_ERROR_PARAMETER3);
  
}

/*****************************************************************************
 * Function: Set RF Frequency
 * Purpose:  This function sets the RF carrier frequency of the modulator, 
 *           this is necessary if frequency entry mode 'absolute' is chosen.
 *           
 *           The instrument compensates frequency responses of RF modulators. 
 *           However, the region of interest is the accessible bandwidth of 200 
 *           MHz centered on the RF carrier frequency. This is why the user 
 *           enters the carrier frequency if frequency entry mode 'absolute' is 
 *           chosen. Outside this region, the instrument does not correct 
 *           anything.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetRFFrequency(
    ViSession   instrSession,
    ViReal64    rfFrequency
)
{
    return rsafq_setViReal64 (instrSession,
                              "EQU:MOD:RFFR",
                              rfFrequency,
                              0.0,100.0e9,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get RF Frequency
 * Purpose:  This function returns the RF carrier frequency of the modulator.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetRFFrequency(
    ViSession   instrSession,
    ViReal64*   rfFrequency
)
{
    return rsafq_getViReal64 (instrSession,
                              "EQU:MOD:RFFR?",
                              rfFrequency,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Equalizer Modulator Points
 * Purpose:  This function queries the number of data points of the current 
 *           equalizer data table of the modulator.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetEqualizerModulatorPoints(
    ViSession   instrSession,
    ViInt32*    numberOfDataPoints
)
{
    return rsafq_getViInt32 (instrSession,
                             "EQU:MOD:POIN?",
                             numberOfDataPoints,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Output State
 * Purpose:  This function switches on or off the active output.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetOutputState(
    ViSession   instrSession,
    ViBoolean   outputState
)
{
    return rsafq_setViBoolean (instrSession,
                               "OUTP:STAT",
                               outputState,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Output State
 * Purpose:  This function returns the output state setting.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetOutputState(
    ViSession   instrSession,
    ViBoolean*  outputState
)
{
    return rsafq_getViBoolean (instrSession,
                               "OUTP:STAT?",
                               outputState,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Active Output
 * Purpose:  This function selects the active outputs. Two outputs cannot be 
 *           active at the same time. The selection offered depends on the 
 *           outputs boards fitted. If there is only one board fitted, e.g. an 
 *           analog one, then there is nothing to select. Only instrument 
 *           version AFQ- B18 is equipped with a digital output board. Then, 
 *           the user decides whether the analog or the digital board shall be 
 *           switched on. If a second digital output board is installed, the 
 *           user has to select either DIGITAL1 or DIGITAL2. Instead, the user 
 *           might decide to install two analog output boards, then, he has to 
 *           select either ANALOG1 or ANALOG2. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetActiveOutput(
    ViSession   instrSession,
    ViInt32 activeOutput
)
{
    return rsafq_setEnum (instrSession,
                          "OUTP:AOUT",
                          activeOutput,
                          activeOutputArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Active Output
 * Purpose:  This function returns the active output.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetActiveOutput(
    ViSession   instrSession,
    ViInt32*    activeOutput
)
{
    return rsafq_getEnum (instrSession,
                          "OUTP:AOUT?",
                          activeOutput,
                          activeOutputArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Configure Analog Output
 * Purpose:  This function sets all parameters for analog output.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ConfigureAnalogOutput(
    ViSession   instrSession,
    ViBoolean   outputState,
    ViInt32     analogOutput,
    ViInt32     type,
    ViReal64    amplitude,
    ViInt32     amplitudeMode,
    ViReal64    iOffset,
    ViReal64    qOffset,
    ViBoolean   enableBias,
    ViReal64    bias
)
{
    ViStatus    rsafq_status;
    
    rsafq_status = rsafq_SetAnalogOutputType (instrSession, analogOutput, type);
    REMAP (3, 4);
    
    rsafq_status = rsafq_SetAnalogOutputAmplitudeMode (instrSession, analogOutput, amplitudeMode);
    REMAP (3, 6);
    
    switch (type)
    {
        case RSAFQ_OUTPUT_ANALOG_TYPE_BALANCED:
            rsafq_status = rsafq_SetAnalogOutputBalancedAmplitude (instrSession, analogOutput, amplitude);
            REMAP (3, 5);
            
            rsafq_status = rsafq_SetAnalogOutputBalancedIOffset (instrSession, analogOutput, iOffset);
            REMAP (3, 7);
            
            rsafq_status = rsafq_SetAnalogOutputBalancedQOffset (instrSession, analogOutput, qOffset);
            REMAP (3, 8);
            
            rsafq_status = rsafq_SetAnalogOutputBalancedBiasState (instrSession, analogOutput, enableBias);
            if (rsafq_status == VI_ERROR_PARAMETER3)
                return VI_ERROR_INSTR_PARAMETER9;
    
            if (enableBias == VI_TRUE)
            {   
                rsafq_status = rsafq_SetAnalogOutputBalancedBias (instrSession, analogOutput, bias);
                if (rsafq_status == VI_ERROR_PARAMETER3)
                    return VI_ERROR_INSTR_PARAMETER10;
            }
        break;
        
        case RSAFQ_OUTPUT_ANALOG_TYPE_UNBALANCED:
            rsafq_status = rsafq_SetAnalogOutputUnbalancedAmplitude (instrSession, analogOutput, amplitude);
            REMAP (3, 5);
            
            rsafq_status = rsafq_SetAnalogOutputUnbalancedIOffset (instrSession, analogOutput, iOffset);
            REMAP (3, 7);
            
            rsafq_status = rsafq_SetAnalogOutputUnbalancedQOffset (instrSession, analogOutput, qOffset);
            REMAP (3, 8);
        break;
        default:
            return VI_ERROR_PARAMETER4;
    }
    
    CHECKERR (rsafq_SetOutputState (instrSession, outputState));

    RETURN;
}

/*****************************************************************************
 * Function: Set Analog Output Type
 * Purpose:  This function defines the type of the selected analog output. 
 * 			 If unbalanced is selected, the signal is output at the I and Q 
 *			 connectors. If balanced is selected, the signal is output at 
 * 			 the I, I BAR, Q and Q BAR connectors.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetAnalogOutputType(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViInt32 type
)
{
    ViChar  buffer[BUFFER_SIZE];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 2);
    CHECKPARAM (rsafq_invalidViInt32Range (type,0,1),3);
    
    sprintf (buffer,"OUTP:ANAL%ld:TYPE", analogOutput);
    
    return rsafq_setEnum (instrSession,
                          buffer,
                          type,
                          typeArr,
                          VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Analog Output Type
 * Purpose:  This function returns the type of the selected analog output.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetAnalogOutputType(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViInt32*    type
)
{
    ViChar    command[256];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 2);
    //CHECKERR (rsafq_checkInstrument (instrSession, RSAFQ_INSTRUMENT_AFQ100B, VI_TRUE));
    
    sprintf (command,"OUTP:ANAL%ld:TYPE?", analogOutput);
    
    return rsafq_getEnum (instrSession,
                          command,
                          type,
                          typeArr,
                          VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Analog Output Amplitude Mode
 * Purpose:  This function sets the amplitude mode. Two of them are possible, 
 *           Electronic or Attenuators. They determine how precise the signals 
 *           will be adjusted.
 *           
 *           Using the Electronic mode, the instrument generates signal 
 *           amplitudes by scaling the waveforms. The adjustment can be 
 *           performed very fast. However, the level resolution is low if a 
 *           small signal amplitude (<< 1V) is selected.
 *           
 *           Using the Attenuators mode, the instrument generates signal 
 *           amplitudes by applying a gauge circuit to the waveforms. To do so, 
 *           the instrument switches a relay. The adjustement requires some 
 *           time. However, the level resolution is high, especially for small 
 *           signal amplitudes (<< 1V).
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetAnalogOutputAmplitudeMode(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViInt32 amplitudeMode
)
{
    ViChar  buffer[BUFFER_SIZE];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 2);
    CHECKPARAM (rsafq_invalidViInt32Range (amplitudeMode,0,1),3);
    
    sprintf (buffer,"OUTP:ANAL%ld:AMOD", analogOutput);
    
    return rsafq_setEnum (instrSession,
                          buffer,
                          amplitudeMode,
                          amplitudeModeArr,
                          VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Analog Output Amplitude Mode
 * Purpose:  This function returns the amplitude mode.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetAnalogOutputAmplitudeMode(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViInt32*    amplitudeMode
)
{
    ViChar  command[256];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 2);
    
    sprintf (command,"OUTP:ANAL%ld:AMOD?", analogOutput);
    
    return rsafq_getEnum (instrSession,
                          command,
                          amplitudeMode,
                          amplitudeModeArr,
                          VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Analog Output Impedance
 * Purpose:  This function sets the output impedance.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetAnalogOutputImpedance(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViInt32 outputImpedance
)
{
#if 0    
    ViChar  buffer[BUFFER_SIZE];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 2);
    CHECKPARAM (rsafq_invalidViInt32Range (outputImpedance,0,1),3);
    
    sprintf (buffer,"OUTP:ANAL%ld:OIMP", analogOutput);
    
    return rsafq_setEnum (instrSession,
                          buffer,
                          outputImpedance,
                          outputImpedanceArr,
                          VI_ERROR_PARAMETER3);
#endif
    
    return RSAFQ_ERROR_INSTRUMENT_NSUP_FUNC;    
}

/*****************************************************************************
 * Function: Get Analog Output Impedance
 * Purpose:  This function returns the output impedance.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetAnalogOutputImpedance(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViInt32*    outputImpedance
)
{
#if 0
    ViChar  command[256];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 2);
    
    sprintf (command,"OUTP:ANAL%ld:OIMP?", analogOutput);
    
    return rsafq_getEnum (instrSession,
                          command,
                          outputImpedance,
                          outputImpedanceArr,
                          VI_ERROR_PARAMETER3);
#endif
    
    return RSAFQ_ERROR_INSTRUMENT_NSUP_FUNC;    
}

/*****************************************************************************
 * Function: Set Analog Output Equalizer State
 * Purpose:  This function sets the equalizer state.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetAnalogOutputEqualizerState(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViBoolean   equalizerState
)
{
    ViChar  buffer[BUFFER_SIZE];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 2);
    CHECKPARAM (rsafq_invalidViBooleanRange (equalizerState), 3);
    
    sprintf (buffer,"OUTP:ANAL%ld:EQU:STAT", analogOutput);
    
    return rsafq_setViBoolean (instrSession,
                               buffer,
                               equalizerState,
                               VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Analog Output Equalizer State
 * Purpose:  This function returns the equalizer state.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetAnalogOutputEqualizerState(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViBoolean*  equalizerState
)
{
    ViChar  command[256];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 2);
    
    sprintf (command,"OUTP:ANAL%ld:EQU:STAT?", analogOutput);
    
    return rsafq_getViBoolean (instrSession,
                               command,
                               equalizerState,
                               VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Analog Output Balanced Amplitude
 * Purpose:  This function sets the output voltage for the balanced signal. 
 *           Small differences at the inputs can be taken into account via the 
 *           impairment settings offset and bias.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetAnalogOutputBalancedAmplitude(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViReal64    amplitude
)
{
    ViChar  buffer[BUFFER_SIZE];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 2);
    
    sprintf (buffer,"OUTP:ANAL%ld:BAL:AMPL", analogOutput);
    
    return rsafq_setViReal64 (instrSession,
                              buffer,
                              amplitude,
                              0.0,1.5,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Analog Output Balanced Amplitude
 * Purpose:  This function returns the output voltage for the balanced signal.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetAnalogOutputBalancedAmplitude(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViReal64*   amplitude
)
{
    ViChar  command[256];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 2);
    
    sprintf (command,"OUTP:ANAL%ld:BAL:AMPL?", analogOutput);
    
    return rsafq_getViReal64 (instrSession,
                              command,
                              amplitude,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Analog Output Unbalanced Amplitude
 * Purpose:  This function sets the output voltage for the unbalanced signal. 
 *           Small differences at the inputs can be taken into account via 
 *           impairment settings.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetAnalogOutputUnbalancedAmplitude(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViReal64    amplitude
)
{
    ViChar  buffer[BUFFER_SIZE];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 2);
    
    sprintf (buffer,"OUTP:ANAL%ld:UNB:AMPL", analogOutput);
    
    return rsafq_setViReal64 (instrSession,
                              buffer,
                              amplitude,
                              0.0,0.75,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Analog Output Unbalanced Amplitude
 * Purpose:  This function returns the output voltage for the unbalanced 
 *           signal.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetAnalogOutputUnbalancedAmplitude(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViReal64*   amplitude
)
{
    ViChar  buffer[BUFFER_SIZE];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 2);

    sprintf (buffer,"OUTP:ANAL%ld:UNB:AMPL?", analogOutput);
    
    return rsafq_getViReal64 (instrSession,
                              buffer,
                              amplitude,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Analog Output Balanced I Offset
 * Purpose:  This function sets an offset between the inverting and the 
 *           non-inverting output.
 *           
 *           The set value is set half in the positive and half in the 
 *           negative direction:
 *           
 *           A positive offset is set with half of the value in positive 
 *           direction at the non-inverting outputs, in negative direction at 
 *           the inverting outputs, respectively.
 *           
 *           For example, if a 10.0 mV offset value is set for the I signal 
 *           component, the I signal will have a 5.0 mV offset and the output 
 *           signal at`I will have a - 5.0 mV offset.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetAnalogOutputBalancedIOffset(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViReal64    iOffset
)
{
    ViChar  buffer[BUFFER_SIZE];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 2);
    
    sprintf (buffer,"OUTP:ANAL%ld:BAL:OFFS:I", analogOutput);
    
    return rsafq_setViReal64 (instrSession,
                              buffer,
                              iOffset,
                              -10.0,10.0,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Analog Output Balanced I Offset
 * Purpose:  This function returns an offset between the inverting and the 
 *           non-inverting output.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetAnalogOutputBalancedIOffset(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViReal64*   iOffset
)
{
    ViChar  command[256];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 3);
    
    sprintf (command,"OUTP:ANAL%ld:BAL:OFFS:I?", analogOutput);
    
    return rsafq_getViReal64 (instrSession,
                              command,
                              iOffset,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Analog Output Balanced Q Offset
 * Purpose:  This function sets an offset between the inverting and the 
 *           non-inverting output. 
 *           
 *           The set value is set half in the positive and half in the 
 *           negative direction:
 *           
 *           A positive offset is set with half of the value in positive 
 *           direction at the non-inverting outputs, in negative direction at 
 *           the inverting outputs, respectively.
 *           
 *           For example, if a 10.0 mV offset value is set for the Q signal 
 *           component, the I signal will have a 5.0 mV offset and the output 
 *           signal at`Q will have a - 5.0 mV offset.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetAnalogOutputBalancedQOffset(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViReal64    qOffset
)
{
    ViChar  buffer[BUFFER_SIZE];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 2);
    CHECKPARAM (rsafq_invalidViReal64Range (qOffset,-10.0,10.0), 3);
    
    sprintf (buffer,"OUTP:ANAL%ld:BAL:OFFS:Q", analogOutput);
    
    return rsafq_setViReal64 (instrSession,
                              buffer,
                              qOffset,
                              -10.0,10.0,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Analog Output Balanced Q Offset
 * Purpose:  This function returns an offset between the inverting and the 
 *           non-inverting output.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetAnalogOutputBalancedQOffset(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViReal64*   qOffset
)
{
    ViChar  command[256];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 3);
    
    sprintf (command,"OUTP:ANAL%ld:BAL:OFFS:Q?", analogOutput);
    
    return rsafq_getViReal64 (instrSession,
                              command,
                              qOffset,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Analog Output Unbalanced I Offset
 * Purpose:  This function sets the DC offset of the I path.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetAnalogOutputUnbalancedIOffset(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViReal64    iOffset
)
{
    ViChar  buffer[BUFFER_SIZE];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 2);
    CHECKPARAM (rsafq_invalidViReal64Range (iOffset,-10.0,10.0), 3);
    
    sprintf (buffer,"OUTP:ANAL%ld:UNB:OFFS:I", analogOutput);
    
    return rsafq_setViReal64 (instrSession,
                              buffer,
                              iOffset,
                              -10.0,10.0,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Analog Output Unbalanced I Offset
 * Purpose:  This function returns the DC offset of the I path.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetAnalogOutputUnbalancedIOffset(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViReal64*   iOffset
)
{
    ViChar  command[256];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 2);
    
    sprintf (command,"OUTP:ANAL%ld:UNB:OFFS:I?", analogOutput);
    
    return rsafq_getViReal64 (instrSession,
                              command,
                              iOffset,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Analog Output Unbalanced Q Offset
 * Purpose:  This function sets the DC offset of the Q path.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetAnalogOutputUnbalancedQOffset(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViReal64    qOffset
)
{
    ViChar  buffer[BUFFER_SIZE];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 2);
    CHECKPARAM (rsafq_invalidViReal64Range (qOffset,-10.0,10.0), 3);
    
    sprintf (buffer,"OUTP:ANAL%ld:UNB:OFFS:Q", analogOutput);
    
    return rsafq_setViReal64 (instrSession,
                              buffer,
                              qOffset,
                              -10.0,10.0,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Analog Output Unbalanced Q Offset
 * Purpose:  This function returns the DC offset of the Q path.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetAnalogOutputUnbalancedQOffset(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViReal64*   qOffset
)
{
    ViChar  command[256];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 2);
    
    sprintf (command,"OUTP:ANAL%ld:UNB:OFFS:Q?", analogOutput);
    
    return rsafq_getViReal64 (instrSession,
                              command,
                              qOffset,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Analog Output Balanced Bias State
 * Purpose:  This function enables/disables bias amplifier.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetAnalogOutputBalancedBiasState(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViBoolean   enableBias
)
{
    ViChar  buffer[BUFFER_SIZE];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 2);
    CHECKPARAM (rsafq_invalidViInt32Range (enableBias,0,1), 3);
    
    sprintf (buffer,"OUTP:ANAL%ld:BAL:BIAS:STAT", analogOutput);
    
    return rsafq_setViBoolean (instrSession,
                               buffer,
                               enableBias,
                               VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Analog Output Balanced Bias State
 * Purpose:  This function returns the bias state.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetAnalogOutputBalancedBiasState(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViBoolean*  enableBias
)
{
    ViChar  command[256];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 2);
    
    sprintf (command,"OUTP:ANAL%ld:BAL:BIAS:STAT?", analogOutput);
    
    return rsafq_getViBoolean (instrSession,
                               command,
                               enableBias,
                               VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Analog Output Balanced Bias
 * Purpose:  This function sets the bias. A DC voltage is superimposed upon 
 *           the symmetric I and Q signal.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetAnalogOutputBalancedBias(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViReal64    bias
)
{
    ViChar  buffer[BUFFER_SIZE];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 2);
    CHECKPARAM (rsafq_invalidViReal64Range (bias,-2.5,2.5), 3);
    
    sprintf (buffer,"OUTP:ANAL%ld:BAL:BIAS", analogOutput);
    
    return rsafq_setViReal64 (instrSession,
                              buffer,
                              bias,
                              -2.5,2.5,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Analog Output Balanced Bias
 * Purpose:  This function returns the bias.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetAnalogOutputBalancedBias(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViReal64*   bias
)
{
    ViChar  command[256];
    
    CHECKPARAM (rsafq_invalidViInt32Range (analogOutput,1,2), 2);
    
    sprintf (command,"OUTP:ANAL%ld:BAL:BIAS?", analogOutput);
    
    return rsafq_getViReal64 (instrSession,
                              command,
                              bias,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Configure Digital Output
 * Purpose:  This function sets all parameters for digital output.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ConfigureDigitalOutput (ViSession   instrSession,
                                                ViBoolean   outputState,
                                                ViInt32     digitalOutput,
                                                ViInt32     port, 
                                                ViInt32     resolution,
                                                ViReal64    amplitude,
                                                ViReal64    iOffset,
                                                ViReal64    qOffset, 
                                                ViReal64    rate,
                                                ViInt32     source,
                                                ViInt32     i0_maps_to,
                                                ViInt32     i1_maps_to,
                                                ViInt32     q0_maps_to,
                                                ViInt32     q1_maps_to
)
{
    ViStatus    rsafq_status;
	rate;
	source;
    
    rsafq_status = rsafq_SetDigitalOutputInterfacePort (instrSession, digitalOutput, port);
    REMAP(3, 4);
    
    rsafq_status = rsafq_SetDigitalOutputSignalResolution (instrSession, digitalOutput, resolution);
    REMAP (3, 5);

    rsafq_status = rsafq_SetDigitalOutputSignalAmplitude (instrSession, digitalOutput, amplitude);
    REMAP (3, 6);

    rsafq_status = rsafq_SetDigitalOutputSignalIOffset (instrSession, digitalOutput, iOffset);
    REMAP (3, 7);

    rsafq_status = rsafq_SetDigitalOutputSignalQOffset (instrSession, digitalOutput, qOffset);
    REMAP (3, 8);
    
    switch (port)
    {   
        case RSAFQ_OUTPUT_DIGITAL_PORT1:
        
        break;
        
        case RSAFQ_OUTPUT_DIGITAL_PORT2:
#if 0
            rsafq_status = rsafq_SetDigitalOutputClockRate (instrSession, digitalOutput, rate);
            if (rsafq_status == VI_ERROR_PARAMETER3)
                return VI_ERROR_INSTR_PARAMETER9;
    
            rsafq_status = rsafq_SetDigitalOutputClockSource (instrSession, digitalOutput, source);
            if (rsafq_status == VI_ERROR_PARAMETER3)
                return VI_ERROR_INSTR_PARAMETER10;
#endif          
            rsafq_status = rsafq_SetDigitalOutputPinI0Assignment (instrSession, digitalOutput, i0_maps_to);
            if (rsafq_status == VI_ERROR_PARAMETER3)
                return VI_ERROR_INSTR_PARAMETER11;
        
            rsafq_status = rsafq_SetDigitalOutputPinI1Assignment (instrSession, digitalOutput, i1_maps_to);
            if (rsafq_status == VI_ERROR_PARAMETER3)
                return VI_ERROR_INSTR_PARAMETER12;
            
            rsafq_status = rsafq_SetDigitalOutputPinQ0Assignment (instrSession, digitalOutput, q0_maps_to);
            if (rsafq_status == VI_ERROR_PARAMETER3)
                return VI_ERROR_INSTR_PARAMETER13;
            
            rsafq_status = rsafq_SetDigitalOutputPinQ1Assignment (instrSession, digitalOutput, q1_maps_to);
            if (rsafq_status == VI_ERROR_PARAMETER3)
                return VI_ERROR_INSTR_PARAMETER14;
        break;
        
        default:
            return VI_ERROR_PARAMETER4;
    }
        
    CHECKERR (rsafq_SetOutputState (instrSession, outputState));
    
    RETURN;
}

/*****************************************************************************
 * Function: Set Digital Output Clock Rate
 * Purpose:  This function sets the output clock frequency.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetDigitalOutputClockRate(
    ViSession   instrSession,
    ViInt32 digitalOutput,
    ViReal64    rate
)
{
#if 0
    ViChar  buffer[BUFFER_SIZE];
    INIT;
    
    CHECKOPT (B18);
    
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    CHECKPARAM (rsafq_invalidViReal64Range (rate,1e3,400e6), 3);
    
    sprintf (buffer,"OUTP:DIG%ld:CLOCK", digitalOutput);
    
    return rsafq_setViReal64 (instrSession,
                              buffer,
                              rate,
                              1e3,400e6,
                              VI_ERROR_PARAMETER3);
#endif
    
    return RSAFQ_ERROR_INSTRUMENT_NSUP_FUNC;
}

/*****************************************************************************
 * Function: Get Digital Output Clock Rate
 * Purpose:  This function returns the output clock frequency.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetDigitalOutputClockRate(
    ViSession   instrSession,
    ViInt32 digitalOutput,
    ViReal64*   rate
)
{
#if 0
    ViChar  command[256];
    INIT;
    
    CHECKOPT (B18);
    
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    
    sprintf (command,"OUTP:DIG%ld:CLOCK?", digitalOutput);
    
    return rsafq_getViReal64 (instrSession,
                              command,
                              rate,
                              VI_ERROR_PARAMETER3);
#endif
    
    return RSAFQ_ERROR_INSTRUMENT_NSUP_FUNC;
}

/*****************************************************************************
 * Function: Set Digital Output Clock Source
 * Purpose:  This function selects the output clock source.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetDigitalOutputClockSource(
    ViSession   instrSession,
    ViInt32 digitalOutput,
    ViInt32 source
)
{
#if 0
    ViChar  buffer[BUFFER_SIZE];
    INIT;
    
    CHECKOPT (B18);
    
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    CHECKPARAM (rsafq_invalidViInt32Range (source,0,1), 3);
    
    sprintf (buffer,"OUTP:DIG%ld:CLOC:SOUR", digitalOutput);
    
    return rsafq_setEnum (instrSession,
                          buffer,
                          source,
                          sourceArr,
                          VI_ERROR_PARAMETER3);
#endif
    
    return RSAFQ_ERROR_INSTRUMENT_NSUP_FUNC;
}

/*****************************************************************************
 * Function: Get Digital Output Clock Source
 * Purpose:  This function returns the output clock source.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetDigitalOutputClockSource(
    ViSession   instrSession,
    ViInt32 digitalOutput,
    ViInt32*    source
)
{
#if 0
    ViChar  command[256];
    INIT;
    
    CHECKOPT (B18);
    
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    
    sprintf (command,"OUTP:DIG%ld:CLOC:SOUR?", digitalOutput);
    
    return rsafq_getEnum (instrSession,
                          command,
                          source,
                          sourceArr,
                          VI_ERROR_PARAMETER3);
#endif
    return RSAFQ_ERROR_INSTRUMENT_NSUP_FUNC;
}

/*****************************************************************************
 * Function: Set Digital Output Clock Slope
 * Purpose:  This function sets the polarity of the active clock for 
 *           triggering digital signal transmission via Port2. Clock and data 
 *           are in phase. Either the rising or the falling edge of the data 
 *           clock can be selected.
 *           
 *           Note(s):
 *           (1) This function is available on AFQ100B only.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetDigitalOutputClockSlope (
	ViSession 	instrSession,
    ViInt32 	digitalOutput,
	ViInt32 	slope
)
{
    ViChar      command[BUFFER_SIZE];
	INIT;
	CHECKVER(AFQ100B);
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    
    sprintf (command,"SOUR:OUTP:DIG%ld:CLOC:SLOP ", digitalOutput);

	return rsafq_setEnum (instrSession,
                          command,
                          slope,
                          inputSlopeArr,
                          VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Digital Output Clock Slope 
 * Purpose:  This function queries the polarity of the active clock for 
 *           triggering digital signal transmission via Port2.
 *           
 *           Note(s):
 *           
 *           (1) This function is available on AFQ100B only.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetDigitalOutputClockSlope (ViSession instrSession,
                                                    ViInt32 digitalOutput,
                                                    ViInt32 *slope)
{
    ViChar      command[BUFFER_SIZE];
	INIT;
	CHECKVER(AFQ100B);
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    
    sprintf (command,"SOUR:OUTP:DIG%ld:CLOC:SLOP?", digitalOutput);

	return rsafq_getEnum (instrSession,
                          command,
                          slope,
                          inputSlopeArr,
                          VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Digital Output Signal Resolution
 * Purpose:  This function sets the signal's digital resolution but does not 
 *           influence analog signals.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetDigitalOutputSignalResolution(
    ViSession   instrSession,
    ViInt32 digitalOutput,
    ViInt32 resolution
)
{
    ViChar  command[256];
    INIT;
    
    CHECKOPT (B18);
    
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    
    sprintf (command,"OUTP:DIG%ld:RES", digitalOutput);
    
    return rsafq_setEnum (instrSession,
                          command,
                          resolution,
                          resolutionArr,
                          VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Digital Output Signal Resolution
 * Purpose:  This function returns the signal's digital resolution.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetDigitalOutputSignalResolution(
    ViSession   instrSession,
    ViInt32 digitalOutput,
    ViInt32*    resolution
)
{
    ViChar  command[256];
    INIT;
    
    CHECKOPT (B18);
    
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    
    sprintf (command,"OUTP:DIG%ld:RES?", digitalOutput);

    return rsafq_getEnum (instrSession,
                          command,
                          resolution,
                          resolutionArr,
                          VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Digital Output Signal Amplitude
 * Purpose:  This function sets the amplitude for a fully modulated input 
 *           signal.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetDigitalOutputSignalAmplitude(
    ViSession   instrSession,
    ViInt32 digitalOutput,
    ViReal64    amplitude
)
{
    ViChar  buffer[BUFFER_SIZE];
    INIT;
    
    CHECKOPT (B18);
    
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);

    sprintf (buffer,"OUTP:DIG%ld:AMPL", digitalOutput);
    
    return rsafq_setViReal64 (instrSession,
                              buffer,
                              amplitude,
                              0.0,1.0,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Digital Output Signal Amplitude
 * Purpose:  This function returns the amplitude for a fully modulated input 
 *           signal.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetDigitalOutputSignalAmplitude(
    ViSession   instrSession,
    ViInt32 digitalOutput,
    ViReal64*   amplitude
)
{
    ViChar  command[256];
    INIT;
    
    CHECKOPT (B18);
    
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    
    sprintf (command,"OUTP:DIG%ld:AMPL?", digitalOutput);
    
    return rsafq_getViReal64 (instrSession,
                              command,
                              amplitude,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Digital Output Signal I Offset
 * Purpose:  This function sets the offset of the digital signal.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetDigitalOutputSignalIOffset(
    ViSession   instrSession,
    ViInt32 digitalOutput,
    ViReal64    iOffset
)
{
    ViChar  buffer[BUFFER_SIZE];
    INIT;
    
    CHECKOPT (B18);
    
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    CHECKPARAM (rsafq_invalidViReal64Range (iOffset,-0.1,0.1), 3);
    
    sprintf (buffer,"OUTP:DIG%ld:OFFS:I", digitalOutput);
    
    return rsafq_setViReal64 (instrSession,
                              buffer,
                              iOffset,
                              -0.1,0.1,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Digital Output Signal I Offset
 * Purpose:  This function returns the offset of the digital signal.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetDigitalOutputSignalIOffset(
    ViSession   instrSession,
    ViInt32 digitalOutput,
    ViReal64*   iOffset
)
{
    ViChar  command[256];
    INIT;
    
    CHECKOPT (B18);
    
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    
    sprintf (command,"OUTP:DIG%ld:OFFS:I?", digitalOutput);
    
    return rsafq_getViReal64 (instrSession,
                              command,
                              iOffset,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Digital Output Signal Q Offset
 * Purpose:  This function sets the offset of the digital signal.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetDigitalOutputSignalQOffset(
    ViSession   instrSession,
    ViInt32 digitalOutput,
    ViReal64    qOffset
)
{
    ViChar  buffer[BUFFER_SIZE];
    INIT;
    
    CHECKOPT (B18);
    
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    CHECKPARAM (rsafq_invalidViReal64Range (qOffset,-0.1,0.1), 3);
    
    sprintf (buffer,"OUTP:DIG%ld:OFFS:Q", digitalOutput);
    
    return rsafq_setViReal64 (instrSession,
                              buffer,
                              qOffset,
                              -0.1,0.1,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Digital Output Signal Q Offset
 * Purpose:  This function returns the offset of the digital signal.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetDigitalOutputSignalQOffset(
    ViSession   instrSession,
    ViInt32 digitalOutput,
    ViReal64*   qOffset
)
{
    ViChar  command[256];
    INIT;
    
    CHECKOPT (B18);
    
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    
    sprintf (command,"OUTP:DIG%ld:OFFS:Q?", digitalOutput);
    
    return rsafq_getViReal64 (instrSession,
                              command,
                              qOffset,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Digital Output Interface Port
 * Purpose:  This function selects a digital output port.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetDigitalOutputInterfacePort(
    ViSession   instrSession,
    ViInt32 digitalOutput,
    ViInt32 port
)
{
    ViChar  buffer[BUFFER_SIZE];
    INIT;
    
    CHECKOPT (B18);
    
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    CHECKPARAM (rsafq_invalidViInt32Range (port,0,1), 3);
    
    sprintf (buffer,"OUTP:DIG%ld:PORT", digitalOutput);
    
    return rsafq_setEnum (instrSession,
                          buffer,
                          port,
                          portArr,
                          VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Digital Output Interface Port
 * Purpose:  This function returns a digital output port.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetDigitalOutputInterfacePort(
    ViSession   instrSession,
    ViInt32 digitalOutput,
    ViInt32*    port
)
{
    ViChar  command[256];
    INIT;
    
    CHECKOPT (B18);
    
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    
    sprintf (command,"OUTP:DIG%ld:PORT?", digitalOutput);
    
    return rsafq_getEnum (instrSession,
                          command,
                          port,
                          portArr,
                          VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Digital Output Pin I0 Assignment
 * Purpose:  This function maps output pin I[0] ( = LSB) to either signal or 
 *           marker 4.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetDigitalOutputPinI0Assignment(
    ViSession   instrSession,
    ViInt32 digitalOutput,
    ViInt32 i_0mapsto
)
{
    ViChar  buffer[BUFFER_SIZE];
    INIT;
    
    CHECKOPT (B18);
    
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    CHECKPARAM (rsafq_invalidViInt32Range (i_0mapsto,0,1), 3);
    
    sprintf (buffer,"OUTP:DIG%ld:PASS:I0", digitalOutput);
    
    return rsafq_setEnum (instrSession,
                          buffer,
                          i_0mapsto,
                          iMapstoArr,
                          VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Digital Output Pin I0 Assignment
 * Purpose:  This function returns mapping of output pin I[0] ( = LSB).
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetDigitalOutputPinI0Assignment(
    ViSession   instrSession,
    ViInt32 digitalOutput,
    ViInt32*    i_0mapsto
)
{
    ViChar  command[256];
    INIT;
    
    CHECKOPT (B18);
    
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    
    sprintf (command,"OUTP:DIG%ld:PASS:I0?", digitalOutput);
    
    return rsafq_getEnum (instrSession,
                          command,
                          i_0mapsto,
                          iMapstoArr,
                          VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Digital Output Pin I1 Assignment
 * Purpose:  This function maps output pin I[1] ( = LSB +1) to either signal 
 *           or marker 3.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetDigitalOutputPinI1Assignment(
    ViSession   instrSession,
    ViInt32 digitalOutput,
    ViInt32 i_1mapsto
)
{
    ViChar  buffer[BUFFER_SIZE];
    INIT;
    
    CHECKOPT (B18);
    
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    CHECKPARAM (rsafq_invalidViInt32Range (i_1mapsto,0,1), 3);
    
    sprintf (buffer,"OUTP:DIG%ld:PASS:I1", digitalOutput);
    
    return rsafq_setEnum (instrSession,
                          buffer,
                          i_1mapsto,
                          iMapstoArr,
                          VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Digital Output Pin I1 Assignment
 * Purpose:  This function returns mapping of output pin I[1] ( = LSB +1).
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetDigitalOutputPinI1Assignment(
    ViSession   instrSession,
    ViInt32 digitalOutput,
    ViInt32*    i_1mapsto
)
{
    ViChar  command[256];
    INIT;
    
    CHECKOPT (B18);
    
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    
    sprintf (command,"OUTP:DIG%ld:PASS:I1?", digitalOutput);
    
    return rsafq_getEnum (instrSession,
                          command,
                          i_1mapsto,
                          iMapstoArr,
                          VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Digital Output Pin Q0 Assignment
 * Purpose:  This function maps output pin Q[0] ( = LSB) to either signal or 
 *           marker 2.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetDigitalOutputPinQ0Assignment(
    ViSession   instrSession,
    ViInt32 digitalOutput,
    ViInt32 q_0mapsto
)
{
    ViChar  buffer[BUFFER_SIZE];
    INIT;
    
    CHECKOPT (B18);
    
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    CHECKPARAM (rsafq_invalidViInt32Range (q_0mapsto,0,1), 3);
    
    sprintf (buffer,"OUTP:DIG%ld:PASS:Q0", digitalOutput);
    
    return rsafq_setEnum (instrSession,
                          buffer,
                          q_0mapsto,
                          iMapstoArr,
                          VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Digital Output Pin Q0 Assignment
 * Purpose:  This function returns mapping of output pin Q[0] ( = LSB).
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetDigitalOutputPinQ0Assignment(
    ViSession   instrSession,
    ViInt32 digitalOutput,
    ViInt32*    q_0mapsto
)
{
    ViChar  command[256];
    INIT;
    
    CHECKOPT (B18);
    
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    
    sprintf (command,"OUTP:DIG%ld:PASS:Q0?", digitalOutput);
    
    return rsafq_getEnum (instrSession,
                          command,
                          q_0mapsto,
                          iMapstoArr,
                          VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set Digital Output Pin Q1 Assignment
 * Purpose:  This function maps output pin Q[1] ( = LSB +1) to either signal 
 *           or marker 1.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetDigitalOutputPinQ1Assignment(
    ViSession   instrSession,
    ViInt32 digitalOutput,
    ViInt32 q_1mapsto
)
{
    ViChar  command[256];
    INIT;
    
    CHECKOPT (B18);
    
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    
    sprintf (command,"OUTP:DIG%ld:PASS:Q1", digitalOutput);
    
    return rsafq_setEnum (instrSession,
                          command,
                          q_1mapsto,
                          iMapstoArr,
                          VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Get Digital Output Pin Q1 Assignment
 * Purpose:  This function returns mapping of output pin Q[1] ( = LSB +1).
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetDigitalOutputPinQ1Assignment(
    ViSession   instrSession,
    ViInt32 digitalOutput,
    ViInt32*    q_1mapsto
)
{
    ViChar  command[256];
    INIT;
    
    CHECKOPT (B18);
    
    CHECKPARAM (rsafq_invalidViInt32Range (digitalOutput,1,2), 2);
    
    sprintf (command,"OUTP:DIG%ld:PASS:Q1?", digitalOutput);
    
    return rsafq_getEnum (instrSession,
                          command,
                          q_1mapsto,
                          iMapstoArr,
                          VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set BERT State
 * Purpose:  This function switches the measurement on/off. Depending on the 
 *           selected mode, either a continuous measurement (BERT:SEQ AUTO) or 
 *           a single measurement (BERT:SEQ SING) is carried out. A single 
 *           measurement must be triggered (:TRIG:BERT).
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetBERTState(
    ViSession   instrSession,
    ViBoolean   bertState
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_setViBoolean (instrSession,
                               "BERT:STAT",
                               bertState,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get BERT State
 * Purpose:  This function returns whether is the BERT measurement switched on 
 *           or off.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetBERTState(
    ViSession   instrSession,
    ViBoolean*  bertState
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_getViBoolean (instrSession,
                               "BERT:STAT?",
                               bertState,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: BERT Start Measurement
 * Purpose:  This function starts a continuous measurement.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_BERTStartMeasurement(
    ViSession   instrSession
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_sendCmd (instrSession, "BERT:STAR");
}

/*****************************************************************************
 * Function: BERT Stop Measurement
 * Purpose:  This function stops an ongoing measurement.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_BERTStopMeasurement(
    ViSession   instrSession
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_sendCmd (instrSession,
                          "BERT:STOP");
}

/*****************************************************************************
 * Function: BERT Measurement Resuts
 * Purpose:  This function queries the result of the last BER measurement. The 
 *           response consists of seven results separated by commas. In the 
 *           first measurement following the start, intermediate results for 
 *           the number of data bits/data blocks, error bits/errors and error 
 *           rate are also queried. In the following measurements (only for 
 *           BERT:SEQ AUTO) only the final results of each single measurement 
 *           are queried.
 *           
 *           
 *           Note:
 *           
 *           At the end of a measurement, the restart of a new one is delayed 
 *           until the first measurement result has been queried with 
 *           BERT:RES?. The resulting brief measurement interruption is 
 *           irrelevant because the subsequent measurement will be synchronized 
 *           within 24 data bits.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_BERTMeasurementResuts(
    ViSession   instrSession,
    ViInt32*    dataBits,
    ViInt32*    errors,
    ViReal64*   bitErrorRate,
    ViInt32*    terminationStatus,
    ViInt32*    clockLineStatus,
    ViInt32*    dataLineStatus,
    ViInt32*    synchronizationStatus
)
{
    ViReal64    values[10];

    INIT;
    
    CHECKOPT (K80);
    
    CHECKERR (rsafq_getViReal64Array (instrSession,
                                      "BERT:RES?",
                                      values,
                                      7,
                                      VI_ERROR_INSTR_INTERPRETING_RESPONSE));

    /* Measured results */
    if (dataBits) *dataBits                             = (ViInt32)values[0];
    if (errors) *errors                                 = (ViInt32)values[1];
    if (bitErrorRate) *bitErrorRate                     = values[2];
    if (terminationStatus) *terminationStatus           = (ViInt32)values[3];
    if (clockLineStatus) *clockLineStatus               = (ViInt32)values[4];
    if (dataLineStatus) *dataLineStatus                 = (ViInt32)values[5];
    if (synchronizationStatus) *synchronizationStatus   = (ViInt32) values[6];

    RETURN;
}

/*****************************************************************************
 * Function: Configure BERT
 * Purpose:  This function is used to select the type of measurement the data 
 *           source and the termination criteria of the BERT measurement. If 
 *           the data is not cyclically continued, the measurement can be 
 *           selectively interrupted and restarted. Certain data sections and 
 *           frames that have been marked as faulty can be excluded from the 
 *           measurement.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ConfigureBERT(
    ViSession   instrSession,
    ViInt32 mode,
    ViInt32 maxDataBits,
    ViInt32 maxErrors,
    ViInt32 prbs,
    ViInt32 dataPolarity,
    ViInt32 clockSlope,
    ViBoolean   externalRestart,
    ViInt32 dataEnable,
    ViInt32 patternIgnore,
    ViReal64    threshold,
    ViInt32 impedance
)
{
    ViStatus    rsafq_status;
    
    CHECKERR (rsafq_SetBERTMode (instrSession, mode));

    rsafq_status = rsafq_SetBERTMaxDataBits (instrSession, maxDataBits);
    REMAP (2, 3);
    
    rsafq_status = rsafq_SetBERTMaxErrors (instrSession, maxErrors);
    REMAP (2, 4);
    
    rsafq_status = rsafq_SetBERTPRBS (instrSession, prbs);
    REMAP (2, 5);
    
    rsafq_status = rsafq_SetBERTDataPolarity (instrSession, dataPolarity);
    REMAP (2, 6);
    
    rsafq_status = rsafq_SetBERTClockSlope (instrSession, clockSlope);
    REMAP (2, 7);
    
    rsafq_status = rsafq_SetBERTExternalRestart (instrSession, externalRestart);
    REMAP (2, 8);
    
    rsafq_status = rsafq_SetBERTDataEnable (instrSession, dataEnable);
    if (rsafq_status == VI_ERROR_PARAMETER2)
        return VI_ERROR_INSTR_PARAMETER9;
    
    rsafq_status = rsafq_SetBERTPatternIgnore (instrSession, patternIgnore);
    if (rsafq_status == VI_ERROR_PARAMETER2)
        return VI_ERROR_INSTR_PARAMETER10;
    
    rsafq_status = rsafq_SetBERTThreshold (instrSession, threshold);
    if (rsafq_status == VI_ERROR_PARAMETER2)
        return VI_ERROR_INSTR_PARAMETER11;
        
    rsafq_status = rsafq_SetBERTImpedance (instrSession, impedance);
    if (rsafq_status == VI_ERROR_PARAMETER2)
        return VI_ERROR_INSTR_PARAMETER12;    
    
    RETURN;
}

/*****************************************************************************
 * Function: Set BERT Mode
 * Purpose:  This function selects the type of measurement: single or 
 *           continuous measurement.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetBERTMode(
    ViSession   instrSession,
    ViInt32 mode
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_setEnum  (instrSession,
                           "BERT:SEQ",
                           mode,
                           bertModeArr,
                           VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get BERT Mode
 * Purpose:  This function returns the type of measurement: single or 
 *           continuous measurement.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetBERTMode(
    ViSession   instrSession,
    ViInt32*    mode
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_getEnum  (instrSession,
                           "BERT:SEQ?",
                           mode,
                           bertModeArr,
                           VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: BERT Execute Single
 * Purpose:  This function triggers a BERT measurement.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_BERTExecuteSingle(
    ViSession   instrSession
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_sendCmd (instrSession, "TRIG:BERT");
}

/*****************************************************************************
 * Function: Set BERT Max Data Bits
 * Purpose:  This function enters the number of transmitted data bits/data 
 *           blocks to be checked before the measurement is terminated. With a 
 *           BER measurement, data that was suppressed by BERT:SETup:DENable is 
 *           not counted. This termination criterion always terminates the 
 *           measurement after the specified number of data bits/data blocks. 
 *           Starting from this point, the fourth value is output with 1 (= 
 *           terminate measurement) if the result is queried with :BERT:RES?.  
 *           If the continuous measurement mode (BERT:SEQ AUTO) has been 
 *           selected, the measurement is restarted once the results have been 
 *           queried.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetBERTMaxDataBits(
    ViSession   instrSession,
    ViInt32 maxDataBits
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_setViInt32 (instrSession,
                             "BERT:SET:MCO",
                             maxDataBits,
                             0,4294967295,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get BERT Max Data Bits
 * Purpose:  This function returns the number of transmitted data bits/data 
 *           blocks to be checked before the measurement is terminated. With a 
 *           BER measurement, data that was suppressed by BERT:SETup:DENable is 
 *           not counted. This termination criterion always terminates the 
 *           measurement after the specified number of data bits/data blocks. 
 *           Starting from this point, the fourth value is output with 1 (= 
 *           terminate measurement) if the result is queried with :BERT:RES?.  
 *           If the continuous measurement mode (BERT:SEQ AUTO) has been 
 *           selected, the measurement is restarted once the results have been 
 *           queried.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetBERTMaxDataBits(
    ViSession   instrSession,
    ViInt32*    maxDataBits
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_getViInt32 (instrSession,
                             "BERT:SET:MCO?",
                             maxDataBits,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set BERT Max Errors
 * Purpose:  This function enters the number of errors to occur before the 
 *           measurement is terminated. This termination criterion always 
 *           terminates the measurement after the specified number of errors. 
 *           Starting from this point, the fourth value is output with 1 (= 
 *           terminate measurement) if the measurement result is queried with 
 *           :BERT:RES?. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetBERTMaxErrors(
    ViSession   instrSession,
    ViInt32 maxErrors
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_setViInt32 (instrSession,
                             "BERT:SET:MERR",
                             maxErrors,
                             0,4294967295,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get BERT Max Errors
 * Purpose:  This function returns the number of errors to occur before the 
 *           measurement is terminated. This termination criterion always 
 *           terminates the measurement after the specified number of errors. 
 *           Starting from this point, the fourth value is output with 1 (= 
 *           terminate measurement) if the measurement result is queried with 
 *           :BERT:RES?.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetBERTMaxErrors(
    ViSession   instrSession,
    ViInt32*    maxErrors
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_getViInt32 (instrSession,
                             "BERT:SET:MERR?",
                             maxErrors,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set BERT PRBS
 * Purpose:  This function selects the PRBS sequence. The data generated by 
 *           the PRBS generator is used as a reference for the measurement.
 *           
 *           Required Option(s):
 *           R&S SMU-K80
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetBERTPRBS(
    ViSession   instrSession,
    ViInt32 prbs
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_setEnum (instrSession,
                          "BERT:SET:TYPE",
                          prbs,
                          prbsArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get BERT PRBS
 * Purpose:  This function returns the PRBS sequence. The data generated by 
 *           the PRBS generator is used as a reference for the measurement.
 *           
 *           Required Option(s):
 *           R&S SMU-K80
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetBERTPRBS(
    ViSession   instrSession,
    ViInt32*    prbs
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_getEnum (instrSession,
                          "BERT:SET:TYPE?",
                          prbs,
                          prbsArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set BERT Clock Slope
 * Purpose:  This function sets the polarity of the active slope of the 
 *           feedback clock.
 *           
 *           Required Option(s):
 *           R&S SMU-K80
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetBERTClockSlope(
    ViSession   instrSession,
    ViInt32 clockSlope
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_setEnum (instrSession,
                          "BERT:SET:CLOC",
                          clockSlope,
                          bertClockSlopeArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get BERT Clock Slope
 * Purpose:  This function returns the polarity of the active slope of the 
 *           feedback clock.
 *           
 *           Required Option(s):
 *           R&S SMU-K80
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetBERTClockSlope(
    ViSession   instrSession,
    ViInt32*    clockSlope
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_getEnum (instrSession,
                          "BERT:SET:CLOC?",
                          clockSlope,
                          bertClockSlopeArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set BERT Data Polarity
 * Purpose:  This function sets the polarity of the feedback data bits.
 *           
 *           Required Option(s):
 *           R&S SMU-K80
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetBERTDataPolarity(
    ViSession   instrSession,
    ViInt32 dataPolarity
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_setEnum (instrSession,
                          "BERT:SET:DATA",
                          dataPolarity,
                          dataPolarityArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get BERT Data Polarity
 * Purpose:  This function returns the polarity of the feedback data bits.
 *           
 *           Required Option(s):
 *           R&S SMU-K80
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetBERTDataPolarity(
    ViSession   instrSession,
    ViInt32*    dataPolarity
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_getEnum (instrSession,
                          "BERT:SET:DATA?",
                          dataPolarity,
                          dataPolarityArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set BERT External Restart
 * Purpose:  This function activates/deactivates an external restart of the 
 *           BER measurement.
 *           
 *           Required Option(s):
 *           R&S SMU-K80
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetBERTExternalRestart(
    ViSession   instrSession,
    ViBoolean   externalRestart
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_setViBoolean (instrSession,
                               "BERT:SET:REST:STAT",
                               externalRestart,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get BERT External Restart
 * Purpose:  This function returns if an external restart of the BER 
 *           measurement is activated or not.
 *           
 *           Required Option(s):
 *           R&S SMU-K80
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetBERTExternalRestart(
    ViSession   instrSession,
    ViBoolean*  externalRestart
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_getViBoolean (instrSession,
                               "BERT:SET:REST:STAT?",
                               externalRestart,
                               VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set BERT Data Enable
 * Purpose:  This function activates/deactivates the use of the Data Enable 
 *           signal and the polarity of the signal if it is used. The Data 
 *           Enable signal marks the data that is actually to be evaluated for 
 *           the BER measurement. Any data in addition to the PRBS sequence is 
 *           masked and thus not evaluated (e.g. sync, preambles, other 
 *           channels, etc that are present in the data bits supplied by the 
 *           DUT).
 *           
 *           Required Option(s):
 *           R&S SMU-K80
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetBERTDataEnable(
    ViSession   instrSession,
    ViInt32 dataEnable
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_setEnum (instrSession,
                          "BERT:SET:DEN",
                          dataEnable,
                          dataEnableArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get BERT Data Enable
 * Purpose:  This function returns if the use of the Data Enable signal and 
 *           the polarity of the signal if it is used is activated or not. The 
 *           Data Enable signal marks the data that is actually to be evaluated 
 *           for the BER measurement. Any data in addition to the PRBS sequence 
 *           is masked and thus not evaluated (e.g. sync, preambles, other 
 *           channels, etc that are present in the data bits supplied by the 
 *           DUT).  
 *           
 *           Required Option(s):
 *           R&S SMU-K80
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetBERTDataEnable(
    ViSession   instrSession,
    ViInt32*    dataEnable
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_getEnum (instrSession,
                          "BERT:SET:DEN?",
                          dataEnable,
                          dataEnableArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set BERT Pattern Ignore
 * Purpose:  This function activates/deactivates ignoring of pure "0" or "1" 
 *           bit sequences at least 32 bits long. Activating excludes faulty 
 *           frames from the measurement. In the case of some mobile radio 
 *           standards, pure "0" or "1" bit sequences are generated when errors 
 *           (e.g. an incorrect checksum) are detected within a frame. These 
 *           sequences, instead of the frame data, are provided for the BER 
 *           measurement and signal that the frame in question should not be 
 *           used for the measurement.
 *           
 *           Required Option(s):
 *           R&S SMU-K80
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetBERTPatternIgnore(
    ViSession   instrSession,
    ViInt32 patternIgnore
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_setEnum (instrSession,
                          "BERT:SET:IGN",
                          patternIgnore,
                          patternIgnoreArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get BERT Pattern Ignore
 * Purpose:  This function returns if ignoring of pure "0" or "1" bit 
 *           sequences at least 32 bits long is activated or not. Activating 
 *           excludes faulty frames from the measurement. In the case of some 
 *           mobile radio standards, pure "0" or "1" bit sequences are 
 *           generated when errors (e.g. an incorrect checksum) are detected 
 *           within a frame. These sequences, instead of the frame data, are 
 *           provided for the BER measurement and signal that the frame in 
 *           question should not be used for the measurement.
 *           
 *           Required Option(s):
 *           R&S SMU-K80
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetBERTPatternIgnore(
    ViSession   instrSession,
    ViInt32*    patternIgnore
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_getEnum (instrSession,
                          "BERT:SET:IGN?",
                          patternIgnore,
                          patternIgnoreArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set BERT Impedance
 * Purpose:  This function sets the impedance of the BERT inputs Clock, Data, 
 *           Data Enable and Restart. 
 *           
 *           Required Option(s):
 *           R&S SMU-K80
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetBERTImpedance(
    ViSession   instrSession,
    ViInt32 impedance
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_setEnum (instrSession,
                          "INP:BERT:IMP",
                          impedance,
                          impedanceArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get BERT Impedance
 * Purpose:  This function returns the impedance of the BERT inputs Clock, 
 *           Data, Data Enable and Restart. 
 *           
 *           Required Option(s):
 *           R&S SMU-K80
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetBERTImpedance(
    ViSession   instrSession,
    ViInt32*    impedance
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_getEnum (instrSession,
                          "INP:BERT:IMP?",
                          impedance,
                          impedanceArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set BERT Threshold
 * Purpose:  This function sets the high/low threshold of the BERT inputs 
 *           Clock, Data, Data Enable and Restart. In the case of positive 
 *           polarity, this threshold determines the point as of which a signal 
 *           is high (active) or low (inactive). 
 *           
 *           Required Option(s):
 *           R&S SMU-K80
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetBERTThreshold(
    ViSession   instrSession,
    ViReal64    threshold
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_setViReal64 (instrSession,
                              "INP:BERT:THR",
                              threshold,
                              0.0,2.0,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get BERT Threshold
 * Purpose:  This function returns high/low threshold of the BERT inputs 
 *           Clock, Data, Data Enable and Restart. In the case of positive 
 *           polarity, this threshold determines the point as of which a signal 
 *           is high (active) or low (inactive). 
 *           
 *           Required Option(s):
 *           R&S SMU-K80
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetBERTThreshold(
    ViSession   instrSession,
    ViReal64*   threshold
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_getViReal64 (instrSession,
                              "INP:BERT:THR?",
                              threshold,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set BERT Display Unit
 * Purpose:  This function sets the unit for the error rate display on the 
 *           screen. BERT:RES? always specifies the error rate as the quotient 
 *           of "Number of errors" and "Number of data items", unaffected by 
 *           this command.
 *           
 *           Required Option(s):
 *           R&S SMU-K80
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetBERTDisplayUnit(
    ViSession   instrSession,
    ViInt32 displayUnit
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_setEnum (instrSession,
                          "BERT:UNIT",
                          displayUnit,
                          displayUnitArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get BERT Display Unit
 * Purpose:  This function returns the unit for the error rate display on the 
 *           screen. BERT:RES? always specifies the error rate as the quotient 
 *           of "Number of errors" and "Number of data items" unaffected by 
 *           this function.
 *           
 *           Required Option(s):
 *           R&S SMU-K80
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetBERTDisplayUnit(
    ViSession   instrSession,
    ViInt32*    displayUnit
)
{
    INIT;
    
    CHECKOPT (K80);
    
    return rsafq_getEnum (instrSession,
                          "BERT:UNIT?",
                          displayUnit,
                          displayUnitArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Clear Status
 * Purpose:  CLEAR STATUS sets the status byte (STB), the standard event 
 *           register (ESR), and the EVENt part of the QUEStionable and 
 *           OPERation register to zero. The function does not change the 
 *           enable and transition parts of the registers. It deletes the 
 *           output buffer.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ClearStatus(
    ViSession   instrSession
)
{
    return rsafq_sendCmd (instrSession,"*CLS");
}

/*****************************************************************************
 * Function: Event Status Enable
 * Purpose:  EVENT STATUS ENABLE sets the event status enable register to the 
 *           specified value. The *ESE? query command returns the content of 
 *           the event status enable register in decimal form.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_EventStatusEnable(
    ViSession   instrSession,
    ViInt32 registerValue
)
{
    return rsafq_setViInt32 (instrSession,
                             "*ESE",
                             registerValue,
                             0,255,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Event Status Query
 * Purpose:  STANDARD EVENT STATUS QUERY returns the content of the event 
 *           status register in decimal form (0...255) and then sets the 
 *           register to zero.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_EventStatusQuery(
    ViSession   instrSession,
    ViInt32*    registerValue
)
{
    return rsafq_getViInt32 (instrSession,
                             "*ESR?",
                             registerValue,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Identification Query
 * Purpose:  IDENTIFICATION QUERY queries the instrument identification. The 
 *           instrument type, serial number and firmware version are returned.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_IdentificationQuery(
    ViSession   instrSession,
    ViChar  _VI_FAR instrumentIdentification[]
)
{
    ViStatus rsafq_status = VI_SUCCESS;

    rsafq_status = rsafq_getViString (instrSession,
                                      "*IDN?",
                                      instrumentIdentification,
                                      256,
                                      VI_ERROR_PARAMETER2);

    return (rsafq_status < VI_SUCCESS) ? rsafq_status : VI_SUCCESS;
}

/*****************************************************************************
 * Function: Individual Status Query
 * Purpose:  INDIVIDUAL STATUS QUERY returns the content of the IST flag in 
 *           decimal form (0 | 1). The IST flag is the status bit which is sent 
 *           during a parallel poll.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_IndividualStatusQuery(
    ViSession   instrSession,
    ViInt32*    istFlag
)
{
    return rsafq_getViInt32 (instrSession,
                             "*IST?",
                             istFlag,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Operation Complete
 * Purpose:  OPERATION COMPLETE sets bit 0 in the event status register when 
 *           all previous commands have been processed. This bit can be used to 
 *           trigger a service request.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_OperationComplete(
    ViSession   instrSession
)
{
    return rsafq_sendCmd (instrSession,"*OPC");
}

/*****************************************************************************
 * Function: Operation Complete Query
 * Purpose:  OPERATION COMPLETE QUERY returns a 1 when all previous commands 
 *           have been processed. It is important to ensure that the timeout 
 *           set at the IEC/IEEE bus is long enough.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_OperationCompleteQuery(
    ViSession   instrSession,
    ViInt32*    opcState
)
{
    return rsafq_getViInt32 (instrSession,
                             "*OPC?",
                             opcState,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Option Identification Query
 * Purpose:  OPTION IDENTIFICATION QUERY queries the options configured in the 
 *           instrument and returns a list of the installed options. The 
 *           options are separated by commas. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_OptionIdentificationQuery(
    ViSession   instrSession,
    ViChar  _VI_FAR optionIdentification[]
)
{
    return rsafq_getViString (instrSession,
                              "*OPT?",
                              optionIdentification,
                              1024,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Parallel Poll Register Enable
 * Purpose:  PARALLEL POLL REGISTER ENABLE sets the parallel poll enable 
 *           register to the specified value. The *PRE? query command returns 
 *           the content of the parallel poll enable register in decimal form.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ParallelPollRegisterEnable(
    ViSession   instrSession,
    ViInt32 registerValue
)
{
    return rsafq_setViInt32 (instrSession,
                             "*PRE",
                             registerValue,
                             0,655,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Parallel Poll Register Query
 * Purpose:  PARALLEL POLL REGISTER QUERY returns the content of the parallel 
 *           poll enable register in decimal form.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ParallelPollRegisterQuery(
    ViSession   instrSession,
    ViInt32*    registerValue
)
{
    return rsafq_getViInt32 (instrSession,
                             "*PRE?",
                             registerValue,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Power On Status Clear
 * Purpose:  POWER ON STATUS CLEAR determines whether the content of the 
 *           ENABle registers are retained or reset at power on.
 *           
 *           The query *PSC? retrieves the contents of the 
 *           Power-on-Status-Clear flag. The answer is either 0 or 1.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_PowerOnStatusClear(
    ViSession   instrSession,
    ViInt32 registerValue
)
{
    return rsafq_setViInt32 (instrSession,
                             "*PSC",
                             registerValue,
                             0,1,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Power On Status Clear Query
 * Purpose:  POWER ON STATUS CLEAR QUERY retrieves the contents of the 
 *           Power-on-Status-Clear flag. The answer is either 0 or 1.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_PowerOnStatusClearQuery(
    ViSession   instrSession,
    ViInt32*    registerValue
)
{
    return rsafq_getViInt32 (instrSession,
                             "*PSC?",
                             registerValue,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Recall
 * Purpose:  RECALL calls up the instrument status which was stored under the 
 *           specified number using the *SAV command or using the 
 *           MMEMory:STORe:STATe command. "0" corresponds to the power-on 
 *           state, unless a new instrument setting was stored under this 
 *           number in the meantime using the MMEMory:STORe:STATe command. This 
 *           function can also be used to call up the 3 intermediate instrument 
 *           states which are stored with *SAV.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_Recall(
    ViSession   instrSession,
    ViInt32 instrumentStatus
)
{
    return rsafq_setViInt32 (instrSession,
                             "*RCL",
                             instrumentStatus,
                             0,10,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Save
 * Purpose:  SAVE stores the current device state under the specified number 
 *           (see also *RCL). This function is used to store the 3 intermediate 
 *           instrument states and to store the current instrument state 
 *           internally when the instrument is switched off. This state is then 
 *           restored at power on.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_Save(
    ViSession   instrSession,
    ViInt32 instrumentStatus
)
{
    return rsafq_setViInt32 (instrSession,
                             "*SAV",
                             instrumentStatus,
                             0,10,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Service Request Enable
 * Purpose:  SERVICE REQUEST ENABLE sets the service request enable register 
 *           to the specified value. Bit 6 (MSS enable bit) remains 0. This 
 *           function determines the conditions under which a service request 
 *           is triggered. The *SRE? query command reads out the content of the 
 *           service request enable register in decimal form. Bit 6 is always 
 *           0.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ServiceRequestEnable(
    ViSession   instrSession,
    ViInt32 registerValue
)
{
    return rsafq_setViInt32 (instrSession,
                             "*SRE",
                             registerValue,
                             0,255,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Service Request Enable Query
 * Purpose:  SERVICE REQUEST ENABLE reads out the content of the service 
 *           request enable register in decimal form. Bit 6 is always 0.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ServiceRequestEnableQuery(
    ViSession   instrSession,
    ViInt32*    registerValue
)
{
    return rsafq_getViInt32 (instrSession,
                             "*SRE?",
                             registerValue,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Status Byte Query
 * Purpose:  READ STATUS BYTE QUERY reads out the content of the status byte 
 *           in decimal form.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_StatusByteQuery(
    ViSession   instrSession,
    ViInt32*    statusByte
)
{
    return rsafq_getViInt32 (instrSession,
                             "*STB?",
                             statusByte,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Trigger
 * Purpose:  TRIGGER triggers all actions which are waiting for a trigger 
 *           event. Specific trigger events can be triggered by means of the 
 *           "TRIGger" command system (sweep and lists) or the SOURce 
 *           subsystems (baseband).
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_Trigger(
    ViSession   instrSession
)
{
    return rsafq_sendCmd (instrSession, "*TRG");
}

/*****************************************************************************
 * Function: Trigger and Wait for OPC
 * Purpose:  This function triggers all actions waiting for a trigger event in 
 *           the currently active window and waits for operation completed 
 *           (OPC) before returning the status code.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_TriggerWaitOPC(
    ViSession   instrSession,
    ViInt32 timeout
)
{
    ViStatus rsafq_status = VI_SUCCESS;
    ViChar buffer[BUFFER_SIZE];

    CHECKPARAM (rsafq_invalidViInt32Range (timeout, 0, 600000), 2);

    CHECKERR (rsafq_waitOPC (instrSession, "*TRG;*OPC?", buffer, timeout));

    return rsafq_status;
}

/*****************************************************************************
 * Function: Self Test Query
 * Purpose:  SELF TEST QUERY triggers all self tests of the instrument and 
 *           outputs an error code in decimal form (see Service Manual supplied 
 *           with the instrument).
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SelfTestQuery(
    ViSession   instrSession,
    ViInt32*    errorCode
)
{
    return rsafq_getViInt32 (instrSession,
                             "*TST?",
                             errorCode,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Wait to Continue
 * Purpose:  WAIT-to-CONTINUE does not allow subsequent commands to be 
 *           processed until all previous commands have been executed and all 
 *           signals are in their transient condition.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_WaittoContinue(
    ViSession   instrSession
)
{
    return rsafq_sendCmd (instrSession, "*WAI");
}

/*****************************************************************************
 * Function: CALibration All
 * Purpose:  This function starts all internal adjustments for which no 
 *           external measuring equipment is needed. The adjustments with 
 *           external measuring equipment are described in the Service Manual 
 *           
 *           Note:
 *           
 *           If the calibration fails, this function returns an error code.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_CALibrationAll(
    ViSession   instrSession,
    ViInt32     timeout,
    ViInt32*    response
)
{
    ViStatus    rsafq_status = VI_SUCCESS;
    ViChar      buffer[BUFFER_SIZE];

    if (rsafq_invalidViInt32Range (timeout, 0, 600000))
        return VI_ERROR_PARAMETER2;
    if (!response)
        return VI_ERROR_PARAMETER3;

    sprintf (buffer, ":CAL:ALL?");

    CHECKERR (rsafq_waitOPC (instrSession, buffer, buffer, timeout));

    if (sscanf (buffer, "%ld", response) != 1)
        return VI_ERROR_INSTR_INTERPRETING_RESPONSE;

    if (*response == 1) /* Calibration failed. */
        rsafq_status = RSAFQ_ERROR_CALIBRATION_FAILED;

    return rsafq_status;
}

/*****************************************************************************
 * Function: CALibration Amplitude
 * Purpose:  This function starts all internal adjustments which affect the 
 *           amplitude at the analog output.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_CALibrationAmplitude(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViInt32 timeout,
    ViInt32*    response
)
{
    ViStatus    rsafq_status = VI_SUCCESS;
    ViChar      buffer[BUFFER_SIZE];

    if (rsafq_invalidViInt32Range (timeout, 0, 600000))
        return VI_ERROR_PARAMETER2;
    if (!response)
        return VI_ERROR_PARAMETER3;

    sprintf (buffer, "CAL:AOUT%ld:AMPL?", analogOutput);

    CHECKERR (rsafq_waitOPC (instrSession, buffer, buffer, timeout));

    if (sscanf (buffer, "%ld", response) != 1)
        return VI_ERROR_INSTR_INTERPRETING_RESPONSE;

    if (*response == 1) /* Calibration failed. */
        rsafq_status = RSAFQ_ERROR_CALIBRATION_FAILED;

    return rsafq_status;
}

/*****************************************************************************
 * Function: Calibration Analog Output Delay 
 * Purpose:  This function adjusts the amplitude of the (selected) analog 
 *           output. The instrument generates a series of signal levels and 
 *           converts them using a precise ADC. The procedure takes about 30 
 *           seconds.
 *           
 *           Note(s):
 *           
 *           (1) This function is available on AFQ100B only.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_CalibrationAnalogOutputDelay(
    ViSession   instrSession,
    ViInt32 analogOutput,
    ViInt32 timeout,
    ViInt32*    response
)
{
    ViChar      buffer[BUFFER_SIZE];

	INIT;
	CHECKVER(AFQ100B);
    CHECKPARAM (rsafq_invalidViInt32Range (timeout, 0, 600000), 2);
    if (!response)
        return VI_ERROR_PARAMETER3;

    sprintf (buffer, "CAL:AOUT%ld:DEL?", analogOutput);

    CHECKERR (rsafq_waitOPC (instrSession, buffer, buffer, timeout));

    if (sscanf (buffer, "%ld", response) != 1)
        return VI_ERROR_INSTR_INTERPRETING_RESPONSE;

    if (*response == 1) /* Calibration failed. */
        rsafq_status = RSAFQ_ERROR_CALIBRATION_FAILED;

    return rsafq_status;
}

/*============================================================================
 * Function: DIAGnostic Modules
 * Purpose:  This function checks the assemblies available in the instrument
 *           using the variant and revision state.
 *===========================================================================*/
ViStatus _VI_FUNC rsafq_DIAGnosticModules (ViSession instrSession,
                                           ViString moduleName,
                                           ViInt32 arraySize,
                                           ViChar moduleConfiguration[])
{
    ViChar      buffer [BUFFER_SIZE];

    CHECKNULL (moduleConfiguration, 4);

    if (!moduleName)
        sprintf (buffer, "DIAG:BGIN?;*OPC?");
    else
        sprintf (buffer, "DIAG:BGIN? '%s';*OPC?", moduleName);

    return rsafq_getCatalog (instrSession,
                             buffer,
                             (ViUInt32) arraySize,
                             moduleConfiguration,
                             VI_ERROR_PARAMETER4);    
}

/*****************************************************************************
 * Function: DIAGnostic Test Points
 * Purpose:  This function queries the test points available in the 
 *           instrument. The test points on the analog output board are queried 
 *           with command DIAG:AOUT:POIN:CAT?. A detailed description of the 
 *           test points can be found in chapter 3 , section "Trouble Shooting" 
 *           of the Service Manual (on CD-ROM, supplied with the instrument).
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_DIAGnosticTestPoints(
    ViSession   instrSession,
    ViInt32     arraySize,
    ViChar      testPoints[]
)
{
    return rsafq_getCatalog (instrSession,
                             ":DIAG:POIN:CAT?;*OPC?",
                             arraySize,
                             testPoints,
                             VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: DIAGnostic Test Point Voltages
 * Purpose:  This function triggers voltage measurement at the specified test 
 *           point and returns the measured voltage. A detailed description of 
 *           the test points can be found in chapter 3 , section "Trouble 
 *           Shooting" of the Service Manual (on CD-ROM, supplied with the 
 *           instrument).
 *           
 *           A list of the available test points can be queried using the 
 *           DIAG:POIN:CAT? command. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_DIAGnosticTestPointVoltages(
    ViSession   instrSession,
    ViString    testPointName,
    ViReal64*   measuredVoltage
)
{
    ViChar      buffer [BUFFER_SIZE];

    CHECKNULL (testPointName, 2);

    sprintf (buffer, ":DIAG:POIN? \"%s\"", testPointName);

    return rsafq_getViReal64 (instrSession,
                              buffer,
                              measuredVoltage,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: DIAGnostic AOUT Test Points
 * Purpose:  This function measures the voltage the test points available in 
 *           the instrument.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_DIAGnosticAOUTTestPoints(
    ViSession   instrSession,
    ViInt32     analogOutput,
    ViInt32     arraySize,
    ViChar      testPoints[]
)
{
    ViChar  buffer [BUFFER_SIZE];

    CHECKPARAM(rsafq_invalidViInt32Range (analogOutput, 1, 2), 2);
        
    sprintf (buffer, ":DIAG:AOUT%ld:POIN:CAT?;*OPC?", analogOutput);

    return rsafq_getCatalog (instrSession,
                             buffer,
                             arraySize,
                             testPoints,
                             VI_ERROR_PARAMETER4);
}

/*****************************************************************************
 * Function: DIAGnostic AOUT Test Point Voltages
 * Purpose:  This function triggers voltage measurement at the specified test 
 *           point of the analog output module and returns the measured 
 *           voltage. A detailed description of the test points can be found in 
 *           chapter 3 , section "Trouble Shooting" of the Service Manual 
 *           supplied with the instrument on CD-ROM.
 *           
 *           A list of the available test points can be queried using the 
 *           DIAG:AOUT:POIN:CAT? command .   
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_DIAGnosticAOUTTestPointVoltages(
    ViSession   instrSession,
    ViInt32     analogOutput,
    ViString    testPointName,
    ViReal64*   measuredVoltage
)
{
    ViChar    command[256];
    
    CHECKPARAM(rsafq_invalidViInt32Range (analogOutput, 1, 2), 2);
    
    sprintf(command,"DIAG:AOUT%ld:POINt? \"%s\"", analogOutput, testPointName);
    
    return rsafq_getViReal64 (instrSession,
                              command,
                              measuredVoltage,
                              VI_ERROR_PARAMETER4);
}

/*****************************************************************************
 * Function: Set Fan Control
 * Purpose:  This function sets the fan speed of the instrument. The full 
 *           speed mode is provided for service purposes.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetFanControl(
    ViSession   instrSession,
    ViInt32 fanSpeed
)
{
    return rsafq_setEnum (instrSession,
                          "DIAG:FCON",
                          fanSpeed,
                          fanSpeedArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get Fan Control
 * Purpose:  This function returns the fan speed of the instrument.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetFanControl(
    ViSession   instrSession,
    ViInt32*    fanSpeed
)
{
    return rsafq_getEnum (instrSession,
                          "DIAG:FCON?",
                          fanSpeed,
                          fanSpeedArr,
                          VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: MMEMory Catalog
 * Purpose:  This function reads out the subdirectories and files in the 
 *           specified directory. If no directory is specified, the default 
 *           directory selected with the MMEM:CDIR command is read out on the 
 *           default drive selected with the MMEM:MSIS command. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_MMEMoryCatalog(
    ViSession   instrSession,
    ViString    directoryPath,
    ViInt32 arraySize,
    ViChar  catalogList[]
)
{
    ViChar  buffer [BUFFER_SIZE];

    if (!directoryPath)
        sprintf (buffer, ":MMEM:CAT?;*OPC?");
    else
        sprintf (buffer, ":MMEM:CAT? '%s';*OPC?", directoryPath);

    return rsafq_getCatalog (instrSession,
                             buffer,
                             arraySize,
                             catalogList,
                             VI_ERROR_PARAMETER4);
}

/*****************************************************************************
 * Function: MMEMory Catalog Length
 * Purpose:  This function reads out the number of files in the specified 
 *           directory. If no directory is specified, the default directory 
 *           selected with the MMEM:CDIR command is read out on the default 
 *           drive selected with the MMEM:MSIS command. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_MMEMoryCatalogLength(
    ViSession   instrSession,
    ViString    directoryPath,
    ViInt32*    catalogLength
)
{
    ViChar  buffer [BUFFER_SIZE];

    if (!directoryPath)
        sprintf (buffer, ":MMEM:CAT:LENG?");
    else
        sprintf (buffer, ":MMEM:CAT:LENG? '%s'", directoryPath);

    return rsafq_getViInt32 (instrSession,
                             buffer,
                             catalogLength,
                             VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: MMEMory Change Directory
 * Purpose:  This function changes the default directory. This directory is 
 *           used for all subsequent MMEM commands if no path is specified with 
 *           them. It is also possible to change to a higher directory using 
 *           two dots '..'.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_MMEMoryChangeDirectory(
    ViSession   instrSession,
    ViString    directoryName
)
{
    return rsafq_setViStringQuoted (instrSession,
                                    ":MMEM:CDIR",
                                    "'",
                                    directoryName,
                                    VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: MMEMory Copy
 * Purpose:  This function copies the first specified file to the second 
 *           specified file. Instead of just a file, this function can also be 
 *           used to copy a complete directory together with all its files. 
 *           
 *           If <destination> is not specified, <source> is copied to the 
 *           MMEM:MSIS drive and the MMEM:CDIR directory. Files which already 
 *           exist with the same name in the destination directory are 
 *           overwritten without an error message.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_MMEMoryCopy(
    ViSession   instrSession,
    ViString    sourcePath,
    ViString    destinationPath
)
{
    ViStatus    rsafq_status;
    ViChar      buffer[BUFFER_SIZE];

    CHECKNULL (sourcePath, 2);

    if (destinationPath)
        sprintf (buffer, ":MMEM:COPY '%s','%s'", sourcePath, destinationPath);
    else
        sprintf (buffer, ":MMEM:COPY '%s'", sourcePath);

    rsafq_status = rsafq_sendCmd (instrSession, buffer);

    return rsafq_status;
}

/*****************************************************************************
 * Function: MMEMory Block Data Write
 * Purpose:  This function writes the block data <binary block data> to the 
 *           file identified by <file_name>. The IEC/IEEE-bus terminator should 
 *           be set to EOI in order to ensure correct data transfer.
 *           
 *           This function can be used to transfer stored instrument settings 
 *           or waveforms directly to the instrument.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_MMEMoryBlockDataWrite(
    ViSession   instrSession,
    ViString    fileName,
    ViInt32 binaryBlockLength,
    //ViConstString   binaryBlockData
    ViChar _VI_FAR binaryBlockData[]
)
{
    ViStatus    rsafq_status    = VI_SUCCESS;
    ViUInt16    intfType        = VI_INTF_GPIB;
    ViChar*     pbuffer;
    ViChar*     p2buf;

    CHECKNULL (fileName, 2);
    CHECKNULL (binaryBlockData, 4);

    CHECKERR (viGetAttribute (instrSession, VI_ATTR_INTF_TYPE, &intfType));

    if (intfType == VI_INTF_GPIB)
        CHECKERR (viWrite (instrSession, (ViBuf)":SYST:COMM:GPIB:LTER EOI\n", 25, VI_NULL));

    p2buf = pbuffer = malloc (strlen (fileName) + binaryBlockLength + 1000);

    p2buf += sprintf (pbuffer, ":MMEM:DATA '%s',#%ld%ld",
                               fileName,
                               (ViInt32)(floor(log10(binaryBlockLength)) + 1),
                               binaryBlockLength);

    memcpy (p2buf, binaryBlockData, binaryBlockLength);
//    p2buf[binaryBlockLength] = '\n';

    rsafq_status = viWrite (instrSession, (ViPBuf)pbuffer, (ViUInt32)((p2buf - pbuffer) + binaryBlockLength), VI_NULL);

    if (pbuffer)
        free (pbuffer);

    if (intfType == VI_INTF_GPIB)
        viWrite (instrSession, (ViBuf)":SYST:COMM:GPIB:LTER STAN\n", 26, VI_NULL);

    return rsafq_status;
}

/*****************************************************************************
 * Function: MMEMory Block Data Read
 * Purpose:  This function transfers the specified file from the R&S SMU to 
 *           the IEC/IEEE bus and then on to the control computer. It is 
 *           important to ensure that the intermediate memory on the control 
 *           computer is large enough to take the file. In this case, the 
 *           setting for the IEC/IEEE-bus terminator is irrelevant.
 *           
 *           This function can be used to read stored instrument settings or 
 *           waveforms directly from the instrument.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_MMEMoryBlockDataRead(
    ViSession   instrSession,
    ViString    fileName,
    ViInt32 binaryBlockLength,
    ViChar  binaryBlockData[]
)
{
    ViStatus    rsafq_status    = VI_SUCCESS;
    ViUInt16    intfType        = VI_INTF_GPIB;
    ViUInt32    count           = 0,
                lengthOfLength  = 0;
    ViChar      buffer[BUFFER_SIZE],
                format[32],
                *pbuffer, *p2buffer;

    CHECKNULL (fileName, 2);

    CHECKERR (viGetAttribute (instrSession, VI_ATTR_INTF_TYPE, &intfType));

    if (intfType == VI_INTF_GPIB)
        CHECKERR (viWrite (instrSession, (ViBuf)":SYST:COMM:GPIB:LTER EOI\n", 25, VI_NULL));

    sprintf (buffer, ":MMEM:DATA? '%s'\n", fileName);

    CHECKERR (viWrite (instrSession, (ViBuf)buffer, (ViUInt32)strlen(buffer), VI_NULL));
    CHECKERR (rsafq_readDataUnknownLength (instrSession, &pbuffer, &count));

    p2buffer = pbuffer;

    if (sscanf (p2buffer, "#%1d", &lengthOfLength) != 1)
        {
        free (pbuffer);
        return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
        }

    sprintf (format, "%%%ldd", lengthOfLength);
    p2buffer += 2;

    if (sscanf (p2buffer, format, &count) != 1)
        {
        free (pbuffer);
        return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
        }

    p2buffer += lengthOfLength;

    if (intfType == VI_INTF_GPIB)
        if ((rsafq_status = viWrite (instrSession, (ViBuf)":SYST:COMM:GPIB:LTER STAN\n", 26, VI_NULL)) < VI_SUCCESS)
            {
            free (pbuffer);
            return rsafq_status;
            }

    if ((rsafq_status = rsafq_checkStatus (instrSession)) < VI_SUCCESS)
        {
        free (pbuffer);
        return rsafq_status;
        }

    if (count > (ViUInt32)binaryBlockLength)
        {
        rsafq_status = RSAFQ_WARNING_SMALL_BUFFER;
        count = binaryBlockLength;
        memcpy (binaryBlockData, p2buffer, count);
        }
    else
        {
        rsafq_status = VI_SUCCESS;
        memcpy (binaryBlockData, p2buffer, count);
        }

    free (pbuffer);

    return rsafq_status;
}

/*****************************************************************************
 * Function: MMEMory Directory Catalog
 * Purpose:  This function reads out the subdirectories of the specified 
 *           directory. If no directory is specified, the default directory 
 *           selected with the MMEM:CDIR command is read out. The directories 
 *           are output in a list (the list entries are separated by commas). 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_MMEMoryDirectoryCatalog(
    ViSession   instrSession,
    ViString    directoryPath,
    ViInt32 arraySize,
    ViChar  catalogList[]
)
{
    ViStatus    rsafq_status;
    ViChar      buffer[BUFFER_SIZE];

    if (!directoryPath)
        sprintf (buffer, ":MMEM:DCAT?;*OPC?");
    else
        sprintf (buffer, ":MMEM:DCAT? '%s';*OPC?", directoryPath);

    rsafq_status = rsafq_getCatalog (instrSession,
                                     buffer,
                                     arraySize,
                                     catalogList,
                                     VI_ERROR_PARAMETER4);

    return rsafq_status;
}

/*****************************************************************************
 * Function: MMemory Directory CatalogLength
 * Purpose:  This function reads out the number of subdirectories in the 
 *           specified directory. If no directory is specified, the directory 
 *           selected with the MMEM:CDIR command is read out.  
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_MMEMoryDirectoryCatalogLength(
    ViSession   instrSession,
    ViString    directoryPath,
    ViInt32*    catalogLength
)
{
    ViStatus    rsafq_status;
    ViChar      buffer[BUFFER_SIZE];

    if (!directoryPath)
        sprintf (buffer, ":MMEM:DCAT:LENG?");
    else
        sprintf (buffer, ":MMEM:DCAT:LENG? '%s'", directoryPath);

    rsafq_status = rsafq_getViInt32 (instrSession,
                                     buffer,
                                     catalogLength,
                                     VI_ERROR_PARAMETER3);

    return rsafq_status;
}

/*****************************************************************************
 * Function: MMEMory Delete
 * Purpose:  This function deletes the specified file.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_MMEMoryDelete(
    ViSession   instrSession,
    ViString    fileName
)
{
    return rsafq_setViStringQuoted (instrSession,
                                    ":MMEM:DEL",
                                    "'",
                                    fileName,
                                    VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: MMEMory Make Directory
 * Purpose:  This function creates a new subdirectory in the specified 
 *           directory. If no directory is specified, a subdirectory is created 
 *           in the default directory. This function can also be used to create 
 *           a directory tree.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_MMEMoryMakeDirectory(
    ViSession   instrSession,
    ViString    directoryName
)
{
    return rsafq_setViStringQuoted (instrSession,
                                    ":MMEM:MDIR",
                                    "'",
                                    directoryName,
                                    VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: MMEMory Move
 * Purpose:  This function renames an existing file ('File Source') if no path 
 *           is specified for 'File Destination'. Otherwise the file is moved 
 *           to the specified path and stored under the original file name or, 
 *           if specified, a new file name.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_MMEMoryMove(
    ViSession   instrSession,
    ViString    fileSource,
    ViString    fileDestination
)
{
    ViStatus    rsafq_status;
    ViChar*     pbuffer;

    CHECKNULL (fileSource, 2);
    CHECKNULL (fileDestination, 3);

    pbuffer = malloc (strlen (fileSource) + strlen (fileDestination) + 20);

    sprintf (pbuffer, ":MMEM:MOVE '%s','%s'", fileSource, fileDestination);

    rsafq_status = rsafq_sendCmd (instrSession, pbuffer);

    free (pbuffer);

    RETURN;
}

/*****************************************************************************
 * Function: MMEMory Drive
 * Purpose:  This function sets the drive (or network resource in the case of 
 *           networks) using <msus> (MSIS = Mass Storage Identification 
 *           String). This setting is effective for all MMEMory functions where 
 *           the drive is not explicitly specified in the parameter.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_MMEMoryDrive(
    ViSession   instrSession,
    ViString    driveName
)
{
    return rsafq_setViStringQuoted (instrSession,
                                    ":MMEM:MSIS",
                                    "'",
                                    driveName,
                                    VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: MMEMory Remove Directory
 * Purpose:  This function deletes the specified subdirectory in the specified 
 *           directory. If no directory is specified, the subdirectory with the 
 *           specified name is deleted in the default directory.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_MMEMoryRemoveDirectory(
    ViSession   instrSession,
    ViString    directoryName
)
{
    return rsafq_setViStringQuoted (instrSession,
                                    ":MMEM:RDIR",
                                    "'",
                                    directoryName,
                                    VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Set Register
 * Purpose:  This function sets the value of selected status register.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_setRegister(
    ViSession   instrSession,
    ViInt32 registerGroup,
    ViInt32 registerType,
    ViInt32 value
)
{
    ViChar  buffer[BUFFER_SIZE];

    CHECKPARAM (rsafq_invalidViInt32Range (registerGroup, 0, 1), 2);
    CHECKPARAM (rsafq_invalidViInt32Range (registerType, 2, 4), 3);

    sprintf (buffer, ":STATus:%s:%s", regGroupArr[registerGroup], regTypeArr[registerType]);

    return rsafq_setViInt32 (instrSession,
                             buffer,
                             value,
                             0, 0,
                             VI_ERROR_PARAMETER4);
}

/*****************************************************************************
 * Function: Get Register
 * Purpose:  This function sets the value of selected status register.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_getRegister(
    ViSession   instrSession,
    ViInt32 registerGroup,
    ViInt32 registerType,
    ViInt32*    value
)
{
    ViChar  buffer[BUFFER_SIZE];

    CHECKPARAM (rsafq_invalidViInt32Range (registerGroup, 0, 1), 2);
    CHECKPARAM (rsafq_invalidViInt32Range (registerType, 0, 1), 3);

    sprintf (buffer, ":STATus:%s:%s?", regGroupArr[registerGroup], regTypeArr[registerType]);

    return rsafq_getViInt32 (instrSession,
                             buffer,
                             value,
                             VI_ERROR_PARAMETER4);
}

/*****************************************************************************
 * Function: Set Register Bit
 * Purpose:  This function sets the value of selected status register.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_setRegisterBit(
    ViSession   instrSession,
    ViInt32 registerGroup,
    ViInt32 bit,
    ViInt32 registerType,
    ViInt32 value
)
{
    ViChar  buffer[BUFFER_SIZE];

    CHECKPARAM (rsafq_invalidViInt32Range (registerGroup, 0, 1), 2);
    if (registerGroup == RSAFQ_REGISTER_OPERATION) 
        CHECKPARAM (rsafq_invalidViInt32Range (bit, 8, 12), 3);
    else if (registerGroup == RSAFQ_REGISTER_QUESTIONABLE)
        CHECKPARAM (rsafq_invalidViInt32Range (bit, 9, 12), 3);
    CHECKPARAM (rsafq_invalidViInt32Range (registerType, 2, 4), 4);

    sprintf (buffer, ":STATus:%s:BIT%ld:%s", regGroupArr[registerGroup], bit, regTypeArr[registerType]);

    return rsafq_setViInt32 (instrSession,
                             buffer,
                             value,
                             0, 0,
                             VI_ERROR_PARAMETER4);
}

/*****************************************************************************
 * Function: Get Register Bit
 * Purpose:  This function sets the value of selected status register.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_getRegisterBit(
    ViSession   instrSession,
    ViInt32 registerGroup,
    ViInt32 bit,
    ViInt32 registerType,
    ViInt32*    value
)
{
    ViChar  buffer[BUFFER_SIZE];

    CHECKPARAM (rsafq_invalidViInt32Range (registerGroup, 0, 1), 2);
    if (registerGroup == RSAFQ_REGISTER_OPERATION) 
        CHECKPARAM (rsafq_invalidViInt32Range (bit, 8, 12), 3);
    else if (registerGroup == RSAFQ_REGISTER_QUESTIONABLE)
        CHECKPARAM (rsafq_invalidViInt32Range (bit, 9, 12), 3);
    CHECKPARAM (rsafq_invalidViInt32Range (registerType, 0, 1), 4);
    
    sprintf (buffer, ":STATus:%s:BIT%ld:%s?", regGroupArr[registerGroup], bit, regTypeArr[registerType]);

    return rsafq_getViInt32 (instrSession,
                             buffer,
                             value,
                             VI_ERROR_PARAMETER4);
}

/*****************************************************************************
 * Function: STATus PRESet
 * Purpose:  This function resets the status registers. All PTRansition parts 
 *           are set to FFFFh (32767), i.e. all transitions from 0 to 1 are 
 *           detected. All NTRansition parts are set to 0, i.e. a transition 
 *           from 1 to 0 in a CONDition bit is not detected. The ENABle parts 
 *           of STATus:OPERation and STATus:QUEStionable are set to 0, i.e. all 
 *           events in these registers are not passed on.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_STATusPreset(
    ViSession   instrSession
)
{
    return rsafq_sendCmd (instrSession, "STATus:PRES");
}

/*****************************************************************************
 * Function: SYSTem All Errors
 * Purpose:  This function returns a list of all errors existing at the time 
 *           when the query is started . This list corresponds to the display 
 *           on the info page under manual control.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SYSTemAllErrors(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViChar  allErrors[]
)
{
    return rsafq_getViString (instrSession,
                              "SYST:SERR?",
                              allErrors,
                              arraySize,
                              VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: SYSTem Error Queue All
 * Purpose:  This function queries all entries in the error queue and then 
 *           deletes them. Positive error numbers denote instrument-specific 
 *           errors, and negative error numbers denote error messages defined 
 *           by SCPI (see Chapter 9). If the error queue is empty, 0 ("No 
 *           error") is returned. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SYSTemErrorQueueAll(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViChar  errorQueue[]
)
{
    return rsafq_getViString (instrSession,
                              ":SYST:ERR:ALL?",
                              errorQueue,
                              arraySize,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: SYSTem Error Queue Code All
 * Purpose:  This function queries all entries in the error queue and then 
 *           deletes them. Only the error numbers are returned and not the 
 *           entire error text. Positive error numbers denote 
 *           instrument-specific errors, and negative error numbers denote 
 *           error messages defined by SCPI (see Chapter 9). If the error queue 
 *           is empty, 0 ("No error") is returned. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SYSTemErrorQueueCodeAll(
    ViSession   instrSession,
    ViInt32 arraySize,
    ViChar  errorQueueCode[]
)
{
    return rsafq_getViString (instrSession,
                              ":SYST:ERR:CODE:ALL?",
                              errorQueueCode,
                              arraySize,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: SYSTem Error Queue Code Next
 * Purpose:  This function queries the oldest entry in the error queue and 
 *           then deletes it. Only the error number is returned and not the 
 *           entire error text. Positive error numbers denote 
 *           instrument-specific errors, and negative error numbers denote 
 *           error messages defined by SCPI (see Chapter 9). If the error queue 
 *           is empty, 0 is returned. 
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SYSTemErrorQueueCodeNext(
    ViSession   instrSession,
    ViInt32*    errorQueueCodeNext
)
{
    return rsafq_getViInt32 (instrSession,
                             "SYST:ERR:CODE?",
                             errorQueueCodeNext,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: SYSTem Error Queue Count
 * Purpose:  This function queries the number of entries in the error queue. 
 *           If the error queue is empty, '0' is returned.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SYSTemErrorQueueCount(
    ViSession   instrSession,
    ViInt32*    errorQueueCount
)
{
    return rsafq_getViInt32 (instrSession,
                             "SYST:ERR:COUN?",
                             errorQueueCount,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Get SYSTem Version
 * Purpose:  This function queries the SCPI version with which the instrument 
 *           complies.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetSYSTemVersion(
    ViSession   instrSession,
    ViChar  scpiVersion[]
)
{
    ViStatus rsafq_status = VI_SUCCESS;

    rsafq_status = rsafq_getViString (instrSession,
                                      ":SYST:VERS?",
                                      scpiVersion,
                                      32,
                                      VI_ERROR_PARAMETER2);

    return (rsafq_status < VI_SUCCESS) ? rsafq_status : VI_SUCCESS;
}

/*****************************************************************************
 * Function: Set SYSTem Date
 * Purpose:  This function sets the date for the instrument-internal calendar.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetSYSTemDate(
    ViSession   instrSession,
    ViString    date
)
{
    ViChar   command[256];
    
    sprintf(command, "SYST:DATE %s", date);
    
    return rsafq_sendCmd (instrSession,command);
}

/*****************************************************************************
 * Function: Get SYSTem Date
 * Purpose:  This function queries the date for the instrument-internal 
 *           calendar.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetSYSTemDate(
    ViSession   instrSession,
    ViChar  _VI_FAR date[]
)
{
    return rsafq_getViString (instrSession,
                              "SYST:DATE?",
                              date,
                              256,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: Set SYSTem Time
 * Purpose:  This function sets the time for the instrument-internal clock.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SetSYSTemTime(
    ViSession   instrSession,
    ViString    time
)
{
    ViChar   command[256];
    
    sprintf(command, "SYST:TIME %s", time);
    
    return rsafq_sendCmd (instrSession,command);
}

/*****************************************************************************
 * Function: Get SYSTem Time
 * Purpose:  This function queries the time for the instrument-internal clock.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetSYSTemTime(
    ViSession   instrSession,
    ViChar  _VI_FAR time[]
)
{
    return rsafq_getViString (instrSession,
                              "SYST:TIME?",
                              time,
                              256,
                              VI_ERROR_PARAMETER3);
}

/*****************************************************************************
 * Function: SYSTem Preset
 * Purpose:  This function triggers an instrument reset. It has the same 
 *           effect as the the *RST command.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_SYSTemPreset(
    ViSession   instrSession
)
{
    return rsafq_sendCmd (instrSession, "SYST:PRES");
}

/*****************************************************************************
 * Function: DEVice Preset
 * Purpose:  This function recovers the initial state of the instrument before 
 *           shipping.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_DEVicePreset(
    ViSession   instrSession
)
{
#if 0    
    return rsafq_sendCmd (instrSession, "DEV:PRES");
#endif
	instrSession;
    
    return RSAFQ_ERROR_INSTRUMENT_NSUP_FUNC;    
}

/*****************************************************************************
 * Function: Get Purge Key
 * Purpose:  This function queries the key necessary for deleting the *.wv 
 *           files.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_GetPurgeKey(
    ViSession   instrSession,
    ViInt32*    key
)
{
    return rsafq_getViInt32 (instrSession,
                             "SOUR:WAV:PURG?",
                             key,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: ARB Waveform Purge
 * Purpose:  This function is used to delete all waveform files from local 
 *           hard disk, which is split into two partitions C:\ and D:\. To 
 *           delete all waveform files on hard disk, two steps are necessary. 
 *           The instrument generates a pseudo random number 
 *           (SOURce:WAVeform:PURGe?) which is the key necessary for deleting 
 *           the *.wv files.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_ARBWaveformPurge(
    ViSession   instrSession,
    ViInt32 key
)
{
    return rsafq_setViInt32 (instrSession,
                             "SOUR:WAV:PURG",
                             key,
                             0,0,
                             VI_ERROR_PARAMETER2);
}

/*****************************************************************************
 * Function: Read To File From Instrument
 * Purpose:  This function is used to read data from the instrument and write 
 *           it to a user specified file on the host computer.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_readToFile(
    ViSession   instrSession,
    ViString    source,
    ViString    destination
)
{
    ViStatus    rsafq_status    = VI_SUCCESS;
    ViUInt16    intfType        = VI_INTF_GPIB;
    ViUInt32    count           = 0,
                lengthOfLength  = 0;
    ViChar      buffer[BUFFER_SIZE],
                format[32],
                *pbuffer, *p2buffer;
    FILE        *targetFile;

    CHECKERR (viGetAttribute (instrSession, VI_ATTR_INTF_TYPE, &intfType));

    if (intfType == VI_INTF_GPIB)
        CHECKERR (viWrite (instrSession, (ViBuf)":SYST:COMM:GPIB:LTER EOI\n", 25, VI_NULL));

    sprintf (buffer, ":MMEM:DATA? '%s'\n", source);

    CHECKERR (viWrite (instrSession, (ViBuf)buffer, (ViUInt32)strlen(buffer), VI_NULL));
    CHECKERR (rsafq_readDataUnknownLength (instrSession, &pbuffer, &count));

    p2buffer = pbuffer;

    if (sscanf (p2buffer, "#%1d", &lengthOfLength) != 1)
        {
        free (pbuffer);
        return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
        }

    sprintf (format, "%%%ldd", lengthOfLength);
    p2buffer += 2;

    if (sscanf (p2buffer, format, &count) != 1)
        {
        free (pbuffer);
        return VI_ERROR_INSTR_INTERPRETING_RESPONSE;
        }

    p2buffer += lengthOfLength;

    if (intfType == VI_INTF_GPIB)
        if ((rsafq_status = viWrite (instrSession, (ViBuf)":SYST:COMM:GPIB:LTER STAN\n", 26, VI_NULL)) < VI_SUCCESS)
            {
            free (pbuffer);
            return rsafq_status;
            }

    if ((rsafq_status = rsafq_checkStatus (instrSession)) < VI_SUCCESS)
        {
        free (pbuffer);
        return rsafq_status;
        }

    /* Store data from instrument to file on host computer */
    if ((targetFile = fopen (destination, "wb")) == VI_NULL)
        {
        if (pbuffer)
            free (pbuffer);
        return VI_ERROR_INSTR_FILE_OPEN;
        }

    if (count != fwrite (p2buffer, sizeof (ViByte), (size_t)(count), targetFile))
        {
        fclose (targetFile);
        if (pbuffer)
            free (pbuffer);
        return VI_ERROR_INSTR_FILE_WRITE;
        }

    fclose (targetFile);

    if (pbuffer) /* If not NULL pointer, then deallocate memory. */
        free (pbuffer);

    if ((rsafq_status = rsafq_checkStatus (instrSession)) < 0)
        return rsafq_status;

    return rsafq_status;
}

/*****************************************************************************
 * Function: Write From File To Instrument
 * Purpose:  This function is used to read data from the host computer and 
 *           write it to a user specified file in the instrument.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_writeFromFile(
    ViSession   instrSession,
    ViString    source,
    ViString    destination
)
{
    ViBoolean   eofile;
    ViUInt32    offset, block;
    ViStatus    rsafq_status;
    ViInt64     fhandle;
    
    rsafq_status = rsafq_writeHugeFileStart (instrSession, source, destination, &fhandle);
                                             
    offset = 0;
    block = 10000;
    do
    {
        rsafq_status = rsafq_contHugeFile (instrSession, fhandle, offset, block, &eofile);
        offset += block;
    } while (eofile == VI_FALSE);
    
    return rsafq_status;
}

/*===========================================================================*/
/* Function: Start Write Huge File To Instrument                             */
/* Purpose:  This function is used to read data from the host computer and   */
/*           write it to a user specified file in the instrument. It is used */
/*           for huge files. It sends only file header. To send file data,   */
/*           call into rssmu_contHugeFile in a loop.                         */
/*===========================================================================*/
ViStatus _VI_FUNC rsafq_writeHugeFileStart (ViSession instrSession,
                                            ViString source,
                                            ViString destination,
                                            ViInt64 *fileHandle)
{
    ViStatus    rsafq_status;
    FILE        *source_file;
    ViUInt32    file_size;
    ViUInt16    intfType        = VI_INTF_GPIB;
    ViChar      buffer [BUFFER_SIZE];
    
    if ((source_file = fopen (source, "rb")) == VI_NULL)
        return VI_ERROR_INSTR_FILE_OPEN;

    /* Get size of file */
    fseek (source_file, 0, SEEK_END);
    file_size = ftell (source_file);
    fseek (source_file, 0, SEEK_SET);

    CHECKERR (viGetAttribute (instrSession, VI_ATTR_INTF_TYPE, &intfType));

    if (intfType == VI_INTF_GPIB)
        viWrite (instrSession, (ViBuf)":SYST:COMM:GPIB:LTER STAN\n", 26, VI_NULL);

    CHECKERR (viSetAttribute (instrSession, VI_ATTR_SEND_END_EN, VI_FALSE));
    
    sprintf (buffer, ":MMEM:DATA '%s',#%ld%ld",
                               destination,
                               (ViInt32)(floor(log10(file_size)) + 1),
                               file_size);
    CHECKERR (viWrite (instrSession, (ViBuf)buffer, (ViUInt32)strlen(buffer), NULL));
    *fileHandle = (ViInt64) source_file;
    
    return rsafq_status;    
}

/*===========================================================================*/
/* Function: Continue Write Huge File To Instrument                          */
/* Purpose:  This function is used to read data from the host computer and   */
/*           write it to a user specified file in the instrument. It is used */
/*           for huge files. To start file transfer, call into               */
/*           rssmu_writeHugeFileStart prior to call into this function.      */
/*===========================================================================*/
ViStatus _VI_FUNC rsafq_contHugeFile (ViSession instrSession,
                                      ViInt64 fileHandle, ViUInt32 offset,
                                      ViUInt32 blockLength, ViBoolean *EOFile)
{
    ViStatus    rsafq_status = VI_SUCCESS;
    ViChar      *pbuffer;
    size_t      read;
    
    pbuffer = (ViChar*) malloc (blockLength);
    
    if (pbuffer == NULL)
        return VI_ERROR_ALLOC;
    
    if (fseek ((FILE*) fileHandle, offset, SEEK_SET) != 0)
    {
        rsafq_status = VI_ERROR_FILE_IO;
        goto cleanup;
    }
    
    read = fread (pbuffer, 1, blockLength, (FILE*) fileHandle);
    
    if (read == 0)
    {
        rsafq_status = VI_ERROR_FILE_IO;
        goto cleanup;
    }
    
    if (read != blockLength)
    {
        viSetAttribute (instrSession, VI_ATTR_SEND_END_EN, VI_TRUE);
    }
    
    rsafq_status = viWrite (instrSession, (ViBuf)pbuffer, (ViUInt32)read, NULL);
    if (read != blockLength)
    {
        ViUInt16    intfType        = VI_INTF_GPIB;

        *EOFile = VI_TRUE;
        fclose ((FILE*) fileHandle);

        viGetAttribute (instrSession, VI_ATTR_INTF_TYPE, &intfType);

        if (intfType == VI_INTF_GPIB)
            viWrite (instrSession, (ViBuf)":SYST:COMM:GPIB:LTER STAN\n", 26, VI_NULL);
        
    }
    else
        *EOFile = VI_FALSE;
    
cleanup:
    if (pbuffer != NULL)
        free (pbuffer);
    return rsafq_status;    
}

/*****************************************************************************
 * Function: State Checking
 * Purpose:  This function switches state checking of the instrument (reading 
 *           of the Standard Event Register and checking it for error) status 
 *           subsystem. Driver functions are using state checking which is by 
 *           default enabled.
 *           
 *           Note:
 *           
 *           (1) In debug mode enable state checking.
 *           
 *           (2) For better bus throughput and instruments performance disable 
 *           state checking.
 *           
 *           (3) When state checking is disabled driver does not check if 
 *           correct instrument model or option is used with each of the 
 *           functions. This might cause unexpected behaviour of the 
 *           instrument.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_errorCheckState(
    ViSession   instrSession,
    ViBoolean   stateChecking
)
{
    ViStatus    rsafq_status = VI_SUCCESS;
    rsafq_instrRange instrPtr;

    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));

    instrPtr -> errorChecking = stateChecking;

    return rsafq_status;
}

/*===========================================================================*/
/* Function: Option Checking                                                 */
/* Purpose:  This function switches ON/OFF option checking of the instrument */
/*===========================================================================*/
ViStatus _VI_FUNC rsafq_optionCheckState (ViSession instrSession,
                                          ViBoolean state)
{
    ViStatus    rsafq_status = VI_SUCCESS;
    rsafq_instrRange instrPtr = NULL;

    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));

    instrPtr -> optionChecking = state;

    return rsafq_status;
    
}

/*****************************************************************************
 * Function: Write To Instrument
 * Purpose:  This function writes commands and queries to the instrument to 
 *           modify parameters and query device settings.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_writeInstrData(
    ViSession   instrSession,
    ViString    writeBuffer
)
{
    ViStatus rsafq_status = VI_SUCCESS;

    CHECKERR (viPrintf (instrSession, "%s\n", writeBuffer));

    return rsafq_status;
}

/*****************************************************************************
 * Function: Read Instrument Data
 * Purpose:  This function reads data from the instrument's output buffer and 
 *           returns it to the specified variable in memory.
 *           
 *           Notes:
 *           
 *           (1) Because the instrument may return both numeric and text data 
 *           in response to queries this function returns the data in string 
 *           format.
 *           
 *           (2) If valid data is not available at the instrument's output 
 *           buffer when this function is called the instrument will hang up 
 *           and the function will not return until it times out.  If the 
 *           time-out is disabled this function will hang indefinitely and it 
 *           may be necessary to reboot the computer to break out.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_readInstrData(
    ViSession   instrSession,
    ViInt32 numberBytesToRead,
    ViChar  readBuffer[],
    ViInt32*    numBytesRead
)
{
    ViStatus    rsafq_status = VI_SUCCESS;
    ViUInt32    tmpCnt;

    if (numBytesRead)
        *numBytesRead = 0L;

    CHECKERR (viRead (instrSession, (ViBuf)readBuffer, numberBytesToRead, &tmpCnt));

    if (numBytesRead)
        *numBytesRead = (ViInt32)tmpCnt;

    return rsafq_status;
}

/*****************************************************************************
 * Function: Reset
 * Purpose:  RESET resets the instrument to a defined default state. The 
 *           function has the same effect as pressing the PRESET key.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_reset(
    ViSession   instrSession
)
{
    ViStatus rsafq_status = VI_SUCCESS;

    /*  Initialize the instrument to a known state.  */
    CHECKERR (viPrintf (instrSession, "*RST;*CLS\n"));
    CHECKERR (rsafq_defaultInstrSetup (instrSession));

    return rsafq_status;
}

/*****************************************************************************
 * Function: Self-Test
 * Purpose:  This function runs the instrument's self test routine and returns 
 *           the test result(s).
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_self_test(
    ViSession   instrSession,
    ViInt16*    selfTestResult,
    ViChar  selfTestMessage[]
)
{
    ViStatus rsafq_status = VI_SUCCESS;

    CHECKERR( viQueryf (instrSession, "*TST?\n", "%hd", selfTestResult));

    sprintf (selfTestMessage, "%s", (*selfTestResult == 0) ? "Self-Test Passed" : "Self-Test Failed");

    return rsafq_status;
}

/*****************************************************************************
 * Function: Error-Query
 * Purpose:  This function reads an error code from the instrument's error 
 *           queue.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_error_query(
    ViSession   instrSession,
    ViInt32*    errorCode,
    ViChar  errorMessage[]
)
{
    ViStatus rsafq_status = VI_SUCCESS;

    CHECKERR (viPrintf (instrSession, ":SYST:ERR?\n"));
    CHECKERR (viScanf (instrSession, "%ld,\"%[^\"]", errorCode, errorMessage));

    return rsafq_status;
}

/*****************************************************************************
 * Function: Error Message
 * Purpose:  This function takes the Status Code returned by the instrument 
 *           driver functions, interprets it and returns it as a user readable 
 *           string.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_error_message(
    ViSession   instrSession,
    ViStatus    statusCode,
    ViChar  _VI_FAR message[]
)
{
    ViStatus rsafq_status = VI_SUCCESS;
    ViInt16 i;

    static rsafq_tStringValPair statusDescArray[] = {
        {VI_WARN_NSUP_ID_QUERY,                 "WARNING: ID Query not supported"},
        {VI_WARN_NSUP_RESET,                    "WARNING: Reset not supported"},
        {VI_WARN_NSUP_SELF_TEST,                "WARNING: Self-test not supported"},
        {VI_WARN_NSUP_ERROR_QUERY,              "WARNING: Error Query not supported"},
        {VI_WARN_NSUP_REV_QUERY,                "WARNING: Revision Query not supported"},
        {VI_ERROR_PARAMETER1,                   "ERROR_LABEL: Parameter 1 out of range"},
        {VI_ERROR_PARAMETER2,                   "ERROR_LABEL: Parameter 2 out of range"},
        {VI_ERROR_PARAMETER3,                   "ERROR_LABEL: Parameter 3 out of range"},
        {VI_ERROR_PARAMETER4,                   "ERROR_LABEL: Parameter 4 out of range"},
        {VI_ERROR_PARAMETER5,                   "ERROR_LABEL: Parameter 5 out of range"},
        {VI_ERROR_PARAMETER6,                   "ERROR_LABEL: Parameter 6 out of range"},
        {VI_ERROR_PARAMETER7,                   "ERROR_LABEL: Parameter 7 out of range"},
        {VI_ERROR_PARAMETER8,                   "ERROR_LABEL: Parameter 8 out of range"},
        {VI_ERROR_FAIL_ID_QUERY,                "ERROR_LABEL: Identification query failed"},
        {VI_ERROR_INV_RESPONSE,                 "ERROR_LABEL: Interpreting instrument response"},
        {VI_ERROR_INSTR_FILE_OPEN,              "ERROR_LABEL: Opening the specified file"},
        {VI_ERROR_INSTR_FILE_WRITE,             "ERROR_LABEL: Writing to the specified file"},
        {VI_ERROR_INSTR_INTERPRETING_RESPONSE,  "ERROR_LABEL: Interpreting the instrument's response"},
        {VI_ERROR_INSTR_PARAMETER8,             "ERROR_LABEL: Parameter 8 out of range"},
        {VI_ERROR_INSTR_PARAMETER9,             "ERROR_LABEL: Parameter 9 out of range"},
        {VI_ERROR_INSTR_PARAMETER10,            "ERROR_LABEL: Parameter 10 out of range"},
        {VI_ERROR_INSTR_PARAMETER11,            "ERROR_LABEL: Parameter 11 out of range"},
        {VI_ERROR_INSTR_PARAMETER12,            "ERROR_LABEL: Parameter 12 out of range"},
        {VI_ERROR_INSTR_PARAMETER13,            "ERROR_LABEL: Parameter 13 out of range"},
        {VI_ERROR_INSTR_PARAMETER14,            "ERROR_LABEL: Parameter 14 out of range"},
        {VI_ERROR_INSTR_PARAMETER15,            "ERROR_LABEL: Parameter 15 out of range"},

        /* --- Instrument specific error codes --- */

        {RSAFQ_ERROR_INSTRUMENT_ERROR,          "ERROR_LABEL: Instrument status error"},
        {RSAFQ_ERROR_INVALID_CONFIGURATION,     "ERROR_LABEL: Instrument configuration error"},
        {RSAFQ_ERROR_INSTRUMENT_OPTION,         "ERROR_LABEL: Option not installed"},
        {RSAFQ_ERROR_INSTRUMENT_NSUP_MODEL,     "ERROR_LABEL: Required Instrument model is not available."},
        {RSAFQ_ERROR_SETTINGS_CONFLICT,         "ERROR_LABEL: Instrument settings conflict."},
        {RSAFQ_ERROR_CALIBRATION_FAILED,        "ERROR_LABEL: Calibration failed."},
        {RSAFQ_ERROR_INSTRUMENT_NSUP_FUNC,      "ERROR_LABEL: Function/command is not supported."},

        /* --- Instrument specific warnings --- */

        {RSAFQ_WARNING_SMALL_BUFFER,            "WARNING: Buffer too small"},

        {VI_NULL, VI_NULL}
    };

    rsafq_status = viStatusDesc (instrSession, statusCode, message);
    if (rsafq_status == VI_WARN_UNKNOWN_STATUS)
        {
        for (i = 0; statusDescArray[i].stringName; i++)
            {
            if (statusDescArray[i].stringVal == statusCode)
                {
                sprintf (message, "%s", statusDescArray[i].stringName);
                return (VI_SUCCESS);
                }
            }
        sprintf (message, "Unknown Error 0x%X", statusCode);
        return (VI_WARN_UNKNOWN_STATUS);
        }

    rsafq_status = VI_SUCCESS;
    return rsafq_status;
}

/*****************************************************************************
 * Function: Revision Query
 * Purpose:  This function returns the revision numbers of the instrument 
 *           driver and instrument firmware, and tells the user with which  
 *           instrument firmware this revision of the driver is compatible.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_revision_query(
    ViSession   instrSession,
    ViChar  instrumentDriverRevision[],
    ViChar  firmwareRevision[]
)
{
    ViStatus rsafq_status = VI_SUCCESS;

    CHECKERR (viPrintf (instrSession, "*IDN?\n"));
    CHECKERR (viScanf (instrSession, "%*[^,],%*[^,],%*[^,],%[\r\n]", firmwareRevision));

    sprintf (instrumentDriverRevision, "%s", RSAFQ_REVISION);

    return rsafq_status;
}

/*****************************************************************************
 * Function: Delay
 * Purpose:  This function suspends the program for the selected number of 
 *           second.
 *****************************************************************************/
ViStatus _VI_FUNC rsafq_delay(
    ViSession   instrSession,
    ViReal64 seconds
)
{
  clock_t StartTime = clock();
  instrSession;
  
  while ((((double)(clock() - StartTime))/CLOCKS_PER_SEC) < (seconds)){
  #ifdef _WIN32
    #ifdef _CVI_
          ProcessSystemEvents ();
    #else
          Sleep(0);
    #endif
  #endif
  }
  return VI_SUCCESS;
}

/*===========================================================================*/
/* Function: Close                                                           */
/* Purpose:  This function closes the instrument.                            */
/*===========================================================================*/
ViStatus _VI_FUNC rsafq_close (ViSession instrSession)
{
    rsafq_instrRange instrPtr;
    ViSession rmSession;
    ViUInt16 intf = VI_INTF_GPIB;
    ViStatus rsafq_status = VI_SUCCESS;

    CHECKERR (viGetAttribute (instrSession, VI_ATTR_RM_SESSION, &rmSession));
    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
    CHECKERR (viGetAttribute (instrSession, VI_ATTR_INTF_TYPE, &intf));

    /* Interface specific operations */
    switch (intf)
        {
        case VI_INTF_ASRL: /* Serial Port. Go to local mode. */
            CHECKERR (viWrite (instrSession, (ViBuf)"@LOC\n", 5, VI_NULL));
            break;
/*
#ifdef VI_INTF_TCPIP

        case VI_INTF_TCPIP: // RSIB Interface. Go to local mode.
            CHECKERR (viSetAttribute (instrSession, RS_ATTR_RSIB_SRE, 0));
            break;

#endif
*/
        }

    if (instrPtr != NULL)
        free (instrPtr);

    rsafq_status = viClose (instrSession);
    viClose (rmSession);

    return rsafq_status;    
}

/*============================================================================/
/*= UTILITY ROUTINES (Non-Exportable Functions) =============================*/
/*============================================================================/

/*===========================================================================*/
/* Function: Initialize Clean Up                                             */
/* Purpose:  This function is used only by the rsafq_init function.  When   */
/*           an error is detected this function is called to close the       */
/*           open resource manager and instrument object sessions and to     */
/*           set the instrSession that is returned from rsafq_init to       */
/*           VI_NULL.                                                        */
/*===========================================================================*/
ViStatus rsafq_initCleanUp (ViSession openRMSession,
                             ViPSession openInstrSession,
                             ViStatus currentStatus)
{
    rsafq_instrRange instrPtr;

    if (viGetAttribute (*openInstrSession, VI_ATTR_USER_DATA, &instrPtr) >= 0)
        if (instrPtr != NULL)
            free (instrPtr);

    viClose (*openInstrSession);
    viClose (openRMSession);
    *openInstrSession = VI_NULL;

    return currentStatus;
}

/*============================================================================/
/*----------- INSERT INSTRUMENT-DEPENDENT UTILITY ROUTINES HERE -------------*/
/*============================================================================/

/*===========================================================================*/
/* Function: Default Instrument Setup                                        */
/* Purpose:  This function sends a default setup to the instrument.  This    */
/*           function is called by the rsafq_reset operation and by the     */
/*           rsafq_init function if the reset option has not been           */
/*           selected.  This function is useful for configuring any          */
/*           instrument settings that are required by the rest of the        */
/*           instrument driver functions such as turning headers ON or OFF   */
/*           or using the long or short form for commands, queries, and data.*/
/*===========================================================================*/
ViStatus rsafq_defaultInstrSetup (ViSession instrSession)
{
    ViChar              buffer[BUFFER_SIZE] = "";
    ViInt32             index;

    ViStatus            rsafq_status = VI_SUCCESS;
    rsafq_instrRange    instrPtr;
    ViInt32             tmp_i;
    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
    tmp_i=0;
    
    /* Determine if the structure has been initialized for the current VISA  */
    /* Session and malloc if it has not.                                     */
    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));

    instrPtr -> errorChecking = VI_TRUE;
    instrPtr -> optionChecking = VI_TRUE;
    strcpy (instrPtr -> instrDriverRevision, RSAFQ_REVISION);

    /* Check for instrument version */
    CHECKERR (viQueryf (instrSession, "*IDN?\n", "%*[^,],%[^,],%*[\r\n]", buffer));
    
    for ( index = 0; instrVersionList[index]; index++ )
        if (!strncmp(buffer, instrVersionList[index], strlen(instrVersionList[index])))
            break;
    if (!instrVersionList[index]) index--;
    instrPtr -> instrVer = index; 
    
    /* Check for instruments options */
    if ((rsafq_status = rsafq_OptionIdentificationQuery (instrSession, buffer)) < VI_SUCCESS)
        return rsafq_status;
    else
        rsafq_status = VI_SUCCESS;
    
    instrPtr->instrOptions.B10  = (strstr (buffer, "AFQ-B10") != VI_NULL);
    instrPtr->instrOptions.B11  = (strstr (buffer, "AFQ-B11") != VI_NULL);
    instrPtr->instrOptions.B18  = (strstr (buffer, "AFQ-B18") != VI_NULL);
    instrPtr->instrOptions.K0   = (strstr (buffer, "AFQ-K0") != VI_NULL);
    instrPtr->instrOptions.K80  = (strstr (buffer, "AFQ-K80") != VI_NULL);
    instrPtr->instrOptions.K6   = (strstr (buffer, "AFQ-K6") != VI_NULL);
    instrPtr->instrOptions.K240 = (strstr (buffer, "AFQ-K240") != VI_NULL);
    instrPtr->instrOptions.K242 = (strstr (buffer, "AFQ-K242") != VI_NULL);
    instrPtr->instrOptions.K243 = (strstr (buffer, "AFQ-K243") != VI_NULL);
    instrPtr->instrOptions.K244 = (strstr (buffer, "AFQ-K244") != VI_NULL);
    instrPtr->instrOptions.K245 = (strstr (buffer, "AFQ-K245") != VI_NULL);
    instrPtr->instrOptions.K246 = (strstr (buffer, "AFQ-K246") != VI_NULL);
    instrPtr->instrOptions.K248 = (strstr (buffer, "AFQ-K248") != VI_NULL);
    instrPtr->instrOptions.K249 = (strstr (buffer, "AFQ-K249") != VI_NULL);
    instrPtr->instrOptions.K250 = (strstr (buffer, "AFQ-K250") != VI_NULL);
    instrPtr->instrOptions.K251 = (strstr (buffer, "AFQ-K251") != VI_NULL);
    instrPtr->instrOptions.K256 = (strstr (buffer, "AFQ-K256") != VI_NULL);
    instrPtr->instrOptions.K261 = (strstr (buffer, "AFQ-K261") != VI_NULL);
    instrPtr->instrOptions.K262 = (strstr (buffer, "AFQ-K262") != VI_NULL);
    
    return rsafq_status;
}
